/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* map key to multiple keys, with macro recording */

#define MXK_MA_PLAY   1
#define MXK_MA_START  2
#define MXK_MA_RECORD 3

struct mxk_entry
{
  unsigned int e_type;
  unsigned int e_code;

  int e_value;
  struct mxk_entry *e_next;
};

struct mxk_array
{
  int a_record;
  int a_play;
  int a_permit; /* permit any keys, not just already defined ones to be changed */

  int a_bypass;
  int a_arm;

  int a_state;

  int a_index; /* pointer to key being recorded */
  struct mxk_entry *a_head; /* first entry of current recording */
  struct mxk_entry *a_tail; /* last entry of current recording */

  unsigned int a_size[2];
  struct mxk_entry **a_table[2];

  char *a_light; /* pattern to run in record state */

  struct mxk_client *a_vector;
  int a_count;
};

static int mxk_start_array(struct list_loop *ls, int id, int start, int stop);
static void mxk_stop_array(struct mxk_array *ma);

static int mxk_run_array(struct list_loop *ls, struct task_loop *tl, int mask);

static int mxk_light_array(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_load_array(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_permission_array(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_bypass_array(struct list_loop *ls, struct task_loop *tx, int special, char *args);

static int mxk_grow_array(struct mxk_array *ma, int press, int key);
static int mxk_fix_array(struct mxk_array *ma, struct mxk_entry *head);
static int mxk_permit_array(struct mxk_array *ma, int press, int key);
static void mxk_free_list_array(struct mxk_entry *head);

/**********************/

static int mxk_start_array(struct list_loop *ls, int id, int record, int play)
{
  struct mxk_array *ma;
  struct task_loop *tx;
  int type;
  char *rs, *ps;

  type = lookup_task_type_loop(ls, "array");
  if(type <= 0){
#ifdef DEBUG
    fprintf(stderr, "array: type not defined\n");
#endif
    return -1;
  }

  ma = malloc(sizeof(struct mxk_array));
  if(ma == NULL){
    return -1;
  }

  ma->a_record = record;
  ma->a_play = play;
  ma->a_permit = 1;

  ma->a_bypass = 0;
  ma->a_arm = 1;

  ma->a_state = MXK_MA_PLAY;
  ma->a_index = 0;
  ma->a_head = NULL;
  ma->a_tail = NULL;

  ma->a_table[0] = ma->a_table[1] = NULL;
  ma->a_size[0] = ma->a_size[1] = 0;

  ma->a_light = NULL;

  ma->a_vector = NULL;
  ma->a_count = 0;

  tx = add_tag_loop(ls, -1, ma, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_array, NULL, type, id);
  if(tx == NULL){
#ifdef DEBUG
    fprintf(stderr, "array: task creation failed\n");
#endif
    mxk_stop_array(ma);
    return -1;
  }

  rs = (record > 0) ? code_to_name(mxk_key_codes, record) : NULL;
  if(rs == NULL){
    rs = "<disabled>";
  }

  ps = (play > 0) ? code_to_name(mxk_key_codes, play) : NULL;
  if(ps == NULL){
    ps = "<disabled>";
  }

  text_loop(ls, tx, "array:%d record=%s, play=%s", id, rs, ps);

  return 0;
}

static void mxk_stop_array(struct mxk_array *ma)
{
  int i, j;
  struct mxk_entry *me;

  if(ma){
    for(j = 0; j < 2; j++){
      if(ma->a_table[j]){
        for(i = 0; i < ma->a_size[j]; i++){
          mxk_free_list_array(ma->a_table[j][i]);
        }
        ma->a_size[j] = 0;
        free(ma->a_table[j]);
        ma->a_table[j] = NULL;
      }
    }

    while(ma->a_head){ /* free stuff in flight */
      me = ma->a_head;
      ma->a_head = me->e_next;
      free(me);
    }
    ma->a_tail = NULL;

    if(ma->a_light){
      free(ma->a_light);
      ma->a_light = NULL;
    }

    mxk_shutdown_client(ma->a_vector, ma->a_count);

    free(ma);
  }
}

static int mxk_light_array(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_array *ma;

  ma = user_loop(ls, tx);
  if(ma){
    if(ma->a_light){
      free(ma->a_light);
      ma->a_light = NULL;
    }
    if(args){
      ma->a_light = copy_word(args, 0);
      if(ma->a_light == NULL){
        return -1;
      }
    }
  }

  return 0;
}

static void mxk_free_list_array(struct mxk_entry *head)
{
  struct mxk_entry *x, *y;

  x = head;
  while(x){
    y = x;
    x = x->e_next;
    free(y);
  }
}

static int mxk_permit_array(struct mxk_array *ma, int press, int key)
{
  if((press < 0) || (press > 1)){
#ifdef DEBUG
    fprintf(stderr, "array: permit: press not in range\n");
#endif
    return 0;
  }

  if(ma->a_permit){
    return 1;
  }

  if(ma->a_record == key){
#ifdef DEBUG
    fprintf(stderr, "array: permit: key matches record\n");
#endif
    return 0;
  }

  if(ma->a_play == key){
#ifdef DEBUG
    fprintf(stderr, "array: permit: key matches play\n");
#endif
    return 0;
  }

  if((key < 0) || (key > ma->a_size[press])){
#ifdef DEBUG
    fprintf(stderr, "array: permit: key outside range\n");
#endif
    return 0;
  }

  if(ma->a_table[press][key] == NULL){
#ifdef DEBUG
    fprintf(stderr, "array: permit: key without mapping\n");
#endif
    return 0;
  }

  return 1;
}

static int mxk_grow_array(struct mxk_array *ma, int press, int key)
{
  struct mxk_entry **tmp;
  int i;

  if((press < 0) || (press > 1)){
#ifdef DEBUG
    fprintf(stderr, "grow array: press out of range %d\n", press);
    abort();
#endif
    return -1;
  }

  if(key < ma->a_size[press]){
    return 0;
  }

  tmp = realloc(ma->a_table[press], sizeof(struct mxk_entry *) * (key + 1));
  if(tmp == NULL){
    return -1;
  }

  for(i = ma->a_size[press]; i <= key; i++){
    tmp[i] = NULL;
  }
  ma->a_table[press] = tmp;
  ma->a_size[press] = key + 1;

  return 0;
}

static int mxk_fix_array(struct mxk_array *ma, struct mxk_entry *head)
{
  struct mxk_entry *tail, *tmp, *current;
  int use, best, i;

  if(head == NULL){
    return 0;
  }

  for(tail = head; tail->e_next; tail = tail->e_next);

  /* O(n^2): presumably the sequences (fixed to chords) are not too long */
  for(i = 0; i < KEY_MAX; i = best){

#ifdef DEBUG
    fprintf(stderr, "fixup: looking at %d\n", i);
#endif

    best = KEY_MAX;
    use = 0;

    for(current = head; current; current = current->e_next){
      if(current->e_type == EV_KEY){
        if((current->e_code > i) && (current->e_code < best)){
          best = current->e_code;
        }
        if(current->e_code == i){
          switch(current->e_value){
            case 0 :
              use--;
              break;
            case 1 :
              use++;
              break;
          }
        }
      }
    }

    switch(use){
      case 0 :
        break;
      case 1 :

        tmp = malloc(sizeof(struct mxk_entry));
        if(tmp == NULL){
          return -1;
        }

        tmp->e_type = EV_KEY;
        tmp->e_code = i;
        tmp->e_value = 0;

        tmp->e_next = NULL;

        /* insert next to tail */
        tail->e_next = tmp;
        tail = tmp;

#ifdef DEBUG
        fprintf(stderr, "array: fixed up missing release for key=0x%x\n", i);
#endif

        break;
      default :
#ifdef DEBUG
        fprintf(stderr, "array: giving up in fixup: use is %d, key=0x%x\n", use, i);
#endif
        return -1;
    }
  }

  /* ensure the chain is terminated by a syn */
  if(tail->e_type != EV_SYN){
    tmp = malloc(sizeof(struct mxk_entry));
    if(tmp == NULL){
      return -1;
    }

    tmp->e_type = EV_SYN;
    tmp->e_code = 0;
    tmp->e_value = 0;
    tmp->e_next = NULL;

    tail->e_next = tmp;
  }

  return 0;
}

static int mxk_permission_array(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_array *ma;

  ma = user_loop(ls, tx);
  if(ma == NULL){
    return -1;
  }

  ma->a_permit = special;

  return 0;
}

static int mxk_bypass_array(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_array *ma;

  ma = user_loop(ls, tx);
  if(ma == NULL){
    return -1;
  }

  ma->a_bypass = special;

  return 0;
}

static int mxk_load_array(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  int idx, key, press, fixup;
  char buffer[MXK_NAME_SIZE];
  struct mxk_array *ma;
  struct mxk_entry *me, *pe, *he;
  struct input_event evt, *vp;

  /* sorry, special field encodes both press and if we should do fixes */
  if(special){
    press = 1;
    fixup = (special > 1) ? 1 : 0;
  } else {
    press = 0;
    fixup = 0;
  }

  ma = user_loop(ls, tx);
  if(ma == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
#ifdef DEBUG
    fprintf(stderr, "array: need key for load\n");
#endif
    return -1;
  }

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    return -1;
  }

  key = name_to_code(mxk_key_codes, buffer);
  if(key < 0){
#ifdef DEBUG
    fprintf(stderr, "array: key not found for load\n");
#endif
    return -1;
  }

  if(ma->a_index == key){ /* curtesy, updates are atomic */
#ifdef DEBUG
    fprintf(stderr, "array: key already being updated\n");
#endif
    return -1;
  }

  if(mxk_grow_array(ma, press, key)){
#ifdef DEBUG
    fprintf(stderr, "array: grow failed\n");
#endif
    return -1;
  }

  vp = &evt;
  me = NULL;
  he = NULL;

  while((idx = next_word(args, idx)) > 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
#ifdef DEBUG
      fprintf(stderr, "load: unable to aquire next array element\n");
#endif
      mxk_free_list_array(he);
      return -1;
    }

    if(mxk_parse_evt(buffer, vp) < 0){
#ifdef DEBUG
      fprintf(stderr, "load: unable to parse next array element\n");
#endif
      mxk_free_list_array(he);
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "load: loaded %d:%d:%d\n", vp->type, vp->code, vp->value);
#endif

    pe = me;
    me = malloc(sizeof(struct mxk_entry));
    if(me == NULL){
      mxk_free_list_array(he);
      return -1;
    }

    me->e_type = vp->type;
    me->e_code = vp->code;
    me->e_value = vp->value;

    me->e_next = NULL;
    if(he){
      pe->e_next = me;
    } else {
      he = me;
    }
  }


  if(fixup){
    if(mxk_fix_array(ma, he)){
      mxk_free_list_array(he);
      return -1;
    }
  }

#ifdef DEBUG
  fprintf(stderr, "load array: updating position [%d][%d]\n", press, key);
#endif
  mxk_free_list_array(ma->a_table[press][key]);
  ma->a_table[press][key] = he;

  return 0;
}

/*************************************************************************/

int mxk_start_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, ax, key[2];
  struct control_loop *cl;
  char *value, *end;
  char buffer[MXK_NAME_SIZE];
  struct mxk_global *mg;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  idx = (-1);
  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, &end, 10);
      switch(end[0]){
        case '\0'  :
        case '\r'  :
        case '\n'  :
        case ' '   :
          break;
        default :
          print_control(cl, cd, 1, "array: first parameter has to be an integer\n");
          return 0;
      }
      idx = next_word(args, idx);
    }
  }

  ax = 0;
  key[1] = key[0] = 0;

  while((idx > 0) && (ax < 2)){

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      print_control(cl, cd, 1, "array: parameter seems too long\n");
      return 0;
    }

    key[ax] = name_to_code(mxk_key_codes, buffer);
    if(key[ax] < 0){
      print_control(cl, cd, 1, "array: unknown keycode name %s\n", buffer);
      return 0;
    }

    idx = next_word(args, idx);
    ax++;

  }

  if(mxk_start_array(ls, id, key[0], key[1])){
    print_control(cl, cd, 1, "array: unable to start task for id %d\n", id);
  } else {
    if(mg->g_verbose > 1){
      print_control(cl, cd, 1, "array: created id %d\n", id);
    }
  }

  return 0;
}

int mxk_bypass_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_OPTIONAL, 1, mxk_bypass_array);

  return 0;
}

int mxk_ignore_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_OPTIONAL, 0, mxk_bypass_array);

  return 0;
}

int mxk_any_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_OPTIONAL, 1, mxk_permission_array);

  return 0;
}

int mxk_preset_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_OPTIONAL, 0, mxk_permission_array);

  return 0;
}

int mxk_release_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_REQUIRED, 0, mxk_load_array);

  return 0;
}

int mxk_press_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_REQUIRED, 1, mxk_load_array);

  return 0;
}

int mxk_quick_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_REQUIRED, 2, mxk_load_array);

  return 0;
}

int mxk_list_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, j, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_array *ma;
  struct mxk_entry *me;
  char *value;
  char *record, *play, *key;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "array");
  if(type <= 0){
    print_control(cl, cd, 1, "array: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
    }
  }

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    count = 0;
    ma = user_loop(ls, tx);
    if(ma){
      if(ma->a_record > 0){
        record = code_to_name(mxk_key_codes, ma->a_record);
        if(record == NULL){
          record = "<unknown>";
        }
      } else {
        record = "<disabled>";
      }
      if(ma->a_play > 0){
        play = code_to_name(mxk_key_codes, ma->a_play);
        if(play == NULL){
          play = "<unknown>";
        }
      } else {
        play = "<disabled>";
      }
      print_control(cl, cd, 0, "array[%d]: record=%s, play=%s, state=%s, permit=%s, unknown=%s\n", tag_loop(ls, tx), record, play, (ma->a_state == MXK_MA_PLAY) ? "play" : "record", ma->a_permit ? "any" : "preset", ma->a_bypass ? "bypass" : "ignore", ma->a_bypass ? "bypass" : "discard");
      for(j = 0; j < 2; j++){
        for(i = 0; i < ma->a_size[j]; i++){
          if(ma->a_table[j][i]){
            count++;
            key = code_to_name(mxk_key_codes, i);
            if(j > 0){
              print_control(cl, cd, 0, "press ");
            } else {
              print_control(cl, cd, 0, "release ");
            }
            if(key){
              print_control(cl, cd, 0, "%s:", key);
            } else {
              print_control(cl, cd, 0, "0x%x:", i);
            }

            me = ma->a_table[j][i];
            while(me){
              print_control(cl, cd, 0, " ");
              mxk_print_evt(cl, cd, me->e_type, me->e_code, me->e_value);
              me = me->e_next;
            }

            print_control(cl, cd, 0, "\n");
          }
        }
      }
      print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");
    } else {
      print_control(cl, cd, 1, "array[%d]: unavailable\n", id);
    }
  }

  return 0;
}

int mxk_light_array_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "array", MXK_MISC_OPTIONAL, 0, mxk_light_array);

  return 0;
}

/*************************************************************************/

void mxk_state_array(struct list_loop *ls, struct task_loop *tl, int state)
{
  struct mxk_array *ma;

  ma = user_loop(ls, tl);

  if(state == ma->a_state){
    return;
  }

  switch(state){
    case MXK_MA_PLAY :
      if(ma->a_light){
        mxk_stop_light(ls, ma->a_light);
      }
      if(ma->a_play){
        ma->a_arm = 0;
      } else {
        ma->a_arm = 1;
      }
#ifdef DEBUG
      if(ma->a_head){
        fprintf(stderr, "array state: problem: something recording while playing\n");
      }
#endif
      break;
    case MXK_MA_START :
      ma->a_arm = 0;
      if(ma->a_light){
        mxk_start_light(ls, ma->a_light);
      }
      break;
    case MXK_MA_RECORD :
      break;
  }

  ma->a_state = state;
}

static int mxk_recording_array(struct list_loop *ls, struct task_loop *tl, struct input_event *vp)
{
  struct mxk_array *ma;

  ma = user_loop(ls, tl);
  switch(ma->a_state){
    case MXK_MA_PLAY :
      if((vp->type == EV_KEY) && (ma->a_record > 0) && (ma->a_record == vp->code)){ /* start */
#ifdef DEBUG
        fprintf(stderr, "array: recording: about to enter record state with key 0x%02x\n", vp->code);
#endif
        return 1;
      } else { /* playing */
        return 0;
      }
    default : /* busy recording */
      return 1;
  }
}

static int mxk_record_array(struct list_loop *ls, struct task_loop *tl, struct input_event *vp)
{
  struct mxk_array *ma;
  struct mxk_entry *me;

  ma = user_loop(ls, tl);
  switch(ma->a_state){

    case MXK_MA_PLAY : /* we have seen record key, now see if it has been released */
      if((vp->type == EV_KEY) && (vp->code == ma->a_record) && (vp->value == 0)){ /* on release */
#ifdef DEBUG
        fprintf(stderr, "array: got 0x%x release: about to record\n", vp->code);
#endif
        mxk_state_array(ls, tl, MXK_MA_START);
#ifdef DEBUG
      } else {
        fprintf(stderr, "array: got 0x%x hit: waiting for release\n", vp->code);
#endif
      }
      break;


    case MXK_MA_START : /* collect the key to be matched */

#ifdef DEBUG
      if(ma->a_index > 0){
        fprintf(stderr, "array: state problem: index set\n");
        abort();
      }
      if(ma->a_head || ma->a_tail){
        fprintf(stderr, "array: state problem: already recording\n");
        abort();
      }
      fprintf(stderr, "array: record: waiting for start: this event type=0x%02x, code=0x%02x, value=%d\n", vp->type, vp->code, vp->value);
#endif

      if((vp->type == EV_KEY) && (vp->value == 0)){ /* on release, otherwise we end up recording the release */
        if(mxk_permit_array(ma, 1, vp->code) && (mxk_grow_array(ma, 1, vp->code) == 0)){
          mxk_state_array(ls, tl, MXK_MA_RECORD);
          ma->a_index = vp->code;
#ifdef DEBUG
          fprintf(stderr, "array: recording for slot 0x%x\n", ma->a_index);
#endif
        } else {
#ifdef DEBUG
          fprintf(stderr, "arrray: problems in growing record, stopping\n");
#endif
          mxk_state_array(ls, tl, MXK_MA_PLAY);
        }
      }
      break;

    case MXK_MA_RECORD : /* collect the keys to be replayed */
      if((vp->type == EV_KEY) && (ma->a_record == vp->code)){

        if(vp->value == 0){
#ifdef DEBUG
          fprintf(stderr, "array: finishing record for 0x%x\n", ma->a_index);
#endif

          if(mxk_fix_array(ma, ma->a_head)){
#ifdef DEBUG
            fprintf(stderr, "array: fix failed\n");
            fprintf(stderr, "array: sequence:");
            for(me = ma->a_head; me; me = me->e_next){
              fprintf(stderr, " 0x%02x:0x%02x:%d", me->e_type, me->e_code, me->e_value);
            }
            fprintf(stderr, "\n");
#endif
            mxk_free_list_array(ma->a_head);
          } else {
            mxk_free_list_array(ma->a_table[1][ma->a_index]);
            ma->a_table[1][ma->a_index] = ma->a_head;
          }

          ma->a_head = NULL;
          ma->a_tail = NULL;
          ma->a_index = 0;

          mxk_state_array(ls, tl, MXK_MA_PLAY);
        }
      } else {
#ifdef DEBUG
        if(ma->a_index >= ma->a_size[1]){
          fprintf(stderr, "array: problem: index %d out of range %d\n", ma->a_index, ma->a_size[1]);
          abort();
        }
#endif

        mxk_send_client(ls, tl, ma->a_vector, ma->a_count, vp->type, vp->code, vp->value);

        switch(vp->type){
          case EV_LED : /* we don't record LED status stuff */
            return 0;
          case EV_SYN : /* don't record duplicate sync's either */
            me = ma->a_tail;
            if(me){
              if(me->e_type == EV_SYN){
                if((me->e_code == vp->code) && (me->e_value == vp->value)){
                  return 0;
                }
              }
            }
            break;
        }

        me = malloc(sizeof(struct mxk_entry));
        if(me){
          me->e_type = vp->type;
          me->e_code = vp->code;
          me->e_value = vp->value;
#ifdef DEBUG
          fprintf(stderr, "array: recording 0x%02x:0x%02x:%d\n", me->e_type, me->e_code, me->e_value);
#endif
          me->e_next = NULL;

          if(ma->a_head == NULL){
            ma->a_head = me;
          } else {
            ma->a_tail->e_next = me;
          }
          ma->a_tail = me;

        } else {
#ifdef DEBUG
          fprintf(stderr, "arrray: problems in record, stopping prematurely\n");
#endif

          mxk_free_list_array(ma->a_head);
          ma->a_head = NULL;
          ma->a_tail = NULL;
          ma->a_index = 0;

          mxk_state_array(ls, tl, MXK_MA_PLAY);
        }
      }
      break;
  }

  return 0;
}

static int mxk_play_array(struct list_loop *ls, struct task_loop *tl, struct input_event *vp)
{
  struct mxk_array *ma;
  struct mxk_entry *me;
  int bypass, press;

  bypass = 0;

  ma = user_loop(ls, tl);
  me = NULL;
  bypass = ma->a_bypass;

  if(ma->a_arm){ /* ready to play array */
    if(ma->a_play > 0){ /* disarm if play prefix key has been set */
      ma->a_arm = 0;
    }
    press = vp->value;
    if((press >= 0) && (press <= 1)){
      if(vp->code < ma->a_size[press]){
        me = ma->a_table[press][vp->code];
      }
#if 0
      if(me == NULL){
        bypass = ma->a_arm ? 0 : ma->a_bypass;
      }
#endif
#ifdef DEBUG
      fprintf(stderr, "array: match[0x%x]=%p, value=%d\n", vp->code, me, vp->value);
#endif
    }
  } else { /* unarmed */
    if(ma->a_play == vp->code){
      ma->a_arm = 1; /* found play macro key, arm */
      bypass = 0;
    }
  }

  if(me){
#ifdef DEBUG
    fprintf(stderr, "array: relaying: start 0x%x:%d:%d\n", me->e_type, me->e_code, me->e_value);
#endif
    while(me){
      mxk_send_client(ls, tl, ma->a_vector, ma->a_count, me->e_type, me->e_code, me->e_value);
      me = me->e_next;
    }
  } else {
    if(bypass){
      mxk_send_client(ls, tl, ma->a_vector, ma->a_count, vp->type, vp->code, vp->value);
#ifdef DEBUG
    } else {
      fprintf(stderr, "array: no bypass, discarding %d\n", vp->code);
#endif
    }
  }

  return 0;
}

static int mxk_run_array(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event *vp;
  struct link_loop *nk;
  struct mxk_array *ma;

  ma = user_loop(ls, tl);

  if(mask & LO_STOP_MASK){
    mxk_state_array(ls, tl, MXK_MA_PLAY);
    mxk_stop_array(ma);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      vp = nk->k_data;
      if(vp){
        if(mxk_recording_array(ls, tl, vp)){
          mxk_record_array(ls, tl, vp);
        } else {
          mxk_play_array(ls, tl, vp);
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_array(void *array, int type, int id)
{
  struct mxk_array *ma;
  struct mxk_client *mc;

  ma = array;

  mc = mxk_insert_client(ma->a_vector, ma->a_count, type, id);
  if(mc == NULL){
    return -1;
  }

  ma->a_vector = mc;
  ma->a_count++;

  return 0;
}

int mxk_startup_array(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "array", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_array);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_array(struct mxk_global *mg)
{
}
