/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* single bit state entity */

#define MXK_BIT_NULL    0x00

#define MXK_BIT_UP      0x01 /* symbol sets state */
#define MXK_BIT_DOWN    0x02 /* clears state */
#define MXK_BIT_TOGGLE  0x03 /* toggles state */

#define MXK_BIT_PRESS   0x04 /* count a press */
#define MXK_BIT_RELEASE 0x08 /* count a release */
#define MXK_BIT_HIT     0x0c /* both count */

#define MXK_BIT_HIPASS  0x10 /* gets sent along if state high */
#define MXK_BIT_LOPASS  0x20 /* gets sent along if state low */
#define MXK_BIT_PASS    0x30 /* gets sent along */

#define MXK_BIT_INVALID 0xff /* use default instead */

struct mxk_bit
{
  int b_state; /* what are we */

  int b_up; /* symbol to emit on entering 1 */
  int b_down; /* symbol for entering 0 */

  int b_sticky; /* only release b_up when state becomes down, ignore b_down */
 
  int b_size;
  unsigned char *b_map; /* zero is default */

  char *b_light; /* pattern to run in up state */

  struct mxk_client *b_vector;
  int b_count;
};

static struct code_name mxk_bit_codes[] = {
  { "up",      MXK_BIT_UP },
  { "down",    MXK_BIT_DOWN },
  { "toggle",  MXK_BIT_TOGGLE },

  { "press",   MXK_BIT_PRESS },
  { "release", MXK_BIT_RELEASE },
  { "hit",     MXK_BIT_HIT },

  { "hipass",  MXK_BIT_HIPASS },
  { "lopass",  MXK_BIT_LOPASS },
  { "pass",    MXK_BIT_PASS },

  { NULL, -1 }
};

static int mxk_start_bit(struct list_loop *ls, int id, int up, int down, int sticky);
static void mxk_stop_bit(struct mxk_bit *mb);

static int mxk_insert_bit(void *bit, int type, int id);
static int mxk_run_bit(struct list_loop *ls, struct task_loop *tl, int mask);

static int mxk_start_mode_bit(struct list_loop *ls, struct task_loop *tl, char *args, int mode);

static int mxk_light_bit(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_key_bit(struct list_loop *ls, struct task_loop *tx, int special, char *args);

/**************************************/

#define MXK_SB_UP      1
#define MXK_SB_DOWN    2
#define MXK_SB_STICKY  3
#define MXK_SB_PAIR    4

int mxk_start_up_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_start_mode_bit(ls, tl, args, MXK_SB_UP);

  return 0;
}

int mxk_start_down_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_start_mode_bit(ls, tl, args, MXK_SB_DOWN);

  return 0;
}

int mxk_start_sticky_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_start_mode_bit(ls, tl, args, MXK_SB_STICKY);

  return 0;
}

int mxk_start_pair_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_start_mode_bit(ls, tl, args, MXK_SB_PAIR);

  return 0;
}

int mxk_light_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "bit", MXK_MISC_OPTIONAL, 0, mxk_light_bit);

  return 0;
}

int mxk_key_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "bit", MXK_MISC_REQUIRED, 0, mxk_key_bit);

  return 0;
}

int mxk_default_bit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "bit", MXK_MISC_OPTIONAL, 1, mxk_key_bit);

  return 0;
}

/***************************************************************/

static int mxk_key_bit(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_bit *mb;
  char buffer[MXK_NAME_SIZE];
  int key, flags, idx, i;
  unsigned char *tmp;

  idx = start_word(args, 0);
  mb = user_loop(ls, tx);

  if(special){ /* default case */
    key = 0;
  } else { /* figure out a key */

    if(idx < 0){
      return -1;
    }

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key <= 0){
      return -1;
    }

    idx = next_word(args, idx);
  }

#ifdef DEBUG
  fprintf(stderr, "key bit: key is %d, size is %d\n", key, mb->b_size);
#endif

  if(key >= mb->b_size){
    tmp = realloc(mb->b_map, sizeof(unsigned char) * (key + 1));
    if(tmp == NULL){
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "key bit: resizing, marking %d to %d inclusive\n", mb->b_size, key);
#endif

    for(i = mb->b_size; i <= key; i++){
      tmp[i] = MXK_BIT_INVALID;
    }

    mb->b_map = tmp;
    mb->b_size = key + 1;
  }

  flags = MXK_BIT_NULL;

  while(idx >= 0){

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "key bit: considering flag %s\n", buffer);
#endif

    i = name_to_code(mxk_bit_codes, buffer);
    if(i < 0){
      return -1;
    }

    flags |= i;

    idx = next_word(args, idx);
  }

  mb->b_map[key] = flags;

#ifdef DEBUG
  fprintf(stderr, "key bit: set map[%d] to 0x%02x\n", key, mb->b_map[key]);
#endif

  return 0;
}

static int mxk_light_bit(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_bit *mb;

  mb = user_loop(ls, tx);
  if(mb){
    if(mb->b_light){
      free(mb->b_light);
      mb->b_light = NULL;
    }
    if(args){
      mb->b_light = copy_word(args, 0);
      if(mb->b_light == NULL){
        return -1;
      }
    }
  }

  return 0;
}

static int mxk_start_mode_bit(struct list_loop *ls, struct task_loop *tl, char *args, int mode)
{
  int cd, idx, id, up, down, sticky, tmp;
  struct control_loop *cl;
  char *value, *end;
  char buffer[MXK_NAME_SIZE];

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  if(args == NULL){
    print_control(cl, cd, 1, "bit: need parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "bit: need nonspace parameters\n");
    return 0;
  }

  value = args + idx;

  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "bit: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  up = (-1);
  down = (-1);
  sticky = 0;

  idx = next_word(args, idx);
  if(idx <= 0){
    print_control(cl, cd, 1, "bit: expected a second parameter\n");
    return 0;
  }

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, cd, 1, "bit: parameter seems too long\n");
    return 0;
  }

  tmp = name_to_code(mxk_key_codes, buffer);
  if(tmp < 0){
    print_control(cl, cd, 1, "bit: unknown keycode name %s\n", buffer);
    return 0;
  }

  switch(mode){
    case MXK_SB_UP      :
      up = tmp;
      break;
    case MXK_SB_DOWN    :
      down = tmp;
      break;
    case MXK_SB_STICKY  :
      up = tmp;
      sticky = 1;
      break;
    case MXK_SB_PAIR    :
      up = tmp;

      idx = next_word(args, idx);
      if(idx <= 0){
        print_control(cl, cd, 1, "bit: insufficient parameters\n");
        return 0;
      }

      if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
        print_control(cl, cd, 1, "bit: parameter seems too long\n");
        return 0;
      }

      down = name_to_code(mxk_key_codes, buffer);
      if(down < 0){
        print_control(cl, cd, 1, "bit: unknown keycode name %s\n", buffer);
        return 0;
      }
      break;
  }

  if(mxk_start_bit(ls, id, up, down, sticky)){
    print_control(cl, cd, 1, "bit: unable to start task for id %d\n", id);
  } else {
    print_control(cl, cd, 1, "bit: created id %d\n", id);
  }

  return 0;
}

static int mxk_start_bit(struct list_loop *ls, int id, int up, int down, int sticky)
{
  struct mxk_bit *mb;
  struct task_loop *tx;
  int type;

  if((down >= 0) && (sticky)){
#ifdef DEBUG
    fprintf(stderr, "bit: conflict: sticky excludes down\n");
#endif
    return -1;
  }

  if((sticky == 0) && (up < 0)){
#ifdef DEBUG
    fprintf(stderr, "bit: conflict: sticky requires up\n");
#endif
    return -1;
  }

  type = lookup_task_type_loop(ls, "bit");
  if(type <= 0){
#ifdef DEBUG
    fprintf(stderr, "bit: type not defined\n");
#endif
    return -1;
  }

  mb = malloc(sizeof(struct mxk_bit));
  if(mb == NULL){
    return -1;
  }

  mb->b_state = 0;

  mb->b_up = up;
  mb->b_down = down;
  mb->b_sticky = sticky;

  mb->b_light = NULL;
  mb->b_map = NULL;
  mb->b_size = 0;

  mb->b_vector = NULL;
  mb->b_count = 0;

  mb->b_map = malloc(sizeof(unsigned char));
  if(mb->b_map == NULL){
    mxk_stop_bit(mb);
    return -1;
  }

  mb->b_size = 1;
  mb->b_map[0] = MXK_BIT_NULL;

#ifdef DEBUG
  fprintf(stderr, "start bit: size is %d\n", mb->b_size);
#endif

  tx = add_tag_loop(ls, -1, mb, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_bit, NULL, type, id);
  if(tx == NULL){
#ifdef DEBUG
    fprintf(stderr, "bit: task creation failed\n");
#endif
    mxk_stop_bit(mb);
    return -1;
  }

  if(mb->b_sticky){
    text_loop(ls, tx, "bit:%d sticky=%d", id, mb->b_up);
  } else {
    text_loop(ls, tx, "bit:%d %d/%d", id, mb->b_up, mb->b_down);
  }

  return 0;
}


static void mxk_stop_bit(struct mxk_bit *mb)
{
  if(mb){
    if(mb->b_map){
      free(mb->b_map);
      mb->b_map = NULL;
    }
    mb->b_size = 0;

    if(mb->b_light){
      free(mb->b_light);
      mb->b_light = NULL;
    }

    mxk_shutdown_client(mb->b_vector, mb->b_count);

    free(mb);
  }
}

static int mxk_run_bit(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event *vp;
  struct mxk_bit *mb;
  struct mxk_global *mg;
  struct link_loop *nk;
  int flags, state;

  mb = user_loop(ls, tl);
  mg = global_loop(ls);

  if(mask & LO_STOP_MASK){
    if(mb->b_state){
      if(mb->b_light){
        mxk_stop_light(ls, mb->b_light);
      }
    }
    mxk_stop_bit(mb);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_KEY :
            if(mb->b_size > vp->code){
              flags = mb->b_map[vp->code];
              if(flags == MXK_BIT_INVALID){
                flags = mb->b_map[0];
              }
            } else {
              flags = mb->b_map[0];
            }

#ifdef DEBUG
            fprintf(stderr, "bit[%d]: code=0x%x(size=0x%x), flags=0x%02x\n", tag_loop(ls, tl), vp->code, mb->b_size, flags);
#endif

            if((vp->value) && !(flags & MXK_BIT_PRESS)){
              flags = MXK_BIT_NULL;
            }
            if((vp->value == 0) && !(flags & MXK_BIT_RELEASE)){
              flags = MXK_BIT_NULL;
            }

            switch(flags & MXK_BIT_TOGGLE){
              case MXK_BIT_UP :
                state = 1;
                break;
              case MXK_BIT_DOWN :
                state = 0;
                break;
              case MXK_BIT_TOGGLE :
                state = 1 - mb->b_state;
                break;
              /* case MXK_BIT_NULL : */
              default :
                state = mb->b_state;
                break;
            }

            if(state && (mb->b_state == 0)){ /* up transition */
              if(mb->b_up > 0){
                mxk_send_client(ls, tl, mb->b_vector, mb->b_count, EV_KEY, mb->b_up, 1);
                if(mb->b_sticky == 0){
                  mxk_send_client(ls, tl, mb->b_vector, mb->b_count, EV_KEY, mb->b_up, 0);
                }
              }
              if(mb->b_light){
                mxk_start_light(ls, mb->b_light);
              }
            }

            if((state == 0) && mb->b_state){ /* down transition */
              if(mb->b_sticky){ /* release up */
                mxk_send_client(ls, tl, mb->b_vector, mb->b_count, EV_KEY, mb->b_up, 0);
              } else if(mb->b_down > 0){ /* toggle down */
                mxk_send_client(ls, tl, mb->b_vector, mb->b_count, EV_KEY, mb->b_down, 1);
                mxk_send_client(ls, tl, mb->b_vector, mb->b_count, EV_KEY, mb->b_down, 0);
              }
              if(mb->b_light){
                mxk_stop_light(ls, mb->b_light);
              }
            }

            mb->b_state = state;

            if(((flags & MXK_BIT_HIPASS) && mb->b_state) ||
               ((flags & MXK_BIT_LOPASS) && (mb->b_state == 0))){
              if((flags & MXK_BIT_HIT) == MXK_BIT_HIT){
                mxk_send_client(ls, tl, mb->b_vector, mb->b_count, vp->type, vp->code, vp->value);
#ifdef DEBUG
              } else {
                fprintf(stderr, "bit[%d]: logic problem: refusing to pass through one half of press/release for code %d\n", tag_loop(ls, tl), vp->code);
                abort();
#endif
              }
            }
          break;
          default :
          /* TODO ? */
#ifdef DEBUG
            fprintf(stderr, "bit: ignoring event type %d\n", vp->type);
#endif
          break;
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_bit(void *bit, int type, int id)
{
  struct mxk_bit *mb;
  struct mxk_client *mc;

  mb = bit;

  mc = mxk_insert_client(mb->b_vector, mb->b_count, type, id);
  if(mc == NULL){
    return -1;
  }

  mb->b_vector = mc;
  mb->b_count++;

  return 0;
}

int mxk_startup_bit(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "bit", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_bit);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_bit(struct mxk_global *mg)
{
}
