/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* split stream into two */

#define MXK_BRANCH_NULL    0x00

#define MXK_BRANCH_UP      0x01 /* symbol sets state */
#define MXK_BRANCH_DOWN    0x02 /* clears state */
#define MXK_BRANCH_TOGGLE  0x03 /* toggles state */

#define MXK_BRANCH_HIPASS  0x10 /* gets sent along if state high */
#define MXK_BRANCH_LOPASS  0x20 /* gets sent along if state low */
#define MXK_BRANCH_PASS    0x30 /* gets sent along */

#define MXK_BRANCH_INVALID 0xff /* use default instead */

struct mxk_branch
{
  int b_state; /* what are we */
  int b_track; /* limited protection against sending releases to different branch */
  int b_defer;

  int b_size;
  unsigned char *b_map; /* zero is default */

  char *b_light[2]; /* pattern to run */
  struct mxk_client *b_vector[2];
  int b_count[2];
};

static struct code_name mxk_branch_codes[] = {
  { "up",      MXK_BRANCH_UP },
  { "down",    MXK_BRANCH_DOWN },
  { "toggle",  MXK_BRANCH_TOGGLE },

  { "hipass",  MXK_BRANCH_HIPASS },
  { "lopass",  MXK_BRANCH_LOPASS },
  { "pass",    MXK_BRANCH_PASS },

  { NULL, -1 }
};

static void mxk_stop_branch(struct mxk_branch *mb);

static int mxk_insert_branch(void *branch, int type, int id);
static int mxk_run_branch(struct list_loop *ls, struct task_loop *tl, int mask);

static int mxk_light_branch(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_state_branch(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_key_branch(struct list_loop *ls, struct task_loop *tx, int special, char *args);

/**************************************/

int mxk_start_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, i;
  struct control_loop *cl;
  char *value;
  struct task_loop *tx;
  struct mxk_branch *mb;
  struct mxk_global *mg;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "branch");
  if(type <= 0){
    print_control(cl, cd, 1, "branch: unable to lookup type for branch\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
#ifdef DEBUG
      fprintf(stderr, "branch: key is %d,%d\n", type, id);
#endif
    }
  }

  mb = malloc(sizeof(struct mxk_branch));
  if(mb == NULL){
    print_control(cl, cd, 1, "branch: unable to allocate branch structure\n");
    return 0;
  }

  mb->b_state = 0;

  mb->b_map = NULL;
  mb->b_size = 0;
  mb->b_track = 0;
  mb->b_defer = 0;

  for(i = 0; i < 2; i++){
    mb->b_light[i] = NULL;
    mb->b_vector[i] = NULL;
    mb->b_count[i] = 0;
  }

  mb->b_map = malloc(sizeof(unsigned char));
  if(mb->b_map == NULL){
    print_control(cl, cd, 1, "branch: unable to allocate branch table\n");
    mxk_stop_branch(mb);
    return 0;
  }

  mb->b_size = 1;
  mb->b_map[0] = MXK_BRANCH_NULL;

#ifdef DEBUG
  fprintf(stderr, "start branch: size is %d\n", mb->b_size);
#endif

  tx = add_tag_loop(ls, -1, mb, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_branch, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "branch: unable to create branch task\n");
    mxk_stop_branch(mb);
    return 0;
  }

  text_loop(ls, tx, "branch:%d", id);
  if(mg->g_verbose > 1){
    print_control(cl, cd, 1, "branch: created id %d\n", id);
  }

  return 0;
}

/**************************************/

int mxk_light_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "branch", MXK_MISC_OPTIONAL, 1, mxk_light_branch);

  return 0;
}

int mxk_dark_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "branch", MXK_MISC_OPTIONAL, 0, mxk_light_branch);

  return 0;
}

int mxk_up_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "branch", MXK_MISC_OPTIONAL, 1, mxk_state_branch);

  return 0;
}

int mxk_down_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "branch", MXK_MISC_OPTIONAL, 0, mxk_state_branch);

  return 0;
}

int mxk_key_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "branch", MXK_MISC_REQUIRED, 0, mxk_key_branch);

  return 0;
}

int mxk_default_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "branch", MXK_MISC_OPTIONAL, 1, mxk_key_branch);

  return 0;
}

int mxk_list_branch_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, j, type, count, flags;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_branch *mb;
  char *value, *key;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "branch");
  if(type <= 0){
    print_control(cl, cd, 1, "branch: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
    }
  }

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    count = 0;
    mb = user_loop(ls, tx);
    if(mb){
      print_control(cl, cd, 0, "branch[%d]: state=%s pattern[up]=%s, pattern[down]=%s pressed=%d\n", tag_loop(ls, tx), mb->b_state ? "up" : "down", mb->b_light[1] ? mb->b_light[1] : "<not set>", mb->b_light[0] ? mb->b_light[0] : "<not set>", mb->b_track);
      for(i = 0; i < mb->b_size; i++){
        if(mb->b_map[i] != MXK_BRANCH_INVALID){

          if(i){
            key = code_to_name(mxk_key_codes, i);
            if(key){
              print_control(cl, cd, 0, "key %s:", key);
            } else {
              print_control(cl, cd, 0, "key 0x%x:", i);
            }
            count++;
          } else {
            print_control(cl, cd, 0, "default action:");
          }
          if(mb->b_map[i]){
            flags = mb->b_map[i];
            for(j = 0; mxk_branch_codes[j].name; j++){
              if((mxk_branch_codes[j].code & flags) == mxk_branch_codes[j].code){
                print_control(cl, cd, 0, " %s", mxk_branch_codes[j].name);
                flags &= ~(mxk_branch_codes[j].code);
              }
            }
          } else {
            print_control(cl, cd, 0, " <none>");
          }
          print_control(cl, cd, 0, "\n");

        }
      }
      print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");
    } else {
      print_control(cl, cd, 1, "branch[%d]: unavailable\n", id);
    }
  }

  return 0;
}

/***************************************************************/

static int mxk_light_branch(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_branch *mb;
  char *string;

  mb = user_loop(ls, tx);
  if(mb == NULL){
    return -1;
  }

#ifdef DEBUG
  if((special < 0) || (special > 1)){
    fprintf(stderr, "branch: logic failure: special has to be a binary value\n");
  }
#endif

  string = mb->b_light[special];

  if(string){
    free(string);
  }

  if(args){
    string = copy_word(args, 0);
    if(string == NULL){
      return -1;
    }
  }

  mb->b_light[special] = string;

  return 0;
}

static int mxk_state_branch(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_branch *mb;

  mb = user_loop(ls, tx);
  if(mb == NULL){
    return -1;
  }

#ifdef DEBUG
  if((special < 0) || (special > 1)){
    fprintf(stderr, "branch: logic failure: special has to be a binary value\n");
  }
#endif

  mb->b_defer = special;
  if(mb->b_track){
    return 0;
  }

  if(mb->b_state != mb->b_defer){
    mb->b_state = mb->b_defer;

    if(mb->b_light[1 - mb->b_state]){
      mxk_stop_light(ls, mb->b_light[1 - mb->b_state]);
    }
    if(mb->b_light[mb->b_state]){
      mxk_start_light(ls, mb->b_light[mb->b_state]);
    }
  }

  return 0;
}

static int mxk_key_branch(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_branch *mb;
  char buffer[MXK_NAME_SIZE];
  int key, flags, idx, i;
  unsigned char *tmp;

  idx = start_word(args, 0);
  mb = user_loop(ls, tx);

  if(special){ /* default case */
    key = 0;
  } else { /* figure out a key */

    if(idx < 0){
      return -1;
    }

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key <= 0){
      return -1;
    }

    idx = next_word(args, idx);
  }

#ifdef DEBUG
  fprintf(stderr, "key branch: key is %d, size is %d\n", key, mb->b_size);
#endif

  if(key >= mb->b_size){
    tmp = realloc(mb->b_map, sizeof(unsigned char) * (key + 1));
    if(tmp == NULL){
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "key branch: resizing, marking %d to %d inclusive\n", mb->b_size, key);
#endif

    for(i = mb->b_size; i <= key; i++){
      tmp[i] = MXK_BRANCH_INVALID;
    }

    mb->b_map = tmp;
    mb->b_size = key + 1;
  }

  flags = MXK_BRANCH_NULL;

  while(idx >= 0){

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "key branch: considering flag %s\n", buffer);
#endif

    i = name_to_code(mxk_branch_codes, buffer);
    if(i < 0){
      return -1;
    }

    flags |= i;

    idx = next_word(args, idx);
  }

  mb->b_map[key] = flags;

#ifdef DEBUG
  fprintf(stderr, "key branch: set map[%d] to 0x%02x\n", key, mb->b_map[key]);
#endif

  return 0;
}

/***************************************************************/

static void mxk_stop_branch(struct mxk_branch *mb)
{
  int i;

  if(mb){
    if(mb->b_map){
      free(mb->b_map);
      mb->b_map = NULL;
    }
    mb->b_size = 0;

    for(i = 0; i < 2; i++){
      if(mb->b_light[i]){
        free(mb->b_light[i]);
        mb->b_light[i] = NULL;
      }
      mxk_shutdown_client(mb->b_vector[i], mb->b_count[i]);
    }

    free(mb);
  }
}

static int mxk_run_branch(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event *vp;
  struct mxk_branch *mb;
  struct mxk_global *mg;
  struct link_loop *nk;
  int flags;

  mb = user_loop(ls, tl);
  mg = global_loop(ls);

#ifdef DEBUG
  if((mb->b_state < 0) || (mb->b_state > 1)){
    fprintf(stderr, "branch: logic problem: state=%d\n", mb->b_state);
    abort();
  }
#endif

  if(mask & LO_STOP_MASK){
    if(mb->b_light[mb->b_state]){
      mxk_stop_light(ls, mb->b_light[mb->b_state]);
    }
    mxk_stop_branch(mb);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_KEY :

            /* decide which flags to use */
            if(mb->b_size > vp->code){
              flags = mb->b_map[vp->code];
              if(flags == MXK_BRANCH_INVALID){
                flags = mb->b_map[0];
              }
            } else {
              flags = mb->b_map[0];
            }

#ifdef DEBUG
            fprintf(stderr, "branch[%d]: code=0x%x(size=0x%x), value=%d, flags=0x%02x\n", tag_loop(ls, tl), vp->code, mb->b_size, vp->value, flags);
#endif

            /* recompute desired state */
            if(vp->value == 1){
              switch(flags & MXK_BRANCH_TOGGLE){
                case MXK_BRANCH_UP :
                  mb->b_defer = 1;
                  break;
                case MXK_BRANCH_DOWN :
                  mb->b_defer = 0;
                  break;
                case MXK_BRANCH_TOGGLE :
                  mb->b_defer = 1 - mb->b_defer;
                  break;
              }
            }

#ifdef DEBUG
            fprintf(stderr, "branch: desired=%d, current=%d, track=%d\n", mb->b_defer, mb->b_state, mb->b_track);
#endif

            /* make real state desired ... */
            if(mb->b_state != mb->b_defer){
              if(mb->b_track == 0){ /* ... but only if all depressed keys are moved out */
#ifdef DEBUG
                fprintf(stderr, "branch: changing state %d->%d\n", mb->b_state, mb->b_defer);
#endif
                mb->b_state = mb->b_defer;

                if(mb->b_light[1 - mb->b_state]){
                  mxk_stop_light(ls, mb->b_light[1 - mb->b_state]);
                }
                if(mb->b_light[mb->b_state]){
                  mxk_start_light(ls, mb->b_light[mb->b_state]);
                }
#ifdef DEBUG
              } else {
                fprintf(stderr, "branch: deferring update: track=%d\n", mb->b_track);
#endif
              }
            }

            /* xmit event */
            if(((flags & MXK_BRANCH_HIPASS) && mb->b_state) ||
               ((flags & MXK_BRANCH_LOPASS) && (mb->b_state == 0))){
              mxk_send_client(ls, tl, mb->b_vector[mb->b_state], mb->b_count[mb->b_state], vp->type, vp->code, vp->value);
            }

            /* keep track of depressed keys */
            switch(vp->value){
              case 0 :
                mb->b_track--;
                if(mb->b_track < 0){
#ifdef DEBUG
                  fprintf(stderr, "branch: warning: track out of sync\n");
                  abort();
#endif
                  mb->b_track = 0;
                  mb->b_defer = mb->b_state;
                }
                break;
              case 1 :
                mb->b_track++;
                break;
            }

          break;

          default :
          /* TODO ? */
#ifdef DEBUG
            fprintf(stderr, "branch: ignoring event type %d\n", vp->type);
#endif
            flags = mb->b_map[0];
            if(((flags & MXK_BRANCH_HIPASS) && mb->b_state) ||
               ((flags & MXK_BRANCH_LOPASS) && (mb->b_state == 0))){
              mxk_send_client(ls, tl, mb->b_vector[mb->b_state], mb->b_count[mb->b_state], vp->type, vp->code, vp->value);
            }
          break;
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_branch(void *branch, int type, int id)
{
  struct mxk_branch *mb;
  struct mxk_client *mc;

  mb = branch;

  mc = mxk_insert_client(mb->b_vector[mb->b_state], mb->b_count[mb->b_state], type, id);
  if(mc == NULL){
    return -1;
  }

  mb->b_vector[mb->b_state] = mc;
  mb->b_count[mb->b_state] = mb->b_count[mb->b_state] + 1;

  return 0;
}

int mxk_startup_branch(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "branch", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_branch);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_branch(struct mxk_global *mg)
{
}
