/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* brasc module: generate 64 virtual keys from home row 6 */

struct mxk_brasc{
  unsigned int e_bits;
  int e_ref;

  struct mxk_client *e_vector;
  int e_count;
};

static void mxk_destroy_brasc(struct mxk_brasc *brasc);
static struct mxk_brasc *mxk_create_brasc();
static int mxk_run_brasc(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************************************************************/

struct mxk_brasc *mxk_create_brasc()
{
  struct mxk_brasc *brasc;

  brasc = malloc(sizeof(struct mxk_brasc));
  if(brasc == NULL){
    return NULL;
  }

  brasc->e_bits = 0;
  brasc->e_ref = 0;

  brasc->e_vector = NULL;
  brasc->e_count = 0;

  return brasc;
}

static void mxk_destroy_brasc(struct mxk_brasc *brasc)
{
  if(brasc == NULL){
    return;
  }

  mxk_shutdown_client(brasc->e_vector, brasc->e_count);
#ifdef DEBUG
  brasc->e_vector = NULL;
  brasc->e_count = 0;
#endif

  free(brasc);
}

/**********************************************************************/

#if 0
int mxk_toggle_brasc(struct list_loop *ls, struct task_loop *tx, int toggle, char *args)
{
  struct mxk_brasc *brasc;

  brasc = user_loop(ls, tx);

  /* do something to brasc state */

  return 0;
}
#endif

/**********************************************************************/

static int mxk_run_brasc(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_brasc *brasc;

  struct input_event *vp;
  struct link_loop *nk;

  int bit, key;

  mg = global_loop(ls);
  brasc = user_loop(ls, tl);

  if(brasc == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_brasc(brasc);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        if(vp->type == EV_KEY){
#ifdef DEBUG
          fprintf(stderr, "brasc: got key %d/%d\n", vp->code, vp->value);
#endif
          key = (-1);
          bit = (-1);
          switch(vp->code){
            case KEY_S :
              bit = 0x1;
              break;
            case KEY_D :
              bit = 0x2;
              break;
            case KEY_F :
              bit = 0x4;
              break;
            case KEY_J :
              bit = 0x8;
              break;
            case KEY_K :
              bit = 0x10;
              break;
            case KEY_L :
              bit = 0x20;
              break;
            case KEY_SPACE :
              bit = 0;
              break;
            case KEY_ENTER :
            case KEY_BACKSPACE :
              key = vp->code;
              break;
          }
          if(bit >= 0){
            switch(vp->value){
              case 0 : /* release */
                switch(brasc->e_ref){
                  case 0 :
                    brasc->e_bits = 0;
                    break;

                  case 1 : /* last key released */
#ifdef DEBUG
                    fprintf(stderr, "brasc: last release: bit mask 0x%02x\n", brasc->e_bits);
#endif

                    key = KEY_MAX + 1 + brasc->e_bits;
                    brasc->e_bits = 0;

                    mxk_send_client(ls, tl, brasc->e_vector, brasc->e_count, EV_KEY, key, 1);
                    mxk_send_client(ls, tl, brasc->e_vector, brasc->e_count, EV_KEY, key, 0);
                    mxk_send_client(ls, tl, brasc->e_vector, brasc->e_count, EV_SYN, 0, 0);

                    /* fall */
                  default :
                    brasc->e_ref--;
                    break;
                }
                break;
              case 1 :
                brasc->e_ref++;
                brasc->e_bits |= bit;
                break;
            }
          } else {
            if((key >= 0) && (vp->value == 0)){
              mxk_send_client(ls, tl, brasc->e_vector, brasc->e_count, EV_KEY, key, 1);
              mxk_send_client(ls, tl, brasc->e_vector, brasc->e_count, EV_KEY, key, 0);
              mxk_send_client(ls, tl, brasc->e_vector, brasc->e_count, EV_SYN, 0, 0);
            }
          }
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

#if 0
int mxk_toggle_brasc_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "brasc", MXK_MISC_OPTIONAL, 1, mxk_toggle_brasc);

  return 0;
}
#endif

int mxk_start_brasc_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_brasc *brasc;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type;
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "brasc");
  if(type <= 0){
    print_control(cl, cd, 1, "brasc: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "brasc: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "brasc: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
  switch(*end){
    case '\n' :
    case ' '  :
    break;
    print_control(cl, cd, 1, "brasc: first parameter needs to be an integer\n");
    return 0;
  }

  brasc = mxk_create_brasc();
  if(brasc == NULL){
    print_control(cl, cd, 1, "brasc: unable to allocate state\n");
    return 0;
  }

  tx = add_tag_loop(ls, -1, brasc, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_brasc, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "brasc: unable to create task\n");
    mxk_destroy_brasc(brasc);
    return 0;
  }

  text_loop(ls, tx, "brasc:%d", id);
  print_control(cl, cd, 1, "brasc: created task\n");

  return 0;
}

int mxk_list_brasc_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_brasc *brasc;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "brasc");
  if(type <= 0){
    print_control(cl, cd, 1, "brasc: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    brasc = user_loop(ls, tx);

    if(brasc){
      print_control(cl, cd, 0, "brasc:%d chord=0x%02x\n", tag_loop(ls, tx), brasc->e_bits);
      count++;
    } else {
      print_control(cl, cd, 1, "brasc: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}
/**********************************************************************/

static int mxk_insert_brasc(void *state, int type, int id)
{
  struct mxk_brasc *brasc;
  struct mxk_client *mc;

  brasc = state;

  mc = mxk_insert_client(brasc->e_vector, brasc->e_count, type, id);
  if(mc == NULL){
    return -1;
  }

  brasc->e_vector = mc;
  brasc->e_count++;

  return 0;
}

int mxk_startup_brasc(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "brasc", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_brasc);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_brasc(struct mxk_global *mg)
{
}
