/*
 * (c) Marc Welz 2008:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* bug (tribute to morse bugs) */

struct mxk_bug{
  unsigned int b_key;
  unsigned int b_datatime;

  unsigned int b_period;
  struct timeval b_press;
  unsigned int b_divisions;
  
  int b_bypass;
  int b_syn;

  struct mxk_client *b_vector;
  int b_count;
};

static void mxk_destroy_bug(struct mxk_bug *bug);
static struct mxk_bug *mxk_create_bug();
static int mxk_run_bug(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************************************************************/

struct mxk_bug *mxk_create_bug(unsigned int key, unsigned int period, unsigned int divisions)
{
  struct mxk_bug *bug;

  if(divisions <= 1){
    return NULL;
  }

  if(period < divisions){
    return NULL;
  }

  bug = malloc(sizeof(struct mxk_bug));
  if(bug == NULL){
    return NULL;
  }

  bug->b_vector = NULL;
  bug->b_count = 0;

  bug->b_divisions = divisions;
  bug->b_key = key;
  bug->b_period = period;

  bug->b_datatime = 0;
  bug->b_bypass = 1;
  bug->b_syn = 0;

  return bug;
}

static void mxk_destroy_bug(struct mxk_bug *bug)
{
  if(bug == NULL){
    return;
  }

  mxk_shutdown_client(bug->b_vector, bug->b_count);
#ifdef DEBUG
  bug->b_vector = NULL;
  bug->b_count = 0;
#endif

  free(bug);
}

/**********************************************************************/

int mxk_datatime_bug(struct list_loop *ls, struct task_loop *tx, int datatime, char *args)
{
  struct mxk_bug *bug;

  bug = user_loop(ls, tx);

  bug->b_datatime = datatime;

  return 0;
}

int mxk_bypass_bug(struct list_loop *ls, struct task_loop *tx, int bypass, char *args)
{
  struct mxk_bug *bug;

  bug = user_loop(ls, tx);

  bug->b_bypass = bypass;

  return 0;
}

/**********************************************************************/

int mxk_bypass_bug_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "bug", MXK_MISC_OPTIONAL, 1, mxk_bypass_bug);

  return 0;
}

int mxk_ignore_bug_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "bug", MXK_MISC_OPTIONAL, 0, mxk_bypass_bug);

  return 0;
}

int mxk_start_bug_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_bug *bug;
  struct mxk_global *mg;
  struct task_loop *tx;
  int i, cd, idx, type, values[4];
  char *end, *ptr;
  char buffer[MXK_NAME_SIZE];

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "bug");
  if(type <= 0){
    print_control(cl, cd, 1, "bug: type not found\n");
    return 0;
  }

  idx = (-1);
  if(args == NULL){
    print_control(cl, cd, 1, "bug: need four parameters\n");
    return 0;
  }

  idx = start_word(args, 0);

  for(i = 0; (idx >= 0) && (i < 4); i++){
    ptr = args + idx;
#ifdef DEBUG
    fprintf(stderr, "bug: parsing parameter[%d]: idx=%d, <%s>\n", i, idx, ptr);
#endif
    if(i == 1){

      if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
        print_control(cl, cd, 1, "single: missing or malformed key parameter\n");
        return 0;
      }

      values[i] = name_to_code(mxk_key_codes, buffer);
      if(values[i] < 0){
        print_control(cl, cd, 1, "bug: unknown key <%s>\n", buffer);
        return 0;
      }
    } else {
      values[i] = strtoul(ptr, &end, 10);
      if(values[i] < 0){
        print_control(cl, cd, 1, "bug: expected natural number\n");
        return 0;
      }
      switch(end[0]){
        case ' ' :
        case '\t' :
        case '\r' :
        case '\n' :
        case '\0' :
          break;
        default :
          print_control(cl, cd, 1, "bug: expected a natural number\n");
          return 0;
      }
    }

    idx = next_word(args, idx);
  }

  if(i < 4){
    print_control(cl, cd, 1, "bug: require four parameters\n");
    return 0;
  }

  if(values[2] < values[3]){
    print_control(cl, cd, 1, "bug: period can not be smaller than divisions\n");
    return 0;
  }

  if(values[3] <= 1){
    print_control(cl, cd, 1, "bug: need more than one interval\n");
    return 0;
  }

  bug = mxk_create_bug(values[1], values[2], values[3]);
  if(bug == NULL){
    print_control(cl, cd, 1, "bug: unable to initialise bug\n");
    return 0;
  }

  tx = add_tag_loop(ls, -1, bug, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_bug, NULL, type, values[0]);
  if(tx == NULL){
    print_control(cl, cd, 1, "bug: unable to create task\n");
    mxk_destroy_bug(bug);
    return 0;
  }

  text_loop(ls, tx, "bug:%d", values[0]);
  print_control(cl, cd, 1, "bug: created task\n");

  return 0;
}

int mxk_list_bug_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_bug *bug;
  char *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "bug");
  if(type <= 0){
    print_control(cl, cd, 1, "bug: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    bug = user_loop(ls, tx);

    if(bug){
      print_control(cl, cd, 0, "bug[%d]: key=", tag_loop(ls, tx));
      ptr = code_to_name(mxk_key_codes, bug->b_key);
      if(ptr){
        print_control(cl, cd, 0, "%s", ptr);
      } else {
        print_control(cl, cd, 0, "0x%x", bug->b_key);
      }
      print_control(cl, cd, 0, " period=%ums divisions=%u unmatched=%s\n", bug->b_period, bug->b_divisions, bug->b_bypass ? "bypass" : "ignore");
      count++;
    } else {
      print_control(cl, cd, 1, "bug: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

/**********************************************************************/

static int mxk_run_bug(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_bug *bug;

  struct input_event *vp;
  struct link_loop *nk;

  struct timeval *tp, *tn;
  unsigned int delta, value;

  mg = global_loop(ls);
  bug = user_loop(ls, tl);

  if(bug == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_bug(bug);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        if((vp->type == EV_KEY) && (vp->code == bug->b_key)){
          switch(vp->value){
            case 1 : /* press */
              if(bug->b_datatime){
                tp = &(vp->time);
              } else {
                tp = now_loop(ls);
              }
              bug->b_press.tv_sec  = tp->tv_sec;
              bug->b_press.tv_usec = tp->tv_usec;
            break;
            case 0 : /* release */
              if(bug->b_datatime){
                tn = &(vp->time);
              } else {
                tn = now_loop(ls);
              }
              tp = &(bug->b_press);

              if(tn->tv_usec < tp->tv_usec){
                if(tn->tv_sec <= tp->tv_sec){
                  delta = 0;
                } else {
                  delta = ((tn->tv_sec - (tp->tv_sec + 1)) * 1000) + (((1000000 + tn->tv_usec) - tp->tv_usec) / 1000);
                }
              } else {
                if(tn->tv_sec < tp->tv_sec){
                  delta = 0;
                } else {
                  delta = ((tn->tv_sec - tp->tv_sec) * 1000) + ((tn->tv_usec - tp->tv_usec) / 1000);
                }
              }

              if(delta >= bug->b_period){
                value = bug->b_divisions - 1;
              } else {
                value = (delta * bug->b_divisions) / bug->b_period;
              }

#ifdef DEBUG
              fprintf(stderr, "bug: (%us-%us) delay is %ums, value=%u\n", tn->tv_sec, tp->tv_sec, delta, value);
#endif
              value += (KEY_MAX + 1);
              
              mxk_send_client(ls, tl, bug->b_vector, bug->b_count, EV_KEY, value, 1);
              mxk_send_client(ls, tl, bug->b_vector, bug->b_count, EV_SYN, 0, 0);
              mxk_send_client(ls, tl, bug->b_vector, bug->b_count, EV_KEY, value, 0);
              mxk_send_client(ls, tl, bug->b_vector, bug->b_count, EV_SYN, 0, 0);
              bug->b_syn = 1;

            break;
            /* ignore autorepeats and others */
          }
        } else {
          if(bug->b_bypass){
            switch(vp->type){
              case EV_SYN :
                if(bug->b_syn == 0){
                  /* WARNING: assumes the SYN type has only one code */
                  mxk_send_client(ls, tl, bug->b_vector, bug->b_count, EV_SYN, 0, 0);
                  bug->b_syn = 1;
                }
                break;
              default :
                mxk_send_client(ls, tl, bug->b_vector, bug->b_count, vp->type, vp->code, vp->value);
                bug->b_syn = 0;
                break;
            }
          }
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

static int mxk_insert_bug(void *state, int type, int id)
{
  struct mxk_bug *bug;
  struct mxk_client *mc;

  bug = state;

  mc = mxk_insert_client(bug->b_vector, bug->b_count, type, id);
  if(mc == NULL){
    return -1;
  }

  bug->b_vector = mc;
  bug->b_count++;

  return 0;
}

int mxk_startup_bug(struct list_loop *ls)
{
  int type;

  /* register type with rest of system */

  type = mxk_enroll_find(ls, "bug", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_bug);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_bug(struct mxk_global *mg)
{
  /* could be useful if dynamic unloading ever gets implemented */
}
