/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sysexits.h>
#include <sched.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* external module */

#define SYSEXEC_BUFFER 128

struct mxk_external{
  char *s_argp;
  char **s_argv;
  int s_argc;

  int s_error;
  pid_t s_pid;
  int s_discard;

  char s_in[SYSEXEC_BUFFER];
  char s_out[SYSEXEC_BUFFER];

  int s_want;
  int s_have;

  struct mxk_client *s_vector;
  int s_count;
};

static void mxk_destroy_external(struct mxk_external *external);
static struct mxk_external *mxk_create_external(char *string);

static struct mxk_external *mxk_create_external(char *string)
{
  struct mxk_external *external;
  char *ptr, **vector;
  int count, idx, end;

  external = malloc(sizeof(struct mxk_external));
  if(external == NULL){
    return NULL;
  }

  external->s_argp = NULL;
  external->s_argv = NULL;

  external->s_argc = 0;

  external->s_pid = 0;
  external->s_error = 0;
  external->s_discard = 0;

  external->s_have = 0;
  external->s_want = 0;

  external->s_vector = NULL;
  external->s_count = 0;

  external->s_argp = strdup(string);
  if(external->s_argp == NULL){
    mxk_destroy_external(external);
    return NULL;
  }

  ptr = external->s_argp;
  count = 0;
  idx = start_word(ptr, 0);

  while(idx >= 0){
    vector = realloc(external->s_argv, (count + 3) * sizeof(char *));
    if(vector == NULL){
      mxk_destroy_external(external);
      return NULL;
    }
    external->s_argv = vector;

    vector[count] = ptr + idx;
    end = end_word(ptr, idx);
#ifdef DEBUG
    fprintf(stderr, "external: arg[%d]=<%s@%d..%d>\n", count, ptr + idx, idx, end);
#endif
    idx = next_word(ptr, idx); /* order critical: needs to be called before clobbering string with \0 */
    if(end > 0){
      ptr[end] = '\0';
    }
    count++;

  }

  if(count < 1){
#ifdef DEBUG
    fprintf(stderr, "external: unable to extract a command\n");
#endif
    mxk_destroy_external(external);
    return NULL;
  }

  vector[count] = NULL;
  external->s_argc = count;

  return external;
}

static int mxk_collect_external(struct mxk_external *external)
{
  int status;

  sched_yield();

  /* PROBLEM: may miss previous children, resulting in zombies */

  if(waitpid(external->s_pid, &status, WNOHANG) > 0){
    if(WIFEXITED(status)){
      if(WEXITSTATUS(status) > 1){
        external->s_error = EIO; /* made up error code */
      }
      external->s_pid = 0;
      return 1;
    }
    /* else: maybe signalled, currently ignored */
  }

  return 0;
}

static void mxk_finish_external(struct list_loop *ls, struct task_loop *tl, int error)
{
  struct mxk_external *external;
  int fd;

  fd = fd_loop(ls, tl);
  external = user_loop(ls, tl);

  set_fd_loop(ls, tl, -1);
  close(fd);
  set_mask_loop(ls, tl, LO_STOP_MASK | LO_WAIT_MASK, 0);

  if(error){
    external->s_error = errno;
  }

  mxk_collect_external(external);
}

static int mxk_parse_external(struct list_loop *ls, struct task_loop *tl)
{
  struct input_event event;
  struct mxk_external *external;
  int i;

  external = user_loop(ls, tl);

  for(i = 0; (i < external->s_have) && (external->s_in[i] != '\n'); i++);

  if(i >= external->s_have){
    if((external->s_have + 1) >= SYSEXEC_BUFFER){
      mxk_finish_external(ls, tl, E2BIG);
      return -1;
    }

    return 0;
  }

  external->s_in[i] = '\0';

#ifdef DEBUG
  fprintf(stderr, "external: attempting to parse <%s>\n", external->s_in);
#endif

  if(mxk_parse_evt(external->s_in, &event) < 0){
    mxk_finish_external(ls, tl, EINVAL);
    return -1;
  }

  mxk_send_client(ls, tl, external->s_vector, external->s_count, event.type, event.code, event.value);
  i++;

  if(i >= external->s_have){
    external->s_have = 0;
    return 0;
  }

  memmove(external->s_in, external->s_in + i, external->s_have - i);
  external->s_have -= i;

  return 1;
}

static int mxk_launch_external(struct mxk_external *external)
{
  int pair[2], fd, m;
  pid_t pid;

  if(socketpair(AF_UNIX, SOCK_STREAM, 0, pair) < 0){
    return -1;
  }

  pid = fork();
  switch (pid) {
  case -1:
    close(pair[0]);
    close(pair[1]);
    return -1;
  case 0:
    m = 0;
    fd = pair[1];
    close(pair[0]);

    if(fd != STDIN_FILENO){
      if(dup2(fd, STDIN_FILENO) != STDIN_FILENO){
        exit(EX_OSERR);
      }
    } else {
      m = 1;
    }

    if(fd != STDOUT_FILENO){
      if(dup2(fd, STDOUT_FILENO) != STDOUT_FILENO){
        exit(EX_OSERR);
      }
    } else {
      m = 1;
    }

    /* TODO: not quite satisfactory - ignoring stderr */
    close(STDERR_FILENO);

    if(m == 0){
      close(fd);
    }

    execvp(external->s_argv[0], external->s_argv);

    exit(EX_UNAVAILABLE);
    return -1;
  default :
    fd = pair[0];
    close(pair[1]);

    fcntl(fd, F_SETFD, FD_CLOEXEC);
    misc_unblock_loop(fd);

    external->s_pid = pid;

    if(mxk_collect_external(external)){
      close(fd);
      return -1;
    }

    return fd;
  }

  /* should not happen - is an exit */
  return -1;
}

static void mxk_destroy_external(struct mxk_external *external)
{
  if(external == NULL){
    return;
  }

  mxk_shutdown_client(external->s_vector, external->s_count);
#ifdef DEBUG
  external->s_vector = NULL;
  external->s_count = 0;
#endif

  if(external->s_argp){
    free(external->s_argp);
    external->s_argp = NULL;
  }

  if(external->s_argv){
    free(external->s_argv);
    external->s_argv = NULL;
  }

  free(external);
}

/**********************************************************************/

int mxk_toggle_external(struct list_loop *ls, struct task_loop *tx, int toggle, char *args)
{
  struct mxk_external *external;

  external = user_loop(ls, tx);

  /* do something to external state */

  return 0;
}

/**********************************************************************/

static int mxk_run_external(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_external *external;
  struct input_event *vp;
  struct link_loop *nk;
  int rr, wr, fd, cr;

  mg = global_loop(ls);
  fd = fd_loop(ls, tl);
  external = user_loop(ls, tl);

  if(external == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_finish_external(ls, tl, 0); /* attempt to collect code */
    mxk_destroy_external(external);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        cr = mxk_string_evt(external->s_out + external->s_want, SYSEXEC_BUFFER - (1 + external->s_want), vp->type, vp->code, vp->value);
#ifdef DEBUG
        fprintf(stderr, "external: converted event to %d byte string\n", cr);
#endif
        if(cr > 0){
          external->s_want += cr;
          external->s_out[external->s_want] = '\n';
          external->s_want++;
        } else {
          /* WARNING: a problem: we don't have infinite buffers, so just drop if full */
          external->s_discard++;
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  if(external->s_want > 0){
    wr = write(fd, external->s_out, external->s_want);
#ifdef DEBUG
    fprintf(stderr, "external: backlog: clearing %d/%d\n", wr, external->s_want);
#endif
    if(wr < 0){
      switch(errno){
        case EAGAIN :
        case EINTR :
          break;
        default :
          mxk_finish_external(ls, tl, errno);
          return 0;
      }
    } else {
      if(wr >= external->s_want){
        external->s_want = 0;
        set_mask_loop(ls, tl, LO_READ_MASK | LO_WAIT_MASK | LO_STOP_MASK, 0);
      } else {
        if(wr > 0){
          memmove(external->s_out, external->s_out + wr, external->s_want - wr);
          external->s_want -= wr;
        }
        set_mask_loop(ls, tl, LO_READ_MASK | LO_WRITE_MASK | LO_STOP_MASK, 0);
        return 0;
      }
    }
  }

  if(mask & LO_READ_MASK){
    rr = read(fd, external->s_in + external->s_have, SYSEXEC_BUFFER - (1 + external->s_have));
#ifdef DEBUG
    fprintf(stderr, "external: read %d+%d\n", rr, external->s_have);
#endif
    switch(rr){
      case -1 :
        switch(errno){
          case EAGAIN :
          case EINTR :
            break;
          default :
            mxk_finish_external(ls, tl, errno);
            return 0;
        }
        break;
      case 0 :
        mxk_finish_external(ls, tl, 0);
        return 0;
      default :
        external->s_have += rr;
        external->s_in[external->s_have] = '\0';
      break;
    }

    while(mxk_parse_external(ls, tl) > 0);
  }

  return 0;
}

/*********************************************************************/

int mxk_toggle_external_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "external", MXK_MISC_OPTIONAL, 1, mxk_toggle_external);

  return 0;
}

int mxk_start_external_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_external *external;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type, fd;
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "external");
  if(type <= 0){
    print_control(cl, cd, 1, "external: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "external: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "external: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "external: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    print_control(cl, cd, 1, "external: nothing to run\n");
    return 0;
  }

  external = mxk_create_external(args + idx);
  if(external == NULL){
    print_control(cl, cd, 1, "external: unable to allocate state\n");
    return 0;
  }

  fd = mxk_launch_external(external);
  if(fd < 0){
    print_control(cl, cd, 1, "external: unable to start subprocess\n");
    return 0;
  }

  tx = add_tag_loop(ls, fd, external, LO_STOP_MASK | LO_WAIT_MASK | LO_READ_MASK, mxk_run_external, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "external: unable to create task\n");
    mxk_destroy_external(external);
    return 0;
  }

  text_loop(ls, tx, "external:%d (%d) %s", id, external->s_pid, external->s_argv[0]);
  print_control(cl, cd, 1, "external: created task\n");

  return 0;
}

int mxk_list_external_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_external *external;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "external");
  if(type <= 0){
    print_control(cl, cd, 1, "external: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    external = user_loop(ls, tx);

    if(external){
      print_control(cl, cd, 0, "external[%d]: ", tag_loop(ls, tx));
      for(i = 0; i < external->s_argc; i++){
        print_control(cl, cd, 0, " %s", external->s_argv[i]);
      }
      print_control(cl, cd, 0, "\n");
      count++;

      if(external->s_pid){
        print_control(cl, cd, 0, "state: up (pid=%d)\n", external->s_pid);
      } else if(external->s_error){
        print_control(cl, cd, 0, "state: failed (%s)\n", strerror(errno));
      } else {
        print_control(cl, cd, 0, "state: finished\n");
      }
      print_control(cl, cd, 0, "discards: %d\n", external->s_discard);

    } else {
      print_control(cl, cd, 1, "external: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}
/**********************************************************************/

static int mxk_insert_external(void *state, int type, int id)
{
  struct mxk_external *external;
  struct mxk_client *mc;

  external = state;

  mc = mxk_insert_client(external->s_vector, external->s_count, type, id);
  if(mc == NULL){
    return -1;
  }

  external->s_vector = mc;
  external->s_count++;

  return 0;
}

int mxk_startup_external(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "external", MXK_EVT_TARGET | MXK_EVT_SOURCE, -1, mxk_insert_external);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_external(struct mxk_global *mg)
{
}
