/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* genb module: chords up to 8 bits */

#define GENB_S 8

struct mxk_genb{
  unsigned int e_bits;
  unsigned int e_press;
  unsigned int e_release;
  int e_ref;
  int e_keys[GENB_S];
  int e_size;

  struct mxk_client *e_vector;
  int e_count;
};

static void mxk_destroy_genb(struct mxk_genb *genb);
static struct mxk_genb *mxk_create_genb();
static int mxk_run_genb(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************************************************************/

struct mxk_genb *mxk_create_genb()
{
  struct mxk_genb *genb;
  int i;

  genb = malloc(sizeof(struct mxk_genb));
  if(genb == NULL){
    return NULL;
  }

  genb->e_bits = 0;
  genb->e_ref = 0;
  genb->e_press = 0;
  genb->e_release = 0;

  for(i = 0; i < GENB_S; i++){
    genb->e_keys[i] = (-1);
  }
  genb->e_size = 0;

  genb->e_vector = NULL;
  genb->e_count = 0;

  return genb;
}

static int mxk_add_genb(struct mxk_genb *genb, int key)
{
  if(genb->e_size >= GENB_S){
    return -1;
  }

  genb->e_keys[genb->e_size] = key;

  genb->e_size++;

  return 0;
}

static void mxk_destroy_genb(struct mxk_genb *genb)
{
  if(genb == NULL){
    return;
  }

  mxk_shutdown_client(genb->e_vector, genb->e_count);
#ifdef DEBUG
  genb->e_vector = NULL;
  genb->e_count = 0;
  genb->e_size = 0;
#endif

  free(genb);
}

/**********************************************************************/

static void mxk_key_genb(struct list_loop *ls, struct task_loop *tl, unsigned int bits, int value)
{
  struct mxk_genb *genb;
  genb = user_loop(ls, tl);

#ifdef DEBUG
  fprintf(stderr, "genb: %s 0x%04x\n", value ? "press" : "release", bits);
#endif

  mxk_send_client(ls, tl, genb->e_vector, genb->e_count, EV_KEY, KEY_MAX + bits, value);
  mxk_send_client(ls, tl, genb->e_vector, genb->e_count, EV_SYN, 0, 0);
}

static int mxk_run_genb(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_genb *genb;

  struct input_event *vp;
  struct link_loop *nk;

  int i;
  unsigned int bit;

  mg = global_loop(ls);
  genb = user_loop(ls, tl);

  if(genb == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_genb(genb);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        if(vp->type == EV_KEY){
#ifdef DEBUG
          fprintf(stderr, "genb: got key %d/%d\n", vp->code, vp->value);
#endif

          bit = 0;
          for(i = 0; i < genb->e_size; i++){
            if(vp->code == genb->e_keys[i]){
              bit = (1 << i);
              break;
            }
          }

          if(bit > 0){
            switch(vp->value){
              case 0 : /* release */
                if(genb->e_press > 0){

                  if(genb->e_release > 0){
#ifdef DEBUG
                    fprintf(stderr, "genb: release: weirdness - need to clear before next press\n");
#endif
                    mxk_key_genb(ls, tl, genb->e_release, 0);
                    genb->e_release = 0;
                  }

                  mxk_key_genb(ls, tl, genb->e_press, 1);
                  genb->e_release = genb->e_press;
                  genb->e_press = 0;
                }

                genb->e_bits &= ~bit;

                if(genb->e_bits == 0){
                  if(genb->e_release > 0){
#ifdef DEBUG
                    fprintf(stderr, "genb: release: last input released\n");
#endif
                    mxk_key_genb(ls, tl, genb->e_release, 0);
                    genb->e_release = 0;
                  }
                }
                break;
              case 1 :
                if(genb->e_release > 0){
                  mxk_key_genb(ls, tl, genb->e_release, 0);
                  genb->e_release = 0;
                }

                genb->e_ref++;
                genb->e_bits |= bit;
                /* propose this as next output */
                genb->e_press = genb->e_bits;
                break;
            }
          }
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

int mxk_start_genb_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_genb *genb;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type, key;
  char buffer[MXK_NAME_SIZE];
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "genb");
  if(type <= 0){
    print_control(cl, cd, 1, "genb: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "genb: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "genb: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
  switch(*end){
    case '\n' :
    case ' '  :
    break;
    print_control(cl, cd, 1, "genb: first parameter needs to be an integer\n");
    return 0;
  }

  genb = mxk_create_genb();
  if(genb == NULL){
    print_control(cl, cd, 1, "genb: unable to allocate state\n");
    return 0;
  }

  while((idx = next_word(args, idx)) > 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      print_control(cl, cd, 1, "genb: unable to load key name\n");
      mxk_destroy_genb(genb);
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key <= 0){
      print_control(cl, cd, 1, "genb: unknown key name <%s>\n", buffer);
      mxk_destroy_genb(genb);
      return -1;
    }

    if(mxk_add_genb(genb, key)){
      print_control(cl, cd, 1, "genb: too many keys to add\n");
      mxk_destroy_genb(genb);
      return -1;
    }
  }

  if(genb->e_size <= 0){
    print_control(cl, cd, 1, "genb: need at least one key\n");
    mxk_destroy_genb(genb);
    return -1;
  }

  tx = add_tag_loop(ls, -1, genb, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_genb, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "genb: unable to create task\n");
    mxk_destroy_genb(genb);
    return 0;
  }

  text_loop(ls, tx, "genb:%d keys=%d", id, genb->e_size);
  print_control(cl, cd, 1, "genb: created task\n");

  return 0;
}

int mxk_list_genb_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_genb *genb;
  char *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "genb");
  if(type <= 0){
    print_control(cl, cd, 1, "genb: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    genb = user_loop(ls, tx);

    if(genb){
      print_control(cl, cd, 0, "genb[%d]: chord=0x%02x size=%d chord=<", tag_loop(ls, tx), genb->e_bits, genb->e_size);
      for(i = 0; i < genb->e_size; i++){
        ptr = code_to_name(mxk_key_codes, genb->e_keys[i]);
        if(ptr){
          print_control(cl, cd, 0, " %s", ptr);
        } else {
          print_control(cl, cd, 0, " %d", genb->e_keys[i]);
        }
      }
      print_control(cl, cd, 0, ">\n");
      count++;
    } else {
      print_control(cl, cd, 1, "genb: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}
/**********************************************************************/

static int mxk_insert_genb(void *state, int type, int id)
{
  struct mxk_genb *genb;
  struct mxk_client *mc;

  genb = state;

  mc = mxk_insert_client(genb->e_vector, genb->e_count, type, id);
  if(mc == NULL){
    return -1;
  }

  genb->e_vector = mc;
  genb->e_count++;

  return 0;
}

int mxk_startup_genb(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "genb", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_genb);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_genb(struct mxk_global *mg)
{
}
