/*
 * (c) Marc Welz 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* Feeble guess module: Test case for input prediction infrastructure.
 * Guess which keypress is going to follow the current one */

struct mxk_guess_item{
  int i_see;
  int i_reply;
};

struct mxk_guess{
  int g_expand;   /* the "tab" key */
  struct mxk_guess_item *g_propose; /* the result of the tab */

  int g_last;     /* last key pressed */
  int g_now;

  int g_size;
  struct mxk_guess_item **g_table;

  struct mxk_client *g_vector;
  int g_count;
};

static void mxk_destroy_guess(struct mxk_guess *guess);
static struct mxk_guess *mxk_create_guess(int size, int expand);
static int mxk_run_guess(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************************************************************/

struct mxk_guess *mxk_create_guess(int size, int expand)
{
  struct mxk_guess *guess;
  struct mxk_guess_item *gi;
  int i;

  guess = malloc(sizeof(struct mxk_guess));
  if(guess == NULL){
    return NULL;
  }

  guess->g_expand = expand;
  guess->g_propose = NULL;
  guess->g_last = -1;
  guess->g_now = -1;

  guess->g_vector = NULL;
  guess->g_count = 0;

  guess->g_table = NULL;
  guess->g_size = 0;

  guess->g_table = malloc(sizeof(struct mxk_guess_item *) * size);
  if(guess->g_table == NULL){
    free(guess);
    return NULL;
  }

  for(i = 0; i < size; i++){
    gi = malloc(sizeof(struct mxk_guess_item));
    guess->g_table[i] = gi;
    if(gi == NULL){
      mxk_destroy_guess(guess);
      return NULL;
    } else {
      gi->i_see = (-1);
      gi->i_reply = (-1);
    }
  }

  guess->g_size = size;

  return guess;
}

static void mxk_destroy_guess(struct mxk_guess *guess)
{
  int i;

  if(guess == NULL){
    return;
  }

  if(guess->g_table){
    for(i = 0; i < guess->g_size; i++){
      if(guess->g_table[i]){
        free(guess->g_table[i]);
        guess->g_table[i] = NULL;
      }
    }
    free(guess->g_table);
    guess->g_table = NULL;
  }

  guess->g_size = 0;

  mxk_shutdown_client(guess->g_vector, guess->g_count);
#ifdef DEBUG
  guess->g_vector = NULL;
  guess->g_count = 0;
#endif

  free(guess);
}

#if 0
/**********************************************************************/

int mxk_toggle_guess(struct list_loop *ls, struct task_loop *tx, int toggle, char *args)
{
  struct mxk_guess *guess;

  guess = user_loop(ls, tx);

  /* do something to guess state */

  return 0;
}

/**********************************************************************/

int mxk_toggle_guess_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "guess", MXK_MISC_OPTIONAL, 1, mxk_toggle_guess);

  return 0;
}
#endif

int mxk_start_guess_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  char buffer[MXK_NAME_SIZE];
  struct control_loop *cl;
  struct mxk_guess *guess;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type, expand;
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "guess");
  if(type <= 0){
    print_control(cl, cd, 1, "guess: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "guess: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "guess: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
  switch(end[0]){
    case '\0'  :
    case '\r'  :
    case '\n'  :
    case ' '   :
      break;
    default :
      print_control(cl, cd, 1, "guess: first parameter has to be an integer\n");
      return 0;
  }
  idx = next_word(args, idx);

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, cd, 1, "guess: parameter seems too long\n");
    return 0;
  }

  expand = name_to_code(mxk_key_codes, buffer);
  if(expand < 0){
    print_control(cl, cd, 1, "guess: unknown key <%s>\n", buffer);
    return 0;
  }

  guess = mxk_create_guess(16, expand);
  if(guess == NULL){
    print_control(cl, cd, 1, "guess: unable to allocate state\n");
    return 0;
  }

  tx = add_tag_loop(ls, -1, guess, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_guess, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "guess: unable to create task\n");
    mxk_destroy_guess(guess);
    return 0;
  }

  text_loop(ls, tx, "guess:%d", id);
  print_control(cl, cd, 1, "guess: created task\n");

  return 0;
}

int mxk_list_guess_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, i, idx, id, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_guess *guess;
  struct mxk_guess_item *gi;
  char *trigger, *value;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "guess");
  if(type <= 0){
    print_control(cl, cd, 1, "guess: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    guess = user_loop(ls, tx);

    if(guess){
      trigger = code_to_name(mxk_key_codes, guess->g_expand);
      print_control(cl, cd, 0, "guess[%d]: expand: %s\n", tag_loop(ls, tx), trigger ? trigger : "<unknown>");
      for(i = 0; i < guess->g_size; i++){
        gi = guess->g_table[i];
        if(gi->i_see > 0){
          trigger = code_to_name(mxk_key_codes, gi->i_see);
        } else {
          trigger = NULL;
        }
        if(gi->i_reply){
          value = code_to_name(mxk_key_codes, gi->i_reply);
        } else {
          value = NULL;
        }
        if(trigger && value){
          print_control(cl, cd, 0, "[%d]: %s -> %s\n", i, trigger, value);
        }
      }
      count++;
    } else {
      print_control(cl, cd, 1, "guess: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

/**********************************************************************/

void reorder_guess(struct mxk_guess *guess, struct mxk_guess_item *item, int pos)
{
  int old, i;

#ifdef DEBUG
  if((pos < 0) || (pos >= guess->g_size)){
    fprintf(stderr, "logic failure: position %d out of range\n", pos);
    abort();
  }
#endif

  for(old = 0; (item != guess->g_table[old]); old++){
#ifdef DEBUG
    if(old >= guess->g_size){
      fprintf(stderr, "logic failure: item %p not in table at %p\n", item, guess->g_table);
      abort();
    }
#endif
  }

  if(old > pos){
    for(i = old; i > pos; i--){
      guess->g_table[i] = guess->g_table[i - 1];
    }
    guess->g_table[pos] = item;
  } else if(old < pos){
    for(i = old; i < pos; i++){
      guess->g_table[i] = guess->g_table[i + 1];
    }
    guess->g_table[pos] = item;
  }

}

static int mxk_run_guess(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_guess *guess;
  struct mxk_guess_item *test;

  struct input_event *vp;
  struct link_loop *nk;

  int i, update;
  char *ptr;

  mg = global_loop(ls);
  guess = user_loop(ls, tl);

  if(guess == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_guess(guess);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
#ifdef DEBUG
        fprintf(stderr, "guess: got (%d,%d,%d)\n", vp->type, vp->code, vp->value);
#endif
        if((vp->type == EV_KEY) && (vp->value == 1)){

          update = 0;

          guess->g_last = guess->g_now;

          if(vp->code == guess->g_expand){
            if(guess->g_propose){
#ifdef DEBUG
              fprintf(stderr, "guess: expanding: %d\n", guess->g_propose->i_reply);
#endif
              mxk_send_client(ls, tl, guess->g_vector, guess->g_count, EV_KEY, guess->g_propose->i_reply, 1);
              mxk_send_client(ls, tl, guess->g_vector, guess->g_count, EV_KEY, guess->g_propose->i_reply, 0);
              mxk_send_client(ls, tl, guess->g_vector, guess->g_count, EV_SYN, 0, 0);
              reorder_guess(guess, guess->g_propose, 0);
              guess->g_now = guess->g_propose->i_reply;

            }

          } else {

            guess->g_now = vp->code;

            test = NULL;
            if(guess->g_last >= 0){
              for(i = 0; i < guess->g_size; i++){
                if(guess->g_table[i]->i_see == guess->g_last){
                  test = guess->g_table[i];
#ifdef DEBUG
                  fprintf(stderr, "guess: found prior match at %d (prev=%d, now=%d)\n", i, test->i_reply, guess->g_now);
#endif
                  if(test->i_reply == guess->g_now){ /* perc up */
                    reorder_guess(guess, test, (i > 2) ? (i - 2) : 0);
                  } else { /* just change */
                    test->i_reply = guess->g_now;
                  }
                }
              }
            }
            if(test == NULL){ /* recycle last if nothing found */
#ifdef DEBUG
              fprintf(stderr, "guess: new sequence %d -> %d\n", guess->g_last, guess->g_now);
#endif
              test = guess->g_table[guess->g_size - 1];
              test->i_reply = guess->g_now;
              test->i_see = guess->g_last;
              reorder_guess(guess, test, guess->g_size / 2);
            }
            if(test == guess->g_propose){
              update = 1;
            }
          }

          /* now propose something */
          for(i = 0; i < guess->g_size; i++){
            if(guess->g_table[i]->i_see == guess->g_now){
#ifdef DEBUG
              fprintf(stderr, "guess: guessing from %d\n", i);
#endif
              test = guess->g_table[i];
              if(guess->g_table[i] != guess->g_propose){
                guess->g_propose = guess->g_table[i];
                update = 1;
              }
              break;
            }
          }

          /* need to update */
          if(update){
            ptr = code_to_name(mxk_key_codes, guess->g_propose->i_reply);
            if(ptr){
              mxk_write_watch(ls, tl, 0, "%s", ptr);
            } else {
              mxk_write_watch(ls, tl, 0, "%d", guess->g_propose->i_reply);
            }
          }

        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

static int mxk_insert_guess(void *state, int type, int id)
{
  struct mxk_guess *guess;
  struct mxk_client *mc;

  guess = state;

  mc = mxk_insert_client(guess->g_vector, guess->g_count, type, id);
  if(mc == NULL){
    return -1;
  }

  guess->g_vector = mc;
  guess->g_count++;

  return 0;
}

int mxk_startup_guess(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "guess", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_guess);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_guess(struct mxk_global *mg)
{
}
