/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <linux/input.h>

#include <loop.h>
#include <code-name.h>
#include <word-work.h>

#include "mxk.h"

static void mxk_free_pattern(struct mxk_pattern *mp);
static int mxk_run_light(struct list_loop *ls, struct task_loop *tl, int mask);

static struct code_name mxk_code_table[] = {
  { "on",     MXK_FLASH_ON },
  { "off",    MXK_FLASH_OFF },
  { "wait",   MXK_FLASH_WAIT },
  { "repeat", MXK_FLASH_REPEAT },
  { NULL, -1 }
};

int mxk_parse_flash(struct mxk_flash *mf, char *buffer)
{
  char *code, *value;

  mf->f_id = 0;

#ifdef DEBUG
  fprintf(stderr, "flash: attempting to parse <%s>\n", buffer);
#endif

  code = strchr(buffer, '/');
  if(code == NULL){
    code = buffer;
  } else {
    code[0] = '\0';
    /* TODO: could use strtoul to detect errors */
    mf->f_id = atoi(buffer);
    code++;
  }

  value = strchr(code, ':');
  if(value){
    value[0] = '\0';
    value++;
  }

  mf->f_code = name_to_code(mxk_code_table, code);
  switch(mf->f_code){
    case MXK_FLASH_ON :
    case MXK_FLASH_OFF :
      if(value == NULL){
        return -1;
      }
      mf->f_value = name_to_code(mxk_led_codes, value);
      if(mf->f_value < 0){
        return -1;
      }
      break;
    case MXK_FLASH_WAIT :
      /* TODO: could use strtoul to detect errors */
      if(value){
        mf->f_value = atoi(value);
      } else {
        mf->f_value = 1;
      }
      break;
    case MXK_FLASH_REPEAT :
      if(value){
        mf->f_value = atoi(value);
      } else {
        mf->f_value = 0;
      }
      break;
    default :
      return -1;
  }

#ifdef DEBUG
fprintf(stderr, "flash: code=%d, value=%d\n", mf->f_code, mf->f_value);
#endif

  return 0;
}

#define FLASH_BUFFER 64

struct mxk_pattern* mxk_parse_pattern(struct list_loop *ls, struct task_loop *tl, char *args, int id)
{
  struct mxk_pattern *mp;
  struct mxk_flash *vc, *mf;
  struct control_loop *cl;
  int fd, idx;
  char buffer[FLASH_BUFFER];

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  if(args == NULL){
    print_control(cl, fd, 1, "need parameters\n");
    return NULL;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, fd, 1, "need nonempty parameters\n");
    return NULL;
  }

  mp = malloc(sizeof(struct mxk_pattern));
  if(mp == NULL){
    print_control(cl, fd, 1, "unable to allocate pattern\n");
    return NULL;
  }

  mp->p_name = NULL;
  mp->p_size = 0;
  mp->p_array = NULL;
  mp->p_used = 0;
  mp->p_id = id;

  mp->p_name = copy_word(args, idx);
  if(mp->p_name == NULL){
    print_control(cl, fd, 1, "unable to extract pattern name\n");
    mxk_free_pattern(mp);
    return NULL;
  }

  while(0 < (idx = next_word(args, idx))){
    vc = realloc(mp->p_array, (1 + mp->p_size) * sizeof(struct mxk_flash));
    if(vc == NULL){
      print_control(cl, fd, 1, "unable to allocate entry for element %d\n", mp->p_size);
      mxk_free_pattern(mp);
      return NULL;
    }

    mp->p_array = vc;
    mf = &(mp->p_array[mp->p_size]);

#ifdef DEBUG
fprintf(stderr, "load: processing at <%s>, into buffer/%d\n", args + idx, FLASH_BUFFER);
#endif

    if(load_word(buffer, FLASH_BUFFER, args, idx) < 0){
      print_control(cl, fd, 1, "element %d too long\n", mp->p_size);
      mxk_free_pattern(mp);
      return NULL;
    }

    if(mxk_parse_flash(mf, buffer)){
      print_control(cl, fd, 1, "unable to parse element %d\n", mp->p_size);
      mxk_free_pattern(mp);
      return NULL;
    }

    mp->p_size = mp->p_size + 1;
  }

  return mp;
}

static void mxk_free_pattern(struct mxk_pattern *mp)
{
  if(mp){
    if(mp->p_name){
      free(mp->p_name);
      mp->p_name = NULL;
    }

    if(mp->p_array){
      free(mp->p_array);
      mp->p_array = NULL;
    }

    free(mp);
  }
}

int mxk_load_pattern_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct mxk_global *mg;
  struct mxk_pattern *mp, **xp;
  struct control_loop *cl;
  int fd;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);
  mg = global_loop(ls);

  mp = mxk_parse_pattern(ls, tl, args, mg->g_pid++);
  if(mp == NULL){
    return 0;
  }

  xp = realloc(mg->g_patterns, sizeof(struct mxk_pattern *) * (mg->g_psize + 1));
  if(xp == NULL){
    print_control(cl, fd, 1, "unable to update array\n");
    mxk_free_pattern(mp);
    return 0;
  }

  /* TODO: could insert sorted, to do lsearch later on */

  mg->g_patterns = xp;
  mg->g_patterns[mg->g_psize] = mp;
  mg->g_psize++;

  return 0;
}

int mxk_list_pattern_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct mxk_global *mg;
  struct mxk_pattern *mp;
  struct control_loop *cl;
  int fd, i;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);
  mg = global_loop(ls);

  for(i = 0; i < mg->g_psize; i++){
    mp = mg->g_patterns[i];
    print_control(cl, fd, 0, "[%d]: %s (%d %s)\n", mp->p_id, mp->p_name, mp->p_size, (mp->p_size == 1) ? "event" : "events");
  }

  print_control(cl, fd, 1, "%d %s\n", mg->g_psize, (mg->g_psize == 1) ? "pattern" : "patterns");

  return 0;
}

struct task_loop *mxk_start_light(struct list_loop *ls, char *pattern)
{
  struct mxk_global *mg;
  struct mxk_pattern *mp;
  struct mxk_light *ml;
  struct task_loop *tx;
  int i, lt, st;

  mg = global_loop(ls);
  lt = lookup_task_type_loop(ls, "light");
  st = lookup_task_type_loop(ls, "source");

  for(i = 0; (i < mg->g_psize) && strcmp(mg->g_patterns[i]->p_name, pattern); i++);

  if(i >= mg->g_psize){
    return NULL;
  }

  mp = mg->g_patterns[i];

  ml = malloc(sizeof(struct mxk_light));
  if(ml == NULL){
    return NULL;
  }

  ml->l_index = 0;
  ml->l_count = 0;
  ml->l_current = mp;
  ml->l_source_type = st;

  tx = add_tag_loop(ls, -1, ml, LO_STOP_MASK | LO_WORK_MASK, mxk_run_light, NULL, lt, mp->p_id);
  if(tx == NULL){
    free(ml);
    return NULL;
  }

  mp->p_used++;
  text_loop(ls, tx, "light: pattern %s", pattern);

  return tx;
}

int mxk_start_light_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int cd, idx;
  char *pattern;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "need a valid pattern as parameter\n");
    return 0;
  }

  pattern = args + idx;
  if(mxk_start_light(ls, pattern) == NULL){
    print_control(cl, cd, 1, "unable to allocate light\n");
    return 0;
  }

  return 0;
}

int mxk_stop_light(struct list_loop *ls, char *pattern)
{
  int type, result, i, id;
  struct task_loop *tx;
  struct mxk_global *mg;
  struct mxk_pattern *mp;

  mg = global_loop(ls);
#ifdef DEBUG
  if(mg == NULL){
    fprintf(stderr, "stop light: major logic problem: no global state\n");
    abort();
  }
#endif

  for(i = 0; (i < mg->g_psize) && strcmp(mg->g_patterns[i]->p_name, pattern); i++);

  if(i >= mg->g_psize){
    return -1;
  }

  mp = mg->g_patterns[i];
  if(mp == NULL){
    return -1;
  }

  type = lookup_task_type_loop(ls, "light");
  if(type <= 0){
    return -1;
  }

  id = mp->p_id;
  result = 0;

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    stop_loop(ls, tx);
    result++;
  }

  return result;
}

int mxk_stop_light_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int cd, idx, count;
  char *pattern;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "need a valid pattern as parameter\n");
    return 0;
  }

  pattern = args + idx;
  count = mxk_stop_light(ls, pattern);

  if(count < 0){
    print_control(cl, cd, 1, "unable to stop light %s\n", pattern);
  } else {
    print_control(cl, cd, 1, "stopped %d %s\n", count, (count == 1) ? "pattern" : "patterns");
  }

  return 0;
}

static int mxk_run_light(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_light *ml;
  struct mxk_flash *mf, *xf;
  struct mxk_pattern *mp;
  struct mxk_global *mg;
  struct task_loop *tx;
  struct timeval tv;
  int value, finish, run, flush;

  mg = global_loop(ls);
  ml = user_loop(ls, tl);

  if(ml == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  mp = ml->l_current;
  if(mp == NULL){
#ifdef DEBUG
    fprintf(stderr, "light: odd, no pattern\n");
#endif
    stop_loop(ls, tl);
    return 0;
  }

  finish = (mask & LO_STOP_MASK) ? 1 : 0;

  if(mask & (LO_STOP_MASK | LO_WORK_MASK | LO_TIME_MASK)){ /* include stop to leave lights consistent  */
    run = finish ? (ml->l_index) : 1; /* skip if at start and already finishing */
    while(run){
      mf = &(mp->p_array[ml->l_index]);
#ifdef DEBUG
      fprintf(stderr, "light: iteration=%d, position=%d/%d\n", ml->l_count, ml->l_index, mp->p_size);
#endif
      ml->l_index++;
      value = 0;
      switch(mf->f_code){
        case MXK_FLASH_ON :
          value = 1;
          /* fall */
        case MXK_FLASH_OFF :
#ifdef DEBUG
          fprintf(stderr, "light: attempting to flash light 0x%02x (searching for type=%d)\n", mf->f_value, ml->l_source_type);
#endif

          /* look ahead to see if we can defer flush */
          flush = 1;
          if(ml->l_index < mp->p_size){
            xf = &(mp->p_array[ml->l_index]);
            switch(xf->f_code){
              case MXK_FLASH_ON :
              case MXK_FLASH_OFF :
                if((xf->f_id == 0) || (xf->f_id == mf->f_id)){
                  flush = 0;
                }
                break;
            }
          }

          for(tx = find_task_tag_loop(ls, NULL, NULL, ml->l_source_type, mf->f_id); tx; tx = find_task_tag_loop(ls, NULL, tx, ml->l_source_type, mf->f_id)){
#ifdef DEBUG
            fprintf(stderr, "light: considering task=%p\n", tx);
#endif
            /* MAYBE: write directly to fd (?) */
            mxk_send_switch(ls, tl, tx, EV_LED, mf->f_value, value);
            if(flush){
              mxk_send_switch(ls, tl, tx, EV_SYN, 0, 0);
            }
          }
          break;
        case MXK_FLASH_WAIT :
          if(finish == 0){
            tv.tv_sec = mf->f_value / 1000;
            tv.tv_usec = (mf->f_value % 1000) * 1000;
            set_time_loop(ls, tl, &tv);
            set_mask_loop(ls, tl, LO_STOP_MASK | LO_TIME_MASK, 0);
          }
          run = 0;
          break;
        case MXK_FLASH_REPEAT :
          set_mask_loop(ls, tl, LO_STOP_MASK | LO_WORK_MASK, 0);
          ml->l_index = 0;
          ml->l_count++;
          if(mf->f_value && (ml->l_count >= mf->f_value)){
            stop_loop(ls, tl);
          }
          run = 0;
          break;
#ifdef DEBUG
        default :
          fprintf(stderr, "light: unknown code=%d, value=%d\n", mf->f_code, mf->f_value);
          run = 0;
          break;
#endif
      }

      if(ml->l_index >= mp->p_size){
        stop_loop(ls, tl); /* ran over pattern edge, end */
        run = 0;
      }
    }

  }

  if(mask & LO_STOP_MASK){
    if(mp){
#ifdef DEBUG
      if(mp->p_used == 0){
        fprintf(stderr, "light: logic problem: reference count out\n");
        abort();
      }
#endif
      mp->p_used--;
    }
    ml->l_current = NULL;
    ml->l_index = 0;
    free(ml);
  }

  return 0;
}

int mxk_startup_light(struct list_loop *ls)
{
  int type;

  /* TODO: possibly a reverse EVT type for events going back to hid */
  type = mxk_enroll_find(ls, "light", MXK_EVT_NULL, 32, NULL);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_light(struct mxk_global *mg)
{
  int i;

  if(mg == NULL){
    return;
  }

  if(mg->g_patterns){
    for(i = 0; i < mg->g_psize; i++){
      mxk_free_pattern(mg->g_patterns[i]);
      mg->g_patterns[i] = NULL;
    }
    free(mg->g_patterns);
    mg->g_patterns = NULL;
    mg->g_psize = 0;
  }
}
