/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>

#include <sys/time.h>
#include <sys/types.h>

#include "loop.h"

#define LO_CONTROL_BUFFER    1024   /* limit of data to be read/written in one go */
#define LO_CONTROL_CMD         16   /* max length of command */
#define LO_CONTROL_VECTOR      32

struct command_loop{
  struct command_loop *m_left, *m_right;
  char m_name[LO_CONTROL_CMD];
  int (*m_cmd)(struct list_loop *ls, struct task_loop *tl, char *args);
  char *m_help;
};

struct control_loop{
  char c_output[LO_CONTROL_BUFFER];
  char c_read[LO_CONTROL_BUFFER];
  int c_need;
  int c_have;

  int c_draining;
  int c_nextline;
  int c_args;

  struct command_loop *c_root;

  int c_vector[LO_CONTROL_VECTOR];
  int c_count;
};

static int flush_print_control(struct control_loop *cl, int fd)
{
  int wr;

  if(cl->c_need == 0){
    return 0;
  }

  wr = write(fd, cl->c_output, cl->c_need);
  if(wr < 0){
    switch(errno){
      case EAGAIN :
      case EINTR  :
        break;
      default :
        return -1;
        break;
    } 
#ifdef DEBUG
    fprintf(stderr, "control: flush of %d bytes failed\n", cl->c_need);
#endif
  } else {
#ifdef DEBUG
    fprintf(stderr, "control: flushed %d/%d bytes\n", wr, cl->c_need);
#endif
    if(wr < cl->c_need){
      memmove(cl->c_output, cl->c_output + wr, cl->c_need - wr);
    }
    cl->c_need -= wr;
  }

  return 0;
}

int print_control(struct control_loop *cl, int fd, int flush, char *str, ...)
{
  va_list ap;
  int len, result;

  result = 0;
  va_start(ap, str);

  len = vsnprintf(cl->c_output + cl->c_need, LO_CONTROL_BUFFER - cl->c_need, str, ap);

  if(len > (LO_CONTROL_BUFFER - cl->c_need)){
    flush_print_control(cl, fd);
    len = vsnprintf(cl->c_output, LO_CONTROL_BUFFER - cl->c_need, str, ap);
  }

  if(len >= 0){
    if((len + cl->c_need) > LO_CONTROL_BUFFER){
      strncpy(cl->c_output + (LO_CONTROL_BUFFER - 4), "...\n", 4);
      cl->c_need = LO_CONTROL_BUFFER;
      result = (-1);
    } else {
      cl->c_need += len;
    }
    if(flush || (cl->c_need > (LO_CONTROL_BUFFER / 2))){
      flush_print_control(cl, fd);
    }
  }

  va_end(ap);

  return result;
}

/***************************************************************/

unsigned long number_vector_control(struct control_loop *cl, int index)
{
  if((index < 0) || (cl->c_count <= index)){
#ifdef DEBUG
    fprintf(stderr, "vector: index out of range %d >= %d\n", index, cl->c_count);
#endif
    return 0;
  }

  return strtoul(cl->c_read + cl->c_vector[index], NULL, 10);
}

unsigned long long bignum_vector_control(struct control_loop *cl, int index)
{
  if((index < 0) || (cl->c_count <= index)){
#ifdef DEBUG
    fprintf(stderr, "vector: index out of range %d >= %d\n", index, cl->c_count);
#endif
    return 0;
  }

  return strtoull(cl->c_read + cl->c_vector[index], NULL, 10);
}

int numeric_vector_control(struct control_loop *cl, int index)
{
  int i, x;

  if((index < 0) || (cl->c_count <= index)){
#ifdef DEBUG
    fprintf(stderr, "vector: index out of range %d >= %d\n", index, cl->c_count);
#endif
    return 0;
  }

  i = cl->c_vector[index];
  x = 0;
  if(cl->c_read[i] == '-'){
    i++;
  }
  while(cl->c_read[i] != '\0'){
    switch(cl->c_read[i]){
      case '0' :
      case '1' :
      case '2' :
      case '3' :
      case '4' :
      case '5' :
      case '6' :
      case '7' :
      case '8' :
      case '9' :
        x++;
        break;
      default:
        return 0;
    }
    i++;
  }

  if(x == 0){
    return 0;
  }
  
  return 1;
}

char *string_vector_control(struct control_loop *cl, int index)
{
  if((index < 0) || (cl->c_count <= index)){
#ifdef DEBUG
    fprintf(stderr, "vector: index out of range %d >= %d\n", index, cl->c_count);
#endif
    return NULL;
  }

  return cl->c_read + cl->c_vector[index];
}

int make_vector_control(struct control_loop *cl)
{
  int i, space, quote;

  cl->c_count = 0;

  if(cl->c_args == 0){
    return 0;
  }

  quote = 0;
  space = 1;
  for(i = cl->c_args; cl->c_read[i] != '\0'; i++){
    switch(cl->c_read[i]){
      case '\"' :
        cl->c_read[i] = '\0';
        if(quote == 0){
          quote = 1;
          space = 1;
        } else {
          quote = 0;
          space = 1;
        }
        break;
      case ' ' :
        if(quote == 0){
          if(space == 0){
            cl->c_read[i] = '\0';
            space = 1;
          }
          break;
        } /* else FALL */
      default :
        if(space == 1){
          cl->c_vector[cl->c_count] = i;
          cl->c_count++;
          if(cl->c_count >= LO_CONTROL_VECTOR){
            return -1;
          }
          space = 0;
        }
        break;
    }
  }

  return cl->c_count;
}

/***************************************************************/

static void del_command_control(struct command_loop *cd)
{
  if(cd){
    del_command_control(cd->m_left);
    del_command_control(cd->m_right);
    cd->m_left = NULL;
    cd->m_right = NULL;
    if(cd->m_help){
      free(cd->m_help);
      cd->m_help = NULL;
    }
    free(cd);
  }
}

int shutdown_control(struct control_loop *cl)
{
#ifdef DEBUG
  fprintf(stderr, "control: deallocting state\n");
#endif
  if(cl){
    del_command_control(cl->c_root);
    cl->c_root = NULL;
    cl->c_have = 0;
    cl->c_need = 0;

    cl->c_nextline = 0;
    cl->c_args = 0;
    cl->c_count = 0;

    cl->c_draining = 1;

    free(cl);
  }
  return 0;
}

/***********************************************************************/

/* WARNING: some commands look deeper into system than abstraction layers */

int link_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int i;
  struct control_loop *cl;
  struct pool_loop *pl;
  int fd;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  for(i = 0; i < ls->l_pool_size; i++){
    pl = &(ls->l_pool[i]);
    print_control(cl, fd, 0, "%s [%d]: used=%d, have=%d, max=%d, pool=%d\n", pl->p_name ? pl->p_name : "<no name>", i, pl->p_used, pl->p_have, pl->p_max, pl->p_pool);
  }

  return 0;
}

#ifdef DEBUG
int debug_link_command(int fd, struct control_loop *cl, struct task_loop *tx, int index)
{
  int count;
  struct link_loop *lk;
  struct task_loop *tz;

  count = 0;

  switch(index){
    case LO_INQ :
      print_control(cl, fd, 0, "input queue:\n");
      break;
    case LO_OUTQ :
      print_control(cl, fd, 0, "output queue:\n");
      break;
    default :
      abort();
      break;
  }

  lk = tx->t_links[index];
  if(lk){
    do{
      tz = lk->k_task[1 - index];
      print_control(cl, fd, 0, "  [%d] type=%d, peer=%p, text: %s\n", count, lk->k_type, tz, tz->t_text);
      count++;
    } while(tx->t_links[index] != lk);
  } 

  print_control(cl, fd, 1, "%d links\n", count);

  return 0;
}

int echo_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int fd, i, count;
  struct control_loop *cl;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  count = make_vector_control(cl);
  for(i = 0; i < count; i++){
    print_control(cl, fd, 0, "[%d]: <%s> %s\n", i, string_vector_control(cl, i), numeric_vector_control(cl, i) ? "number" : "string");
  }

  print_control(cl, fd, 1, "%d arguments\n", count);

  return 0;
}

int debug_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int value;
  struct task_loop *tx;
  struct control_loop *cl;
  int fd;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  if(args == NULL){
    value = 0;
    tx = ls->l_head;
  } else {
    value = atoi(args);
    for(tx = ls->l_head; (tx != NULL) && (tx != (struct task_loop *) value); tx = tx->t_next);
  }

  if(tx){
    print_control(cl, fd, 0, "task: %p\n", tx);
    print_control(cl, fd, 0, "text: %s\n", tx->t_text);

    debug_link_command(fd, cl, tx, LO_INQ);
    debug_link_command(fd, cl, tx, LO_OUTQ);
    
  } else {
    print_control(cl, fd, 1, "task 0x%x not found\n", value);
  }
  
  return 0;
}
#endif

int status_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
#define LO_TIME_BUFFER 64
  struct control_loop *cl;
  struct job_loop *jz;
  int fd;
  char buffer[LO_TIME_BUFFER];
  struct tm *tp;
  unsigned int min, sec;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  jz = &(ls->l_job[0]);

  print_control(cl, fd, 0, "tasks: current=%d, maximum=%d, completed=%u\n", jz->j_used, jz->j_max, ls->l_was);

  tp = localtime(&(ls->l_start));
  strftime(buffer, LO_TIME_BUFFER - 1, "%Y-%m-%d %H:%M:%S", tp);
  buffer[LO_TIME_BUFFER - 1] = '\0';

  sec = (ls->l_now.tv_sec - ls->l_start) % 60;
  min = (ls->l_now.tv_sec - ls->l_start) / 60;

  print_control(cl, fd, 0, "time: start=%s, up=%d:%02d:%02d\n", buffer, min / 60, min % 60, sec);

  return 0;
#undef LO_TIME_BUFFER
}

int ps_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int fd;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  for(tl = ls->l_head; tl; tl = tl->t_next){
    print_control(cl, fd, 0, "%p: %2d,%2d %c%c%c%c%c%c%c %2d ", tl, tl->t_type, tl->t_tag,
          (tl->t_mask & LO_WORK_MASK)  ? 'k' : '-', 
          (tl->t_mask & LO_READ_MASK)  ? 'r' : '-', 
          (tl->t_mask & LO_WRITE_MASK) ? 'w' : '-', 
          (tl->t_mask & LO_TIME_MASK)  ? 't' : '-', 
          (tl->t_mask & LO_STOP_MASK)  ? 's' : '-', 
          (tl->t_mask & LO_MORE_MASK)  ? 'm' : '-', 
          (tl->t_mask & LO_WAIT_MASK)  ? 'a' : '-', 
          tl->t_fd);
    if(tl->t_have[LO_INQ]){
      print_control(cl, fd, 0, "%d", tl->t_have[LO_INQ]);
    } else {
      print_control(cl, fd, 0, "*");
    }
    print_control(cl, fd, 0, (tl->t_new) ? "+" : ":");
    if(tl->t_have[LO_OUTQ]){
      print_control(cl, fd, 0, "%d ", tl->t_have[LO_OUTQ]);
    } else {
      print_control(cl, fd, 0, "* ");
    }
    if(tl->t_mask & LO_TIME_MASK){
      print_control(cl, fd, 0, "%2ld.%06ld ", tl->t_interval.tv_sec, tl->t_interval.tv_usec);
    } else {
      print_control(cl, fd, 0, "          ");
    }
    print_control(cl, fd, 0, "%s\n", tl->t_text);
  }

  return 0;
}

static int do_help_command(struct control_loop *cl, struct command_loop *cd, int fd)
{
  if(cd){
    do_help_command(cl, cd->m_left, fd);
    if(cd->m_help){
      print_control(cl, fd, 0, "%s:\t%s\n", cd->m_name, cd->m_help);
    } else {
      print_control(cl, fd, 0, "%s\n", cd->m_name);
    }
    do_help_command(cl, cd->m_right, fd);
  }

  return 0;
}

int help_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  int fd;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  return do_help_command(cl, cl->c_root, fd);
}


int quit_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
#ifdef DEBUG
  fprintf(stderr, "control: running quit\n");
#endif

  stop_loop(ls, tl);

  return 0;
}

int halt_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
#ifdef DEBUG
  fprintf(stderr, "control: running halt\n");
#endif

  ls->l_up = 0;

  return 0;
}

/***********************************************************************/

int add_command_control(struct control_loop *cl, char *name, int (*cmd)(struct list_loop *ls, struct task_loop *tl, char *args))
{
  return add_command_help_control(cl, name, cmd, NULL);
}

int add_command_help_control(struct control_loop *cl, char *name, int (*cmd)(struct list_loop *ls, struct task_loop *tl, char *args), char *help)
{
  struct command_loop *cd, *cx;

  cd = malloc(sizeof(struct command_loop));
  if(cd == NULL){
    return -1;
  }

  if(help){
    cd->m_help = strdup(help);
    if(cd->m_help == NULL){
      free(cd);
      return -1;
    }
  } else {
    cd->m_help = NULL;
  }
  cd->m_left = NULL;
  cd->m_right = NULL;
  cd->m_cmd = cmd;
  memset(cd->m_name, '\0', LO_CONTROL_CMD);
  strncpy(cd->m_name, name, LO_CONTROL_CMD - 1);

  if(cl->c_root == NULL){
    cl->c_root = cd;
    return 0;
  }

  cx = cl->c_root;
  while(cx){
    if(strcmp(cd->m_name, cx->m_name) > 0){
      if(cx->m_right){
        cx = cx->m_right;
      } else {
        cx->m_right = cd;
        return 0;
      }
    } else {
      if(cx->m_left){
        cx = cx->m_left;
      } else {
        cx->m_left = cd;
        return 0;
      }
    }
  }

#ifdef DEBUG
  fprintf(stderr, "add_command: odd: failed to traverse tree\n");
#endif

  return -1;
}

void *default_setup_control(struct list_loop *ls, void *data, struct sockaddr *addr, int fd)
{
  struct control_loop *cl;
  int result = 0;

  cl = malloc(sizeof(struct control_loop));
  if(cl == NULL){
    return NULL;
  }

  cl->c_have = 0;
  cl->c_read[0] = '\0';

  cl->c_need = 0;
  cl->c_output[0] = '\0';

  cl->c_draining = 0;
  cl->c_nextline = 0;
  cl->c_args = 0;
  cl->c_count = 0;

  cl->c_root = NULL;

  result += add_command_help_control(cl, "quit",   &quit_command,   "disconnect control channel");
  result += add_command_help_control(cl, "help",   &help_command,   "this help");
  result += add_command_help_control(cl, "ps",     &ps_command,     "task resources");
  result += add_command_help_control(cl, "link",   &link_command,   "message resources");
  result += add_command_help_control(cl, "halt",   &halt_command,   "shut down the server");
  result += add_command_help_control(cl, "status", &status_command, "uptime and processing information");

#ifdef DEBUG
  result += add_command_control(cl, "debug", &debug_command);
  result += add_command_control(cl, "echo",  &echo_command);
#endif

  if(result){
    shutdown_control(cl);
    return NULL;
  }

  return cl;
}

/***************************************************************/

static void cleanup_control(struct control_loop *cl)
{
  if(cl->c_nextline){
    if(cl->c_nextline < cl->c_have){
      memmove(cl->c_read, cl->c_read + cl->c_nextline, cl->c_have - cl->c_nextline);
      cl->c_have -= cl->c_nextline;
    } else {
      cl->c_have = 0;
    }
    cl->c_read[cl->c_have] = '\0';
    cl->c_args = 0;
    cl->c_count = 0;
    cl->c_nextline = 0;
  }
}

#define CSTATE_END     0
#define CSTATE_START   1
#define CSTATE_CMD     2
#define CSTATE_SPACE   3
#define CSTATE_ARGS    4
#define CSTATE_CRLF    5
#define CSTATE_NEXT    6

static int prepare_control(struct control_loop *cl)
{
  int state, args, word, last, i;

  if(cl->c_nextline){
#ifdef DEBUG
    fprintf(stderr, "prepare: nextline already set\n");
#endif
    return 0;
  }

  word = 0;
  last = 0;
  args = 0;
  state = CSTATE_START;

  for(i = 0; (i < (LO_CONTROL_BUFFER - 1)); i++){
    switch(state){
      case CSTATE_START :
        switch(cl->c_read[i]){
          case '\n' :
          case '\r' :
          case ' '  :
            cl->c_read[i] = '\0';
            cl->c_nextline = i + 1;
            return 0;
          case '\0' :
            cl->c_have = i;
            return 1;
          default :
            state = CSTATE_CMD;
            break;
        }
        break;
      case CSTATE_CMD :
        switch(cl->c_read[i]){
          case '\n' :
          case '\r' :
            word = i;
            last = i;
            state = CSTATE_CRLF;
            break;
          case ' '  :
            word = i;
            state = CSTATE_SPACE;
            break;
          case '\0' :
            cl->c_have = i;
            return 1;
        }
        break;
      case CSTATE_SPACE :
        switch(cl->c_read[i]){
          case ' ' :
            break;
          case '\n' :
          case '\r' :
            last = i;
            state = CSTATE_CRLF;
            break;
          case '\0' :
            cl->c_have = i;
            return 1;
          default :
            state = CSTATE_ARGS;
            args = i;
            break;
        }
        break;
      case CSTATE_ARGS :
        switch(cl->c_read[i]){
          case '\n' :
          case '\r' :
            last = i;
            state = CSTATE_CRLF;
            break;
          case '\0' :
            cl->c_have = i;
            return 1;
          default :
            break;
        }
        break;
      case CSTATE_CRLF :
        switch(cl->c_read[i]){
          case '\n' :
          case '\r' :
            break;
          default :
            cl->c_nextline = i;
            cl->c_args = args;
            cl->c_count = 0;
            cl->c_read[word] = '\0';
            cl->c_read[last] = '\0';
            return 0;
        }
        break;
    }
  }

  return -1;

}

static struct command_loop *find_control(struct control_loop *cl)
{
  struct command_loop *cd, *cm;
  int i;

  cd = cl->c_root;

  i = 0;
  while(cd && (cl->c_read[i] != '\0')){
    if(cl->c_read[i] == cd->m_name[i]){
      i++;
    } else {
      if(cl->c_read[i] < cd->m_name[i]){
#ifdef DEBUG
        fprintf(stderr, "control: comparing [%d]=%c against [%d]@%s, taking left=%p\n", i, cl->c_read[i], i, cd->m_name, cd->m_left);
#endif
        cd = cd->m_left;
      } else {
#ifdef DEBUG
        fprintf(stderr, "control: comparing [%d]=%c against [%d]@%s, taking right=%p\n", i, cl->c_read[i], i, cd->m_name, cd->m_right);
#endif
        cd = cd->m_right;
      }
      i = 0;
    }
  }

  if(cd == NULL){
    return NULL;
  }

  if(cd->m_name[i] == '\0'){
#ifdef DEBUG
    fprintf(stderr, "control: exact match: <%s>=<%s>\n", cl->c_read, cd->m_name);

#endif
    return cd; 
  }

#ifdef DEBUG
  fprintf(stderr, "control: testing for ambiguities\n");
#endif

  if(cd->m_left){ 
    cm = cd->m_left;
    while(cm->m_right){
      cm = cm->m_right;
    }
    if(!strncmp(cm->m_name, cl->c_read, i)){
#ifdef DEBUG
      fprintf(stderr, "control: leftside hit: %s/%d\n", cm->m_name, i);
#endif
      return NULL;
    }
  }

  if(cd->m_right){ 
    cm = cd->m_right;
    while(cm->m_left){
      cm = cm->m_left;
    }
    if(!strncmp(cm->m_name, cl->c_read, i)){
#ifdef DEBUG
      fprintf(stderr, "control: rightside hit: %s/%d\n", cm->m_name, i);
#endif
      return NULL;
    }
  }

#ifdef DEBUG
  fprintf(stderr, "control: seems unique\n");
#endif

  return cd;
}

int run_control(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct control_loop *cl;
  struct command_loop *cd;
  int fd, rr, plan, result;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  if(cl == NULL){
    return 0;
  }

  if(mask & LO_STOP_MASK){
#ifdef DEBUG
    fprintf(stderr, "command: ending\n");
#endif
    shutdown_control(cl);
    return 0;
  }

  if(mask & LO_WRITE_MASK){
    if(flush_print_control(cl, fd) < 0){
      stop_loop(ls, tl);
    }
  }

  if(mask & LO_READ_MASK){
    rr = read(fd, cl->c_read + cl->c_have, (LO_CONTROL_BUFFER - 1) - cl->c_have);
    switch(rr){
      case -1 : 
        switch(errno){
          case EAGAIN :
          case EINTR  :
            break;
          default :
            close(fd);
            stop_loop(ls, tl);
            break;
        }
        break;
      case  0 :
        /* WARNING: may need to drain io */
        cl->c_draining = 1;
        break;
      default :
        cl->c_have += rr;
        break;
    }
    
    cl->c_read[cl->c_have] = '\0';
#ifdef DEBUG
    fprintf(stderr, "control: read %d, total now %d, content=<%s>\n", rr, cl->c_have, cl->c_read);
#endif
  }

  result = prepare_control(cl);

#ifdef DEBUG
  fprintf(stderr, "control: prepare returns=%d, nextline=%d, args=%d, cmd=<%s>, have=%d\n", result, cl->c_nextline, cl->c_args, cl->c_read, cl->c_have);
#endif

  if(result == 0){
    switch(cl->c_read[0]){
      case '\0' :
      case '\n' :
      case '\r' :
      case '#'  :
        break;
      default :
        cd = find_control(cl);
        if(cd){
#ifdef DEBUG
          fprintf(stderr, "control: attempting to run <%s> with arg <%s>\n", cd->m_name, (cl->c_args ? (cl->c_read + cl->c_args) : NULL));
#endif
          result = (*(cd->m_cmd))(ls, tl, cl->c_args ? (cl->c_read + cl->c_args) : NULL);
          if(result < 0){ /* force immediate stop, eg when fd is stolen */
            stop_loop(ls, tl);
            return 0;
          }
        } else {
          print_control(cl, fd, 0, "unknown or incomplete command: try help\n");
        }
        break;
    }
    flush_print_control(cl, fd);
  } else {
    if(result < 0){
      print_control(cl, fd, 1, "unusuable command stream: giving up\n");
      stop_loop(ls, tl);
    }
  }

  cleanup_control(cl);

  /* figure out how to schedule ourselves */
  plan = LO_STOP_MASK;
  if(cl->c_need > 0){
    plan |= LO_WRITE_MASK;
  }
  if((cl->c_need < (LO_CONTROL_BUFFER / 2)) && (cl->c_draining == 0)){
    plan |= LO_READ_MASK;
  }
  if(prepare_control(cl) == 0){
    plan |= LO_WORK_MASK;
  } else {
    if(cl->c_draining && (cl->c_need == 0)){
      plan = LO_STOP_MASK;
      stop_loop(ls, tl);
    }
  }

  set_mask_loop(ls, tl, plan, 0);

  return 0;
}

