/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <pwd.h>
#include <netdb.h>
#include <sched.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>

#include "loop.h"

#define LO_CONSOLE_BUFFER 1024
#define LO_CONSOLE_DRAIN  250000

static int connect_console(char *name);
static int feed_console(int cfd, int bfd, int afd);
static int drain_console(int cfd, int afd);
static void parent_console(int cfd, int flags, char *profile, pid_t pid);

static int connect_console(char *name)
{
  int fd;
  struct sockaddr_un sa;

  sa.sun_family = AF_UNIX;
  strncpy(sa.sun_path, name, sizeof(sa.sun_path));

  if(name == NULL){
    return -1;
  }

  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if(fd < 0){
    return -1;
  }

  if(connect(fd, (struct sockaddr *)&sa, sizeof(sa)) != 0){
    close(fd);
    return -1;
  }

  return fd;
}

static int drain_console(int cfd, int afd)
{
  char buffer[LO_CONSOLE_BUFFER];
  int rr, i;
  struct timeval tv;
  fd_set fsr;

  for(;;){
    tv.tv_sec = 0;
    tv.tv_usec = LO_CONSOLE_DRAIN;

    FD_ZERO(&fsr);
    for(i = 0; i < 2; i++){
      FD_SET(cfd, &fsr);
    }

    rr = select(cfd + 1, &fsr, NULL, NULL, &tv);
    if(rr > 0){
      rr = read(cfd, buffer, LO_CONSOLE_BUFFER);
    }

    if(rr <= 0){
      return (rr < 0) ? (-1) : 0;
    }

    if(write(afd, buffer, rr) != rr){
      return -1;
    }
  }

  close(cfd);

  return 0;
}

static int feed_console(int cfd, int bfd, int afd)
{
  int max, rr, wr, i;
  int rfd[2], wfd[2];
  char buffer[LO_CONSOLE_BUFFER];
  fd_set fsr;

  rfd[0] = cfd;
  rfd[1] = bfd;
  /* WARNING: write array is inverted */
  wfd[1] = cfd;
  wfd[0] = afd;

  max = (-1);
  for(i = 0; i < 2; i++){
    if(rfd[i] > max){
      max = rfd[i];
    }
    if(wfd[i] > max){
      max = wfd[i];
    }
  }
  max++;
  
  for(;;){
    FD_ZERO(&fsr);
    for(i = 0; i < 2; i++){
      FD_SET(rfd[i], &fsr);
    }
    rr = select(max, &fsr, NULL, NULL, NULL);
    if(rr < 0){
      fprintf(stderr, "console: select failed: %s\n", strerror(errno));
      return -1;
    }
    for(i = 0; i < 2; i++){
      if(FD_ISSET(rfd[i], &fsr)){
        rr = read(rfd[i], buffer, LO_CONSOLE_BUFFER);
        switch(rr){
          case -1 :
            fprintf(stderr, "console: read failed: %s\n", strerror(errno));
            return -1;
          case 0 :
            /* WARNING: tell caller which side has finished */
            return (rfd[i] == bfd) ? 1 : 0;
            /* default does the write below */
        }
        wr = write(wfd[i], buffer, rr);
        if(wr < rr){
          if(wr < 0){
            fprintf(stderr, "console: write failed: %s\n", strerror(errno));
          } else {
            fprintf(stderr, "console: short write\n");
          }
          return -1;
        }
      }
    }
  }

  /* not reached */
}

static void parent_console(int cfd, int flags, char *profile, pid_t pid)
{
  int code, afd, status, again, result;

  code = 1;
  result = EX_OK;

  /* send script file to server */
  if((flags & LO_FILE_CONSOLE) && profile){
    afd = open(profile, O_RDONLY);
    if(afd < 0){
      fprintf(stderr, "console: unable to open command file %s: %s\n", profile, strerror(errno));
    } else {
      code = feed_console(cfd, afd, STDOUT_FILENO);
      close(afd);
    }
  }

  /* attach interactive console to server */
  if((flags & LO_RUN_CONSOLE) && (code > 0)){
    fprintf(stderr, "console: ready\n");
    code = feed_console(cfd, STDIN_FILENO, STDOUT_FILENO);
  }

  /* cleanup */
  if(code > 0){
    drain_console(cfd, STDOUT_FILENO);
  } else {
    if(code < 0){
      fprintf(stderr, "console: link to server process failed\n");
      result = EX_SOFTWARE;
    }
  }

  sched_yield();

  /* collect status of (potentially) exiting server */
  if((flags & LO_WAIT_CONSOLE) && (pid > 0)){
    for(again = 1; again; ){
      if (waitpid(pid, &status, (again == 1) ? WNOHANG : 0) > 0) {	/* got a child */
        if (WIFEXITED(status)) {
          result = WEXITSTATUS(status);
          if(result){
            fprintf(stderr, "console: abnormal server exit: code %d\n", result);
          } else {
            fprintf(stderr, "console: normal server exit\n");
          }
        } else if (WIFSIGNALED(status)) {
          fprintf(stderr, "console: server killed by signal %d\n", WTERMSIG(status));
          fflush(stderr);
          result = EX_SOFTWARE;
        }
        again = 0;
      } else {
        if(flags & LO_DETACH_CONSOLE){
          fprintf(stderr, "console: server now running detached from terminal\n");
          again = 0;
        } else {
          fprintf(stderr, "console: server continues running on terminal\n");
          /* WARNING: sneaky, again > 1 is different, see above */
          again++;
        }
      }
    }
  }

  /* exit ourselves */
  exit(result);
}

/*************************************************/

/* old interface */

int console_fd(int detach)
{
  int flags;

  flags = LO_RUN_CONSOLE;
  if(detach){
    flags |= LO_DETACH_CONSOLE;
  }
 
  return start_profile_console(NULL, flags);
}

int start_profile_console(char *profile, int flags)
{
  int pair[2];
  pid_t pid;
#if 0
  struct sigaction sag;
#endif

  if(socketpair(AF_UNIX, SOCK_STREAM, 0, pair) < 0){
    return -1;
  }

  pid = fork();
  switch (pid) {
  case -1:
    return -1;
  case 0:
    if(flags & LO_DETACH_CONSOLE){
      setsid();
    }

    close(pair[0]);
    misc_unblock_loop(pair[1]);

    if(pair[1] != STDOUT_FILENO){
      close(STDOUT_FILENO);
    }

    if(pair[1] != STDIN_FILENO){
      if(dup2(pair[1], STDIN_FILENO) != STDIN_FILENO){
        return -1;
      }
      close(pair[1]);
    }

    fcntl(STDIN_FILENO, F_SETFD, FD_CLOEXEC);
    fcntl(STDOUT_FILENO, F_SETFD, FD_CLOEXEC);

    return STDIN_FILENO;
  default:
    close(pair[1]);

    fflush(stdout);

    parent_console(pair[0], flags | LO_WAIT_CONSOLE, profile, pid);

    /* never returns */
    return -1;
  }
}

void commit_attach_console(char *name, int flags)
{
  int fd;

  fd = connect_console(name);
  if(fd < 0){
    return;
  }
  
  parent_console(fd, flags, NULL, 0);
  /* does not return */
}

struct task_loop *add_console_loop(struct list_loop *ls, int fd, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), char *name)
{
  struct task_loop *tl;

  tl = add_loop(ls, fd, data, mask, call, NULL, 0);
  if(tl == NULL){
    /* WARNING: up to caller to clean up fd and data */
    return NULL;
  }
  
  text_loop(ls, tl, "con active: %s", name ? name : "batch");

  return tl;
}

