/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sched.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "loop.h"

#define LO_MON_BUFFER   64000

static volatile int monitor_enable = 0;
static volatile int monitor_check = 1;
static volatile int monitor_run = 1;
static volatile int monitor_detached = 0;

static int monitor_up = (-1);

#define LO_BAND_ESCAPE  27
#define LO_BAND_EXIT     3  /* C */
#define LO_BAND_DETACH  26  /* Z */
#define LO_BAND_ENABLE  17  /* Q */
#define LO_BAND_DISABLE 19  /* S */

static void monitor_signal(int s)
{
  switch(s){
    case SIGHUP :
    case SIGINT : /* detach from terminal */
      monitor_check = 1;
      monitor_detached = 1;
      break;
    case SIGUSR1 : /* enable io */
      monitor_check = 1;
      monitor_enable = 1;
      break;
    case SIGUSR2 : /* disable io */
      monitor_check = 1;
      monitor_enable = 0;
      break;
    case SIGTERM : /* stop */
      monitor_run = 0;
      break;
    /* case SIGALRM : nothing: just have system call return -1 */
  }
}

static int fancy_open(char *name, int flags)
{
  int fd;
  int bits;
  mode_t mode;

  if(name){
    mode = S_IWUSR | S_IRUSR;
    if(LO_MON_PUBLIC){
      mode |= S_IRGRP | S_IROTH;
    }

    if(flags & LO_MON_FIFO){
      mknod(name, mode | S_IFIFO, 0);
    }

    bits = O_CREAT | O_NOCTTY | O_WRONLY;
    if(flags & LO_MON_TRUNCATE){
      bits |= O_TRUNC;
    } else {
      bits |= O_APPEND;
    }
    if(flags & LO_MON_SYNC){
      bits |= O_SYNC;
    } else if(!(flags & LO_MON_FIFO)){
      bits |= O_NONBLOCK;
    }

    alarm(2);
    fd = open(name, bits, mode);
    alarm(0);
#ifdef DEBUG
    if(fd < 0){
      fprintf(stderr, "unable to open %s: %s\n", name, strerror(errno));
    }
#endif
  } else { 
    fd = (-1);
  }

  return fd;
}

static int sync_chase(char *buffer, int index)
{
  int i;

  for(i = index; i < LO_MON_BUFFER; i++){
    if(buffer[i] == '\n'){
      return ((i + 1) < LO_MON_BUFFER) ? (i + 1) : (LO_MON_BUFFER - 1);
    }
  }
  for(i = 0; i < index; i++){
    if(buffer[i] == '\n'){
      return ((i + 1) < LO_MON_BUFFER) ? (i + 1) : (LO_MON_BUFFER - 1);
    }
  }

  return index;
}

static void inband_monitor(char *buffer, int len)
{
  static int mark = 0;
  int i;

  for(i = 0; i < len; i++){
    switch(buffer[i]){
      case LO_BAND_ESCAPE :
        mark = 1;
        buffer[i] = '\n';
        break;
      default :
        if(mark){
          monitor_check = 1;
          switch(buffer[i]){
            case LO_BAND_ENABLE : /* xon */
              monitor_enable = 1;
              break;
            case LO_BAND_DETACH : /* go into background */
              monitor_detached = 1;
              break;
            case LO_BAND_DISABLE : /* xoff */
              monitor_enable = 0;
              break;
            case LO_BAND_EXIT : /* xoff */
              monitor_run = 0;
              break;
            default : /* nevermind */
              monitor_check = 0;
              break;
          }
          buffer[i] = '\f';
          mark = 0;
        }
        break;
    }

  }
}

static int run_monitor(int efd, char *name, int flags)
{
  int ffd;
  char buffer[LO_MON_BUFFER];
  int index, chase, clobber, restart, detached, error;
  int rr, wr;
  struct sigaction sag;
  sigset_t sst;

  misc_close_loop(STDOUT_FILENO, efd, -1);
  misc_close_loop(STDIN_FILENO, efd, -1);

  if(name && !(flags & LO_MON_DEFER)){
    setsid();
    close(STDERR_FILENO);
    detached = 1;
  } else {
    detached = 0;
  }

  sigfillset(&(sag.sa_mask));
  sag.sa_flags = 0;
  sag.sa_handler = monitor_signal;

  if(name){
    if(flags & LO_MON_DEFER){
      if(sigaction(SIGHUP, &sag, NULL)){ /* on hangup cut over to file */
        return EX_OSERR;
      }
      if(sigaction(SIGINT, &sag, NULL)){ /* same */
        return EX_OSERR;
      }
    }
  }
  if(sigaction(SIGUSR1, &sag, NULL)){ /* open file, strictly only an option of name is set (not tested, since inband will still do) */
    return EX_OSERR;
  }
  if(sigaction(SIGUSR2, &sag, NULL)){ /* close file */
    return EX_OSERR;
  }
  if(sigaction(SIGTERM, &sag, NULL)){ /* stop */
    return EX_OSERR;
  }
  if(sigaction(SIGALRM, &sag, NULL)){ /* have open fail with -1 */
    return EX_OSERR;
  }

  sag.sa_flags = SA_RESTART;
  sag.sa_handler = SIG_IGN;
  if(sigaction(SIGPIPE, &sag, NULL)){ /* ignore write failures */
    return EX_OSERR;
  }

  sigemptyset(&sst);
  sigaddset(&sst, SIGHUP);
  sigaddset(&sst, SIGINT);
  sigaddset(&sst, SIGTERM);
  sigaddset(&sst, SIGUSR1);
  sigaddset(&sst, SIGUSR2);

  sigprocmask(SIG_UNBLOCK, &sst, NULL);

  if(flags & LO_MON_START){
    monitor_enable = 1;
  }

  memset(buffer, '\0', LO_MON_BUFFER);
  buffer[0] = '\n';
  index = 1;
  chase = 0;

  restart = 0;
  clobber = 0;
  error = 0;

  ffd = (-1);

  while(monitor_run){

    if(efd < 0){
      pause();
    } else {
      rr = read(efd, buffer + index, LO_MON_BUFFER - index);
      switch(rr){
        case -1: 
          switch(errno){
            case EAGAIN :
            case EINTR :
              break;
            default :
              close(efd);
              efd = (-1);
              error = 1;
              break;
          }
          break;
        case  0 :
          close(efd);
          efd = (-1);
          break;
        default :
          inband_monitor(buffer + index, rr);
          if((chase > index) && (chase < (index + rr))){
            clobber++;
          }
          index += rr;
          if(index >= LO_MON_BUFFER){
            index = 0;
          }
          break;
      }
    }

    if(efd < 0){ /* monitored entity has gone away */
      if((error == 0) && (detached == 0)){ /* exit if no problem and not yet detached (implicit to DETACH flag) */
        monitor_run = 0;
      } 
      if((flags & LO_MON_EXIT) && (error == 0)){ /* exit if no problem and asked to exit */
        monitor_run = 0;
      }
      if((flags & LO_MON_ABORT) && (ffd < 0)){  /* if no io and asked to do io, do io */
        monitor_check = 1;
        monitor_enable = 1;
      }
#ifdef DEBUG
      fprintf(stderr, "efd = (-1): run=%d, check=%d, enable=%d\n", monitor_run, monitor_check, monitor_enable);
#endif
    }

    if(monitor_check){
      monitor_check = 0;
      sigprocmask(SIG_BLOCK, &sst, NULL);
      if(monitor_detached && (detached == 0)){ /* detect transition */
        setsid();
        close(STDERR_FILENO);
        detached = 1;
      } else {
        if(detached){
          close(ffd);
        }
      }
      ffd = (-1); /* ffd now disabled, check if we need to (re)enable  */
      if(monitor_enable){ 
        if(detached){
          ffd = fancy_open(name, flags);
        } else {
          ffd = STDERR_FILENO;
        }
        restart = 1;
      }
    } else {
      if((ffd < 0) && (efd > 0)){
        continue; /* WARNING: saves us system calls */
      }
      sigprocmask(SIG_BLOCK, &sst, NULL);
    }

    if(ffd >= 0){ /* should output */

      if(clobber || restart){
#ifdef DEBUG
        fprintf(stderr, "resync: clobber=%d, restart=%d\n", clobber, restart);
#endif
        clobber = 0;
        restart = 0;
        /* TODO: something to indicate restart or clobber */
        chase = sync_chase(buffer, index);
      }

      wr = 0;
      if(chase > index){
        wr = write(ffd, buffer + chase, LO_MON_BUFFER - chase);
        if(wr > 0){
          chase += wr;
          if(chase >= LO_MON_BUFFER){
            chase = 0;
          }
        }
      }
      if(chase < index){
        wr = write(ffd, buffer + chase, index - chase);
        if(wr > 0){
          chase += wr;
        }
      }

      if(wr < 0){
        switch(errno){
          case EAGAIN :
          case EINTR :
            break;
          default :
            if(detached){
              close(ffd);
            }
            ffd = (-1);
            error = 1;
            break;
        }
      }
    }

    sigprocmask(SIG_UNBLOCK, &sst, NULL);
    
  }

  return EX_OK;
}

/******************************************************/

void error_enable()
{
  if(monitor_up > 0){
    fprintf(stderr, "%c%c", LO_BAND_ESCAPE, LO_BAND_ENABLE);
    fflush(stderr);
  }
}

void error_detach()
{
  if(monitor_up > 0){
    fprintf(stderr, "%c%c", LO_BAND_ESCAPE, LO_BAND_DETACH);
    fflush(stderr);
  }
}

void error_disable()
{
  if(monitor_up > 0){
    fprintf(stderr, "%c%c", LO_BAND_ESCAPE, LO_BAND_DISABLE);
    fflush(stderr);
  }
}

void error_absolve()
{
#ifdef DEBUG
  int fd;
#endif

  if(monitor_up > 0){
#if 0 /* two processes are writing, rely on close() to notify reader */
    fprintf(stderr, "%c%c", LO_BAND_ESCAPE, LO_BAND_EXIT);
#endif
    fflush(stderr);
#ifdef DEBUG
    /* stderr now /dev/null or to a broken pipe */
    fd = open("/dev/null", O_WRONLY);
    if((fd >= 0) && (fd != STDERR_FILENO)){
      dup2(fd, STDERR_FILENO);
      close(fd);
    }
#else
    /* WARNING: stderr now recyclable */
    close(STDERR_FILENO);
#endif

    monitor_up = 0;
  }
}


void error_exit(int code, char *format, ...)
{
  va_list ap;

  if(monitor_up > 0){
    if(format){
      va_start(ap, format);
      vfprintf(stderr, format, ap);
      va_end(ap);
    }

    fflush(stderr);
    monitor_up = 0;
  }

  exit(code);
}

int error_monitor(char *name, int flags)
{
  int ep[2];
  int status;
  int result;
  pid_t pid;
  
  /* Sets up a child process. Child will capture parent stderr */
  
  if(monitor_up >= 0){
    return -1;
  }

  if(pipe(ep)){
    return -1;
  }

  /* we expect stderr to be valid, otherwise defer does not work */
  if((ep[0] == STDERR_FILENO) || (ep[1] == STDERR_FILENO)){
    close(ep[0]);
    close(ep[1]);
    return -1;
  }

  fflush(stderr);

  pid = fork();
  switch(pid){
    case -1 :
      close(ep[0]);
      close(ep[1]);
      return -1;

    case 0 :
      close(ep[1]);

      result = run_monitor(ep[0], name, flags);

      exit(result);
      return result; /* consistency */

    default : /* parent */
      close(ep[0]);

      sched_yield();
      sched_yield();

      result = waitpid(pid, &status, WNOHANG);

      if(result == pid){
        close(ep[1]);
        return -1;
      }

      if(dup2(ep[1], STDERR_FILENO) == STDERR_FILENO){
        monitor_up = 1;
        close(ep[1]);
        return 0; /* ok */
      }

      close(ep[1]);
      return -1;
  }

  /* WARNING: switch contains all returns */
}
