/*
 * (c) Marc Welz 1999, 2004, 2005, 2006: 
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer 
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#ifndef LOOP_H_
#define LOOP_H_

#include <stdarg.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define LO_READ_MASK  0x01  /* fd can be read */
#define LO_WRITE_MASK 0x02  /* fd can be written */
#define LO_TIME_MASK  0x04  /* timer expires */
#define LO_WORK_MASK  0x08  /* just run, mkay */
#define LO_STOP_MASK  0x10  /* about to stop, clean up */
#define LO_MORE_MASK  0x20  /* can send more messages (under local out quota) */
#define LO_WAIT_MASK  0x40  /* received a message */

#define LO_BREAK_MASK 0x80  /* OLD, UNUSED */

struct list_loop;
struct task_loop;
struct link_loop;
struct pool_loop;

#define LO_INQ              0
#define LO_OUTQ             1
#define LO_FREEQ       LO_INQ

#define LO_NEW_FORM      0x01
#define LO_ACK_FORM      0x02
#define LO_BOTH_FORM  (LO_NEW_FORM | LO_ACK_FORM)

struct link_loop{
  int k_type;
  void *k_data;
  struct link_loop *k_back[2];
  struct link_loop *k_next[2];
  struct task_loop *k_task[2];
};

struct pool_loop{
  int p_type;
  void (*p_empty)(struct list_loop *ls, void *data);
  struct link_loop *p_free;
  int p_used;
  int p_have;

  int p_max; 
  int p_pool;

  char *p_name;
};

struct job_loop{
  int j_type;

  int j_used;
  int j_max;

  char *j_name;
};

#define LO_TEXT_SIZE 32

#define LO_STOP_NOT         0
#define LO_STOP_PREPARE     1
#define LO_STOP_FINISH      2

struct task_loop{
  int (*t_call)(struct list_loop *ls, struct task_loop *tl, int mask);
  unsigned short t_mask;
  unsigned char t_stop;
  struct timeval t_interval;
  struct timeval t_timeout;
  int t_fd;
  void *t_data;
  int t_type;
  int t_tag;
  struct task_loop *t_next;
  char t_text[LO_TEXT_SIZE];
  struct link_loop *t_links[2];
  struct link_loop *t_new;
  int t_have[2];
  int t_max[2];
};

struct list_loop{
  int l_max;
  int l_mfd;
  int l_up;
  struct timeval l_now;
  void *l_data;
  struct task_loop *l_head;
  struct task_loop *l_spare;
  int l_spare_size;

  struct pool_loop *l_pool; /* [0] is special */
  int l_pool_size;

  struct job_loop *l_job;
  int l_job_size;

  time_t l_start;
  unsigned int l_was;
};

struct list_loop *startup_loop(void *gb, int tasks, int links);
int shutdown_loop(struct list_loop *ls);

#define LO_FAVOUR_IO     -1
#define LO_FAVOUR_BALANCE 0
#define LO_FAVOUR_CPU     1

int main_loop(struct list_loop *ls);
int main_loop_mode(struct list_loop *ls, int mode);

struct task_loop *add_loop(struct list_loop *ls, int fd, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);
struct task_loop *add_tag_loop(struct list_loop *ls, int fd, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type, int tag);

int stop_loop(struct list_loop *ls, struct task_loop *tl);
int unwind_loop(struct list_loop *ls, struct task_loop *tl); /* like a pop off the stack */

/* modes: if below zero remove flag from set, above or it with existing. If zero overwrite completely */
int set_fd_loop(struct list_loop *ls, struct task_loop *tl, int fd);
int set_mask_loop(struct list_loop *ls, struct task_loop *tl, int mask, int mode);

int set_user_loop(struct list_loop *ls, struct task_loop *tl, void *data);
int set_type_loop(struct list_loop *ls, struct task_loop *tl, int type);
int set_time_loop(struct list_loop *ls, struct task_loop *tl, struct timeval *tv);
int set_call_loop(struct list_loop *ls, struct task_loop *tl, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask));
void set_max_loop(struct list_loop *ls, struct task_loop *tl, int queue, int max);

struct link_loop *send_link_loop(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type);
int discard_link_loop(struct list_loop *ls, struct link_loop *nk);
struct link_loop *receive_link_loop(struct list_loop *ls, struct task_loop *tl);
int ack_link_loop(struct list_loop *ls, struct task_loop *tl, struct link_loop *nk);
void reset_link_loop(struct list_loop *ls, struct task_loop *tl);
struct link_loop *work_link_loop(struct list_loop *ls, struct task_loop *tl);
int detach_link_loop(struct list_loop *ls, struct link_loop *nk);
int reply_link_loop(struct list_loop *ls, struct link_loop *nk);
int respond_link_loop(struct list_loop *ls, struct link_loop *nk);

struct task_loop *find_task_loop(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type);
struct task_loop *find_task_tag_loop(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type, int tag);

void *user_loop(struct list_loop *ls, struct task_loop *tl);
int fd_loop(struct list_loop *ls, struct task_loop *tl);
int mask_loop(struct list_loop *ls, struct task_loop *tl);
int tag_loop(struct list_loop *ls, struct task_loop *tl);
int type_loop(struct list_loop *ls, struct task_loop *tl);
char *type_name_loop(struct list_loop *ls, struct task_loop *tl);
void *global_loop(struct list_loop *ls);
struct timeval *now_loop(struct list_loop *ls);
void text_loop(struct list_loop *ls, struct task_loop *tl, char *str, ...);

void *alloc_link_loop(struct list_loop *ls, struct link_loop *nk, int type, int size);

/* WARNING: logic change - null empty function now means do nothing */
int claim_link_name_loop(struct list_loop *ls, int lm, int lp, char *name, void (*empty)(struct list_loop *ls, void *data));
int claim_link_loop(struct list_loop *ls, int lm, int lp, void (*empty)(struct list_loop *ls, void *data));
int usage_link_loop(struct list_loop *ls, int type);
int limit_link_loop(struct list_loop *ls, int type);
int lookup_link_type_loop(struct list_loop *ls, char *name);

int claim_task_name_loop(struct list_loop *ls, int tm, int tp, char *name);
int claim_task_loop(struct list_loop *ls, int tm, int tp);
int usage_task_loop(struct list_loop *ls, int type);
int limit_task_loop(struct list_loop *ls, int type);
int lookup_task_type_loop(struct list_loop *ls, char *name);

#if 0
int claim_type_loop(struct list_loop *ls, int tm, int tp, int lm, int lp, void (*empty)(struct list_loop *ls, void *data));
#endif

/* pass to claim_type_loop to free data */
void default_empty_loop(struct list_loop *ls, void *data);

/* api for console interface: monitor stdin/stdout of application */

#define LO_NULL_CONSOLE   0x00
#define LO_DETACH_CONSOLE 0x01
#define LO_RUN_CONSOLE    0x02
#define LO_FILE_CONSOLE   0x04
#define LO_WAIT_CONSOLE   0x08

struct task_loop *add_console_loop(struct list_loop *ls, int fd, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), char *name);
int start_profile_console(char *profile, int flags);
void commit_attach_console(char *name, int flags);

/* old interface */
int console_fd(int detach);

/* api for socket interface: run something on listening ip or unix socket */

struct task_loop *add_unix_listen_loop(struct list_loop *ls, char *name, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type); 
struct task_loop *add_unix_loop(struct list_loop *ls, char *name, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), int type); 

struct task_loop *add_tcp_joint_listen_loop(struct list_loop *ls, char *name, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);
struct task_loop *add_tcp_split_listen_loop(struct list_loop *ls, char *address, char *port, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);
struct task_loop *add_tcp_joint_loop(struct list_loop *ls, char *name, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), int type);
struct task_loop *add_tcp_split_loop(struct list_loop *ls, char *address, char *port, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), int type);

struct task_loop *connect_tcp_joint_loop(struct list_loop *ls, char *name, void *data, void (*peer)(struct list_loop *ls, void *data, struct sockaddr *addr), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);
struct task_loop *connect_tcp_split_loop(struct list_loop *ls, char *address, char *port, void *data, void (*peer)(struct list_loop *ls, void *data, struct sockaddr *addr), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);
struct task_loop *connect_tcp_direct_loop(struct list_loop *ls, struct sockaddr *addr, void *data, void (*peer)(struct list_loop *ls, void *data, struct sockaddr *addr), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);

struct task_loop *add_udp_joint_peer_loop(struct list_loop *ls, char *from, char *to, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);
struct task_loop *add_udp_split_peer_loop(struct list_loop *ls, char *from_name, char *from_port, char *to_name, char *to_port, void *data, int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type);

/* api for commandline interface: can define a replacement for default_setup_control 
 * which calls add_command_control to extend command set. commands may use print_control 
 * for output, output per event needs to be bounded, otherwise excess will be elided. 
 * run needs to be added with READ and STOP flags. 
 */ 

struct control_loop;

void *default_setup_control(struct list_loop *ls, void *data, struct sockaddr *addr, int fd);
int shutdown_control(struct control_loop *cl);
int run_control(struct list_loop *ls, struct task_loop *tl, int mask);

int print_control(struct control_loop *cl, int fd, int flush, char *str, ...);

int make_vector_control(struct control_loop *cl);
char *string_vector_control(struct control_loop *cl, int index);
int numeric_vector_control(struct control_loop *cl, int index);
unsigned long number_vector_control(struct control_loop *cl, int index);
unsigned long long bignum_vector_control(struct control_loop *cl, int index);

int add_command_control(struct control_loop *cl, char *name, int (*cmd)(struct list_loop *ls, struct task_loop *tl, char *args));
int add_command_help_control(struct control_loop *cl, char *name, int (*cmd)(struct list_loop *ls, struct task_loop *tl, char *args), char *help);

/* resolver infrastructure */

struct task_loop *start_resolver_loop(struct list_loop *ls, char *id, int size, int expire);
int ask_host_resolver_loop(struct list_loop *ls, struct task_loop *tl, char *host);
int get_host_resolver_loop(struct list_loop *ls, struct task_loop *tl, struct in_addr *addr, char *message, int len);

/* error stuff */

#define LO_MON_FIFO     0x01   /* file should be a fifo */
#define LO_MON_TRUNCATE 0x02   /* truncate log file */
#define LO_MON_PUBLIC   0x04   /* make permission more relaxed */
#define LO_MON_SYNC     0x08   /* make writes synchronous */

#define LO_MON_START    0x10   /* enable output at startup */
#define LO_MON_DEFER    0x20   /* defer detach, only background when error_detach is called */
#define LO_MON_ABORT    0x40   /* on abnormal exit emit buffered log */
#define LO_MON_EXIT     0x80   /* exit if parent disappears */

#define LO_MON_SYSLOG  0x100   /* TODO: write syslog entries */

int error_monitor(char *name, int flags); /* start monitor infrastructure (subprocess) */
void error_enable();                      /* (re)start io */
void error_disable();                     /* stop io */
void error_detach();                       /* make error monitor detach from tty */
void error_absolve();                      /* shut down error monitor */
void error_exit(int code, char *format, ...); /* give child a chance to print error exit */

/* support stuff */

int misc_unblock_loop(int fd);
int misc_xnprintf_loop(char *buffer, size_t size, char *format, ...);
void misc_close_loop(int fd, ...);
int misc_drop_loop(char *id, char *dir, int resource);

#endif
