/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

/*
 * sigh, workaround for a recent glibc brainfade - SO_PEERCRED
 * is visible without _GNU_SOURCE, but the matching structure
 * isn't. SO_PEERCRED is a marvelous feature, it should be 
 * sufficient to test just for the define, in case non GNU platforms 
 * decide to implement it too. But no, that would be too simple
 */
#define _GNU_SOURCE

#if 0
struct ucred {
  unsigned int pid;
  unsigned int uid;
  unsigned int gid;
};
#endif

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>

#include <loop.h>

#define LO_UNIX_BACKLOG 8
#define LO_UNIX_PATH  512

struct unix_loop{
  int u_mask;
  void *u_data;
  void *(*u_setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd);
  int (*u_call)(struct list_loop *ls, struct task_loop *tl, int mask);
  struct timeval u_timeout;
  int u_type;
};

static int unix_listen_loop(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct unix_loop *ul;
  struct sockaddr_un sa;
  struct task_loop *tx;
  int fd, nd;
  unsigned int len;
  void *data;
#ifdef SO_PEERCRED
  struct ucred cr;
  struct passwd *pw;
  unsigned int cl;
#endif
  
  ul = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  if(mask & LO_STOP_MASK){
    free(ul);
    return 0;
  }

  if(mask & LO_READ_MASK){
    len = sizeof(struct sockaddr_un);
    nd = accept(fd, (struct sockaddr *) &sa, &len);
    if(nd >= 0){
      if(misc_unblock_loop(nd) == 0){
        tx = add_loop(ls, nd, NULL, ul->u_mask, ul->u_call, (ul->u_mask & LO_TIME_MASK) ? (&(ul->u_timeout)) : NULL, ul->u_type);
        if(tx){
          data = (ul->u_setup) ? (*(ul->u_setup))(ls, ul->u_data, (struct sockaddr *) &sa, nd) : NULL;
          if(data || (ul->u_setup == NULL)){
            set_user_loop(ls, tx, data);
#ifdef SO_PEERCRED
            cl = sizeof(struct ucred);
            if(getsockopt(nd, SOL_SOCKET, SO_PEERCRED, &cr, &cl) == 0) {
              /* FIXME: hey, getpwuid belongs into the resolver */
              pw = getpwuid(cr.uid);
              if(pw){
                text_loop(ls, tx, "unx connected: %s/%u", pw->pw_name, cr.pid);
              } else {
                text_loop(ls, tx, "unx connected: %u/%u", cr.uid, cr.pid);
              }
            } else {
#endif
              text_loop(ls, tx, "unx connected: unknown");
#ifdef SO_PEERCRED
            }
#endif
            return 0;
          }
#ifdef DEBUG
          set_call_loop(ls, tx, NULL);
#endif
          set_mask_loop(ls, tx, 0, 0);
          stop_loop(ls, tl);
          nd = (-1);
        }
      }
      if(nd >= 0){
        close(nd);
      }
    }
  } 

  return -1;
}

struct task_loop *add_unix_loop(struct list_loop *ls, char *name, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), int type)
{
  return add_unix_listen_loop(ls, name, data, setup, mask, call, NULL, type);
}

struct task_loop *add_unix_listen_loop(struct list_loop *ls, char *name, void *data, void *(*setup)(struct list_loop *ls, void *data, struct sockaddr *addr, int fd), int mask, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask), struct timeval *tv, int type)
{
  int fd;
  struct sockaddr_un sa;
  struct unix_loop *ul;
  struct task_loop *tl;
  char chase_source[LO_UNIX_PATH];
  char chase_target[LO_UNIX_PATH];
  int result;

  result = snprintf(chase_source, LO_UNIX_PATH, "%s", name);
  chase_source[LO_UNIX_PATH - 1] = '\0';
  if((result < 0) || (result >= LO_UNIX_PATH)){
#ifdef DEBUG
    fprintf(stderr, "add unix: name %s too long\n", name);
#endif
    return NULL;
  }

  do{
#ifdef DEBUG
    fprintf(stderr, "add unix: chasing %s\n", chase_source);
#endif
    result = readlink(chase_source, chase_target, LO_UNIX_PATH - 1);
    if(result > 0){
      memcpy(chase_source, chase_target, result);
      chase_source[result] = '\0';
    }
  } while(result > 0);

  switch(errno){
    case ENOENT :
    case EINVAL :
      break;
    default :
#ifdef DEBUG
      fprintf(stderr, "add unix: chasing %s failed: %s\n", chase_source, strerror(errno));
#endif
      return NULL;
  }

  sa.sun_family = AF_UNIX;
  result = snprintf(sa.sun_path, sizeof(sa.sun_path) - 1, "%s.%d", chase_source, getpid());
  sa.sun_path[sizeof(sa.sun_path) - 1] = '\0';
  if((result < 0) || (result >= sizeof(sa.sun_path))){
#ifdef DEBUG
    fprintf(stderr, "unix add: path too long %s\n", chase_source);
#endif
    return NULL;
  }

  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if(fd < 0){
    return NULL;
  }

  if(bind(fd, (struct sockaddr *) &sa, sizeof(sa))){
    close(fd);
    return NULL;
  }

  if(listen(fd, LO_UNIX_BACKLOG)){
    close(fd);
    return NULL;
  }

  ul = malloc(sizeof(struct unix_loop));
  if(ul == NULL){
    close(fd);
    return NULL;
  }

  if(rename(sa.sun_path, chase_source)){
#ifdef DEBUG
    fprintf(stderr, "add unix: rename to %s failed: %s\n", chase_source, strerror(errno));
#endif
    free(ul);
    close(fd);
    return NULL;
  }

  ul->u_mask = mask;
  ul->u_data = data;
  ul->u_setup = setup;
  ul->u_call = call;
  ul->u_type = type;

  if(tv){
    ul->u_timeout.tv_sec = tv->tv_sec;
    ul->u_timeout.tv_usec = tv->tv_usec;
  } else {
    ul->u_timeout.tv_sec = 1;
    ul->u_timeout.tv_usec = 0;
  }

  fcntl(fd, F_SETFD, FD_CLOEXEC);

  tl = add_loop(ls, fd, ul, LO_READ_MASK | LO_STOP_MASK, unix_listen_loop, NULL, 0);
  if(tl == NULL){
    free(ul);
    close(fd);
    return NULL;
  }

  text_loop(ls, tl, "unx listen: %s", name);

  return tl;
}
