/*
 * (c) Marc Welz 1999, 2004, 2005, 2006:
 *
 * Only to be distributed under the conditions of the GNU GPL v2 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "priv.h"
#include "loop.h"

/* support functions: api exported to user */

void text_loop(struct list_loop *ls, struct task_loop *tl, char *str, ...)
{
  va_list ap;

  va_start(ap, str);

  vsnprintf(tl->t_text, LO_TEXT_SIZE, str, ap);
  tl->t_text[LO_TEXT_SIZE - 1] = '\0';

  va_end(ap);
}

void *user_loop(struct list_loop *ls, struct task_loop *tl)
{
  return tl->t_data;
}

int fd_loop(struct list_loop *ls, struct task_loop *tl)
{
  return tl->t_fd;
}

int mask_loop(struct list_loop *ls, struct task_loop *tl)
{
  return tl->t_mask;
}

int tag_loop(struct list_loop *ls, struct task_loop *tl)
{
  return tl->t_tag;
}

int type_loop(struct list_loop *ls, struct task_loop *tl)
{
  return tl->t_type;
}

char *type_name_loop(struct list_loop *ls, struct task_loop *tl)
{
  int type;

  type = tl->t_type;

  if((type <= 0) || (type >= ls->l_job_size)){
    return NULL;
  }

  return ls->l_job[type].j_name;
}

void *global_loop(struct list_loop *ls)
{
  return ls->l_data;
}

struct timeval *now_loop(struct list_loop *ls)
{
  return &(ls->l_now);
}

int set_fd_loop(struct list_loop *ls, struct task_loop *tl, int fd)
{
  tl->t_fd = fd;
  return 0;
}

int set_mask_loop(struct list_loop *ls, struct task_loop *tl, int mask, int mode)
{
  if(mode < 0){
    tl->t_mask = (~mask) & (tl->t_mask);
  } else if(mode > 0){
    tl->t_mask = mask | (tl->t_mask);
  } else {
    tl->t_mask = mask;
  }
  return 0;
}

void set_max_loop(struct list_loop *ls, struct task_loop *tl, int queue, int max)
{
#ifdef DEBUG
  switch(queue){
    case LO_INQ :
    case LO_OUTQ :
      break;
    default :
      fprintf(stderr, "max_loop: unknown queue %d\n", queue);
      abort();
  }
#endif

  tl->t_max[queue] = max;
}

int set_call_loop(struct list_loop *ls, struct task_loop *tl, int (*call)(struct list_loop *ls, struct task_loop *tl, int mask))
{
  if(call){
    tl->t_call = call;
    return 0;
  } else {
    if(tl->t_mask == 0){
      return 0;
    } else {
#ifdef DEBUG
      fprintf(stderr, "call: warning: null call with nonzero mask\n");
#endif
      return -1;
    }
  }
}

int set_time_loop(struct list_loop *ls, struct task_loop *tl, struct timeval *tv)
{
  if(tv){
    tl->t_interval.tv_sec = tv->tv_sec;
    tl->t_interval.tv_usec = tv->tv_usec;
  }
#ifdef DEBUG
  fprintf(stderr, "loop: time interval %ld.%06ld\n", tl->t_interval.tv_sec, tl->t_interval.tv_usec);
#endif
  add_time_loop(&(tl->t_timeout), &(ls->l_now), &(tl->t_interval));

  return 0;
}

int set_type_loop(struct list_loop *ls, struct task_loop *tl, int type)
{
  tl->t_type = type;
  return 0;
}

int set_tag_loop(struct list_loop *ls, struct task_loop *tl, int tag)
{
  tl->t_tag = tag;
  return 0;
}

int set_user_loop(struct list_loop *ls, struct task_loop *tl, void *data)
{
  tl->t_data = data;
  return 0;
}

struct task_loop *find_task_loop(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type)
{
  return find_task_tag_loop(ls, tl, tx, type, 0);
}

struct task_loop *find_task_tag_loop(struct list_loop *ls, struct task_loop *tl, struct task_loop *tx, int type, int tag)
{
  struct task_loop *ty;

  if((type < 0) || (tag < 0)){
    return NULL;
  }

  ty = (tx) ? (tx->t_next) : (ls->l_head);

  while(ty && /* items there */
    ((type && (ty->t_type != type)) || /* mismatched type */
    (tag && (ty->t_tag != tag)) || /* mismatched tag */
    (tl == ty) || /* exclude self/special/search */
    (ty->t_stop != LO_STOP_NOT))){ /* shutting down */
    ty = ty->t_next;
  }

  return ty;
}

int lookup_link_type_loop(struct list_loop *ls, char *name)
{
  /* could have a static variable to cache lookup, at cost of making it thread unsafe */
  struct pool_loop *pl;
  int i;

  if(name == NULL){
    return -1;
  }

  for(i = 0; i < ls->l_pool_size; i++){
    pl = &(ls->l_pool[i]);
    if(pl->p_name && (strcmp(pl->p_name, name) == 0)){
      return pl->p_type;
    }
  }

  return -1;
}

int claim_link_name_loop(struct list_loop *ls, int lm, int lp, char *name, void (*empty)(struct list_loop *ls, void *data))
{
  struct pool_loop *tmp, *pl, *pz;
  char *copy;

  /* TODO: possibly check if name unique */

  if(name){
    copy = strdup(name);
    if(copy == NULL){
      return 0;
    }
  } else {
    copy = NULL;
  }

  tmp = realloc(ls->l_pool, sizeof(struct pool_loop) * (ls->l_pool_size + 1));
  if(tmp == NULL){
    if(copy){
      free(copy);
    }
    return 0; /* failure */
  }

  pl = &(tmp[ls->l_pool_size]);
  pz = &(tmp[0]);

  pl->p_type = ls->l_pool_size;
  pl->p_free = NULL;
  pl->p_have = 0;
  pl->p_used = 0;

  pl->p_empty = empty;
  pl->p_max = lm;
  pl->p_pool = lp; /* keep no more than lp unused around (can be scavenged to less) */

  pl->p_name = copy;

  if(lp < 0){ /* if local pool unlimited, up global pool to global maximum */
    pz->p_pool = pz->p_max;
  } else if(pz->p_pool >= 0) { /* if local pool bounded, increment global pool bounded by global maximum */
    pz->p_pool += lp;
    if((pz->p_max >= 0) && (pz->p_pool > pz->p_max)){
      pz->p_pool = pz->p_max;
    }
  }

#ifdef DEBUG
  fprintf(stderr, "quotas: pool: %d/%d, max=%d/%d\n", pl->p_pool, pz->p_pool, pl->p_max, pz->p_max);
#endif

  ls->l_pool = tmp;
  ls->l_pool_size++;

#ifdef DEBUG
  fprintf(stderr, "claim: grabbed link %d\n", pl->p_type);
#endif

  return pl->p_type;
}

int usage_link_loop(struct list_loop *ls, int type)
{
  if((type < 0) || (type >= ls->l_pool_size)){
    return -1;
  }

  return ls->l_pool[type].p_used;
}

int limit_link_loop(struct list_loop *ls, int type)
{
  if((type < 0) || (type >= ls->l_pool_size)){
    return -1;
  }

  return ls->l_pool[type].p_max;
}

int claim_link_loop(struct list_loop *ls, int lm, int lp, void (*empty)(struct list_loop *ls, void *data))
{
  return claim_link_name_loop(ls, lm, lp, NULL, empty);
}

int lookup_task_type_loop(struct list_loop *ls, char *name)
{
  /* could have a static variable to cache lookup, at cost of making it thread unsafe */
  struct job_loop *jl;
  int i;

  if(name == NULL){
    return -1;
  }

  for(i = 0; i < ls->l_job_size; i++){
    jl = &(ls->l_job[i]);
    if(jl->j_name && (strcmp(jl->j_name, name) == 0)){
      return jl->j_type;
    }
  }

  return -1;
}

int claim_task_name_loop(struct list_loop *ls, int tm, int tp, char *name)
{
  struct job_loop *tmp, *jl, *jz;
  char *copy;

  if(name){
    copy = strdup(name);
    if(copy == NULL){
      return 0;
    }
  } else {
    copy = NULL;
  }

  tmp = realloc(ls->l_job, sizeof(struct job_loop) * (ls->l_job_size + 1));
  if(tmp == NULL){
    if(copy){
      free(copy);
    }
    return 0; /* failure */
  }

  jl = &(tmp[ls->l_job_size]);
  jz = &(tmp[0]);

  jl->j_type = ls->l_job_size;
  jl->j_used = 0;
  jl->j_max = tm;

  jl->j_name = copy;

  ls->l_job = tmp;
  ls->l_job_size++;

#ifdef DEBUG
  fprintf(stderr, "claim: grabbed task %d\n", jl->j_type);
#endif

  return jl->j_type;
}

int claim_task_loop(struct list_loop *ls, int tm, int tp)
{
  return claim_task_name_loop(ls, tm, tp, NULL);
}

int usage_task_loop(struct list_loop *ls, int type)
{
  if((type < 0) || (type >= ls->l_job_size)){
    return -1;
  }

  return ls->l_job[type].j_used;
}

int limit_task_loop(struct list_loop *ls, int type)
{
  if((type < 0) || (type >= ls->l_job_size)){
    return -1;
  }

  return ls->l_job[type].j_max;
}

int claim_type_loop(struct list_loop *ls, int tm, int tp, int lm, int lp, void (*empty)(struct list_loop *ls, void *data))
{
  int jx, px;

#ifdef DEBUG
  fprintf(stderr, "warning: claim type not desirable\n");
#endif

  jx = claim_task_loop(ls, tm, tp);
  px = claim_link_loop(ls, lm, lp, empty);

#ifdef DEBUG
  if(jx != px){
    fprintf(stderr, "claim: mixing old and new claim api\n");
    abort();
  }
#endif

  return px;
}

void default_empty_loop(struct list_loop *ls, void *data)
{
  free(data);
}

void *alloc_link_loop(struct list_loop *ls, struct link_loop *nk, int type, int size)
{
  if(nk->k_type == 0){
    nk->k_type = type;
#ifdef DEBUG
  } else {
    if(nk->k_type != type){
      fprintf(stderr, "alloc: logic problem - type mismatch\n");
      abort();
    }
#endif
  }

  if(nk->k_data == NULL){
    nk->k_data = malloc(size);
  }

  return nk->k_data;
}
