/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* modal switch: redirect stream to multiple modes */

struct mxk_mode_ops{
  char *o_light;
  int o_mode;
  int o_flags;

  struct mxk_client *o_vector;
  int o_count;
};

struct mxk_mode{
  int m_mode;
  int m_track;
  int m_defer;

  int m_size;
  unsigned char *m_map; /* zero is ignored */

  int m_top;
  struct mxk_mode_ops *m_ops;
};

#define MXK_MODE_NULL    0x0
#define MXK_MODE_PASS    0x1  /* let the char through */
#define MXK_MODE_CHANGE  0x2  /* change the current mode */
#define MXK_MODE_DEFER   0x4  /* look at global mode to see if passes */

#define MXK_MODE_CYCLE   0x10 /* cycle through modes */
#define MXK_MODE_SKIP    0x20 /* skip over given entry when cycling */

static struct code_name mxk_mode_codes[] = {
  { "ignore",     MXK_MODE_NULL },
  { "pass",       MXK_MODE_PASS },
  { "change",     MXK_MODE_CHANGE },
  { "defer",      MXK_MODE_DEFER },
  { "cycle",      MXK_MODE_CYCLE },
  { "skip",       MXK_MODE_SKIP },

  { NULL, -1 }
};

static void mxk_stop_mode(struct mxk_mode *mm);

static int mxk_insert_mode(void *mode, int type, int id);
static int mxk_run_mode(struct list_loop *ls, struct task_loop *tl, int mask);

static int mxk_setup_mode(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_state_mode(struct list_loop *ls, struct task_loop *tx, int special, char *args);
static int mxk_key_mode(struct list_loop *ls, struct task_loop *tx, int special, char *args);

/**************************************/

int mxk_start_mode_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type;
  struct control_loop *cl;
  char *value;
  struct task_loop *tx;
  struct mxk_mode *mm;
  struct mxk_mode_ops *mo;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "mode");
  if(type <= 0){
    print_control(cl, cd, 1, "mode: unable to lookup type for mode\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
#ifdef DEBUG
      fprintf(stderr, "mode: key is %d,%d\n", type, id);
#endif
    }
  }

  mm = malloc(sizeof(struct mxk_mode));
  if(mm == NULL){
    print_control(cl, cd, 1, "mode: unable to allocate mode structure\n");
    return 0;
  }

  mm->m_mode = 0;

  mm->m_map = NULL;
  mm->m_size = 0;

  mm->m_track = 0;
  mm->m_defer = 0;

  mm->m_top = 0;
  mm->m_ops = NULL;

  mo = malloc(sizeof(struct mxk_mode_ops));
  if(mo == NULL){
    free(mm);
    return 0;
  }

  mo->o_light = NULL;
  mo->o_mode = 0;
  mo->o_flags = 0;
  mo->o_vector = NULL;
  mo->o_count = 0;

  mm->m_ops = mo;
  mm->m_top = 1;

  tx = add_tag_loop(ls, -1, mm, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_mode, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "mode: unable to create mode task\n");
    mxk_stop_mode(mm);
    return 0;
  }

  text_loop(ls, tx, "mode:%d", id);
  print_control(cl, cd, 1, "mode: created id %d\n", id);

  return 0;
}

/**************************************/

int mxk_setup_mode_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "mode", MXK_MISC_REQUIRED, 0, mxk_setup_mode);

  return 0;
}

int mxk_state_mode_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "mode", MXK_MISC_REQUIRED, 0, mxk_state_mode);

  return 0;
}

int mxk_key_mode_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "mode", MXK_MISC_REQUIRED, 0, mxk_key_mode);

  return 0;
}

int mxk_list_mode_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, tag, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_mode *mm;
  struct mxk_mode_ops *mo;
  char *value, *key, *name;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "mode");
  if(type <= 0){
    print_control(cl, cd, 1, "mode: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
    }
  }

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    count = 0;
    mm = user_loop(ls, tx);
    if(mm){
      tag = tag_loop(ls, tx);
      print_control(cl, cd, 0, "mode[%d]: current state=%d, total states=%d\n", tag, mm->m_mode, mm->m_top);
      for(i = 0; i < mm->m_top; i++){
        mo = &(mm->m_ops[i]);
        name = code_to_name(mxk_mode_codes, mo->o_flags);
        print_control(cl, cd, 0, "state %d: light pattern=%s, next state=%d, relay key=%s, links=%d\n", i, mo->o_light ? mo->o_light : "<not set>", mo->o_mode, name, mo->o_count);
      }
      for(i = 1; i < mm->m_size; i++){
        if((mm->m_map[i] > 0) && (mm->m_map[i] < mm->m_top)){
          key = code_to_name(mxk_key_codes, i);
          if(key){
            print_control(cl, cd, 0, "key %s:", key);
          } else {
            print_control(cl, cd, 0, "key 0x%x:", i);
          }
          mo = &(mm->m_ops[mm->m_map[i]]);
          name = code_to_name(mxk_mode_codes, mo->o_flags);
          print_control(cl, cd, 0, "light pattern=%s, next state=%d, relay key=%s\n", mo->o_light ? mo->o_light : "<not set>", mo->o_mode, name);
          count++;
        }
      }
      print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "key" : "keys");
    } else {
      print_control(cl, cd, 1, "mode[%d]: unavailable\n", id);
    }
  }

  return 0;
}

/***************************************************************/

static int mxk_setup_mode(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_mode *mm;
  struct mxk_mode_ops *mo;
  char buffer[MXK_NAME_SIZE];
  char *value;
  int idx, mode, i, flags, run;

  mm = user_loop(ls, tx);
  if(mm == NULL){
    return -1;
  }

  if(args == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  value = args + idx;
  mode = strtoul(value, NULL, 10);

  if((mode < 0) || (mode >= 256)){
#ifdef DEBUG
    fprintf(stderr, "mode: out of range (no mode than 2^8 modes)\n");
#endif
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr, "mode: updating mode %d\n", mode);
#endif

  if(mode >= mm->m_top){
#ifdef DEBUG
    fprintf(stderr, "mode: need to resize (%d->%d)\n", mm->m_top, mode + 1);
#endif
    mo = realloc(mm->m_ops, sizeof(struct mxk_mode_ops) * (mode + 1));
    if(mo == NULL){
      return -1;
    }

    mm->m_ops = mo;
    for(i = mm->m_top; i <= mode; i++){
      mo = &(mm->m_ops[i]);

      mo->o_light = NULL;
      mo->o_flags = MXK_MODE_NULL;
      mo->o_mode = i;

      mo->o_vector = NULL;
      mo->o_count = 0;
    }

    mm->m_top = mode + 1;
  }

  /* clear everything */
  mo = &(mm->m_ops[mode]);
  if(mo->o_light){
    free(mo->o_light);
    mo->o_light = NULL;
  }
  mo->o_mode = mode;
  mo->o_flags = MXK_MODE_NULL;

  for(run = 1; run;){
    idx = next_word(args, idx);
    if(idx < 0){
      return 0;
    }
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }
    flags = name_to_code(mxk_mode_codes, buffer);
    if(flags >= 0){
      mo->o_flags |= flags;
    } else {
      if(isdigit(buffer[0])){
        run = 0;
      } else {
#ifdef DEBUG
        fprintf(stderr, "mode: unknown relay type %s\n", buffer);
#endif
        return -1;
      }
    }
  }

  /* next mode, if there is one */
  value = args + idx;
  mo->o_mode = strtoul(value, NULL, 10);
  if(mo->o_mode < 0){
    mo->o_mode = 0;
    return -1;
  }

  /* light value */
  idx = next_word(args, idx);
  if(idx < 0){
    return 0;
  }
  mo->o_light = copy_word(args + idx, 0);
  if(mo->o_light == NULL){
    return -1;
  }
#ifdef DEBUG
  fprintf(stderr, "mode: setting light to <%s>\n", mo->o_light);
#endif

  return 0;
}

static int mxk_state_mode(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_mode *mm;
  struct mxk_mode_ops *mo;
  char *value, *light;
  int idx, mode;

  mm = user_loop(ls, tx);
  if(mm == NULL){
    return -1;
  }

  if(args == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  value = args + idx;
  mode = strtoul(value, NULL, 10);

  if((mode < 0) || (mode >= mm->m_top)){
#ifdef DEBUG
    fprintf(stderr, "mode: state out of range\n");
#endif
    return -1;
  }

  if(mm->m_track){
    mm->m_defer = mode;

  } else {
    if(mm->m_mode != mode){

      mo = &(mm->m_ops[mm->m_mode]);
      light = mo->o_light;

      mm->m_mode = mode;
      mo = &(mm->m_ops[mm->m_mode]);
      mm->m_defer = mm->m_mode;

      if((light == NULL) || (mo->o_light == NULL) || strcmp(light, mo->o_light)){
        if(light){
          mxk_stop_light(ls, light);
        }
        if(mo->o_light){
          mxk_start_light(ls, mo->o_light);
        }
      }
    }
  }

  return 0;
}

static int mxk_key_mode(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_mode *mm;
  char buffer[MXK_NAME_SIZE];
  int key, mode, idx, i;
  char *value;
  unsigned char *tmp;

  idx = start_word(args, 0);
  mm = user_loop(ls, tx);

#ifdef DEBUG
  fprintf(stderr, "mode: updating key\n");
#endif

  if(idx < 0){
    return -1;
  }

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    return -1;
  }

  key = name_to_code(mxk_key_codes, buffer);
  if(key <= 0){
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr, "key mode: key is %d(%s), size is %d\n", key, buffer, mm->m_size);
#endif

  if(key >= mm->m_size){
    tmp = realloc(mm->m_map, sizeof(unsigned char) * (key + 1));
    if(tmp == NULL){
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "key mode: resizing, marking %d to %d inclusive\n", mm->m_size, key);
#endif

    for(i = mm->m_size; i <= key; i++){
      tmp[i] = 0;
    }

    mm->m_map = tmp;
    mm->m_size = key + 1;
  }


  idx = next_word(args, idx);
  if(idx < 0){
    mode = 0;
  } else {
    value = args + idx;
    mode = strtoul(value, NULL, 10);

    if((mode < 0) || (mode >= mm->m_top)){
#ifdef DEBUG
      fprintf(stderr, "mode: state out of range\n");
#endif
      return -1;
    }
  }

  mm->m_map[key] = mode;

#ifdef DEBUG
  fprintf(stderr, "key mode: set map[%d] to mode %d\n", key, mode);
#endif

  return 0;
}

/***************************************************************/

static void mxk_stop_mode(struct mxk_mode *mm)
{
  int i;
  struct mxk_mode_ops *mo;

  if(mm){
    if(mm->m_map){
      free(mm->m_map);
      mm->m_map = NULL;
    }
    mm->m_size = 0;

    for(i = 0; i < mm->m_top; i++){
      mo = &(mm->m_ops[i]);

      if(mo->o_light){
        free(mo->o_light);
        mo->o_light = NULL;
      }

      mxk_shutdown_client(mo->o_vector, mo->o_count);
      mo->o_vector = NULL;
      mo->o_count = 0;
    }

    if(mm->m_ops){
      free(mm->m_ops);
      mm->m_ops = NULL;
    }

    mm->m_top = 0;

    free(mm);
  }
}

static int mxk_run_mode(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event *vp;
  struct mxk_mode *mm;
  struct mxk_mode_ops *mo, *to;
  struct mxk_global *mg;
  struct link_loop *nk;
  char *light;
  int index, run;

  mm = user_loop(ls, tl);
  mg = global_loop(ls);

#ifdef DEBUG
  if((mm->m_mode < 0) || (mm->m_mode >= mm->m_top)){
    fprintf(stderr, "mode: logic problem: mode=%d out of range\n", mm->m_mode);
    abort();
  }
#endif

  mo = &(mm->m_ops[mm->m_mode]);

  if(mask & LO_STOP_MASK){
    if(mo->o_light){
      mxk_stop_light(ls, mo->o_light);
    }
    mxk_stop_mode(mm);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_KEY :

            index = 0;
            if(mm->m_size > vp->code){
              index = mm->m_map[vp->code];
              if((index < 0) || (index >= mm->m_top)){
                /* prefer not to disallow out of range entries at setup, as this makes constructing circular modes awkward */
#ifdef DEBUG
                fprintf(stderr, "mode: warning: changing unknown mode %d to 0\n", mm->m_defer);
#endif
                index = 0;
              }
            }

            to = &(mm->m_ops[index]);
            if(vp->value == 1){ /* press key only */
              if(to->o_flags & MXK_MODE_CHANGE){
#ifdef DEBUG
                fprintf(stderr, "mode: 0x%x -> index=%d: change to new mode %d\n", vp->code, index, to->o_mode);
#endif
                mm->m_defer = to->o_mode;
              }
              if(to->o_flags & MXK_MODE_CYCLE){
                mm->m_defer = mm->m_mode;
                for(run = 1; run;){
                  mm->m_defer++;
                  if(mm->m_defer >= mm->m_top){
                    mm->m_defer = 0;
                  }
                  if(mm->m_defer == mm->m_mode){
                    run = 0;
                  }
                  if(!(mm->m_ops[mm->m_defer].o_flags & MXK_MODE_SKIP)){
                    run = 0;
                  }
                }
#ifdef DEBUG
                fprintf(stderr, "mode: 0x%x -> index=%d: cycle to mode %d\n", vp->code, index, mm->m_defer);
#endif
              }
            }

            /* keep track of depressed keys */
            switch(vp->value){
              case 0 :
                mm->m_track--;
                if(mm->m_track < 0){
#ifdef DEBUG
                  fprintf(stderr, "mode: warning: track out of sync\n");
#endif
                  mm->m_track = 0;
                  mm->m_defer = mm->m_mode;
                }
                break;
              case 1 :
                mm->m_track++;
                break;
            }

#ifdef DEBUG
            fprintf(stderr, "mode[%d]: current=%d, next=%d, local=%d(0x%x): code=0x%02x\n", tag_loop(ls, tl), mm->m_mode, mm->m_defer, to->o_mode, to->o_flags, vp->code);
#endif

            /* make real state desired ... */
            if(mm->m_mode != mm->m_defer){
              if(mm->m_track == 1){ /* ... but only if all depressed keys are moved out */
#ifdef DEBUG
                fprintf(stderr, "mode: changing mode %d->%d\n", mm->m_mode, mm->m_defer);
#endif

                light = mo->o_light;

                mm->m_mode = mm->m_defer;
                mo = &(mm->m_ops[mm->m_mode]);

                if((light == NULL) || (mo->o_light == NULL) || strcmp(light, mo->o_light)){
                  if(light){
                    mxk_stop_light(ls, light);
                  }
                  if(mo->o_light){
                    mxk_start_light(ls, mo->o_light);
                  }
                }
#ifdef DEBUG
              } else {
                fprintf(stderr, "mode: deferring update: track=%d\n", mm->m_track);
#endif
              }
            }

            /* change from local to global if needed */
            if(to->o_flags & MXK_MODE_DEFER){
              to = &(mm->m_ops[mm->m_mode]);
            }
            if(to->o_flags & MXK_MODE_PASS){
              /* xmit event */
#ifdef DEBUG
              fprintf(stderr, "mode[%d]: forwarding 0x%x(%s)\n", tag_loop(ls, tl), vp->code, code_to_name(mxk_key_codes, vp->code));
#endif
              mxk_send_client(ls, tl, mo->o_vector, mo->o_count, vp->type, vp->code, vp->value);
            }

          break;

          default :
          /* TODO ? */
#ifdef DEBUG
            fprintf(stderr, "mode: ignoring event type %d\n", vp->type);
#endif
          break;
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_mode(void *mode, int type, int id)
{
  struct mxk_mode *mm;
  struct mxk_mode_ops *mo;
  struct mxk_client *mc;

  mm = mode;

#ifdef DEBUG
  if((mm->m_mode < 0) || (mm->m_mode >= mm->m_top)){
    fprintf(stderr, "mode: logic failure: inserting into nonexistant mode\n");
    return -1;
  }
#endif

  mo = &(mm->m_ops[mm->m_mode]);

  mc = mxk_insert_client(mo->o_vector, mo->o_count, type, id);
  if(mc == NULL){
    return -1;
  }

  mo->o_vector = mc;
  mo->o_count = mo->o_count + 1;

  return 0;
}

int mxk_startup_mode(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "mode", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_mode);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_mode(struct mxk_global *mg)
{
}
