/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <linux/kdev_t.h>
#include <linux/input.h>

#include <loop.h>
#include <code-name.h>


#include "mxk.h"

#ifdef DEBUG
static void mxk_dump_queue(struct mxk_queue *mq, FILE *fp, char *prefix);
#endif

static int mxk_find_queue(struct mxk_queue *mq, int code, int id, int index);
static void mxk_shiftup_queue(struct mxk_queue *mq, int pos, int index);
static void mxk_insert_queue(struct mxk_queue *mq, int code, int id);

struct mxk_queue *mxk_start_queue(int size)
{
  struct mxk_queue *mq;
  int i;

  mq = malloc(sizeof(struct mxk_queue));
  if(mq == NULL){
    return NULL;
  }

  mq->q_queue = NULL;
  for(i = 0; i < 4; i++){
    mq->q_index[i] = 0;
  }
  mq->q_index[MXK_QUEUE_TOTAL_SIZE] = size;

  mq->q_fancy = 0;
  mq->q_repeat = 0;

  mq->q_queue = malloc(sizeof(struct mxk_qitem) * size);
  if(mq->q_queue == NULL){
    mxk_stop_queue(mq);
    return NULL;
  }

  for(i = 0; i < size; i++){
    mq->q_queue[i].i_code = 0;
    mq->q_queue[i].i_id = 0;
  }

  return mq;
}

void mxk_stop_queue(struct mxk_queue *mq)
{
  int i;

  if(mq){
    if(mq->q_queue){
      free(mq->q_queue);
      mq->q_queue = NULL;
    }

    for(i = 0; i < 4; i++){
      mq->q_index[i] = 0;
    }

    free(mq);
  }
}

/***********************************************************************/

void mxk_update_queue(struct list_loop *ls, struct mxk_queue *mq, int code, int value)
{
  int pos;
#ifdef DEBUG
  int i, p;

  p = 0;
  for(i = 0; i < 4; i++){
    if(p > mq->q_index[i]){
      fprintf(stderr, "queue: logic problem: index not ascending (%d)\n", i);
      abort();
    }
    p = mq->q_index[i];
  }
#endif

  pos = mxk_find_queue(mq, code, 0, MXK_QUEUE_NEW_RELEASE);
  if(value){ /* press */
    if(pos < 0){ /* not found */
      mxk_insert_queue(mq, code, 0);
    }
  } else { /* release */
    if(pos >= mq->q_index[MXK_QUEUE_OLD_PRESS]){
      mxk_shiftup_queue(mq, pos, MXK_QUEUE_OLD_RELEASE);
    } else if(pos >= 0){
      mxk_shiftup_queue(mq, pos, MXK_QUEUE_NEW_RELEASE);
    } /* else not found */
  }

#ifdef DEBUG
  mxk_dump_queue(mq, stderr, "postupdate");
#endif
}

void mxk_consume_queue(struct mxk_queue *mq, int length)
{
  struct mxk_qitem *mi, *xi;
  int d;

  if(length <= 0){
    return;
  }

  if(length >= mq->q_index[MXK_QUEUE_OLD_PRESS]){
#ifdef DEBUG
    fprintf(stderr, "consume: consume=%d, range=(%d,%d,%d)\n", length, mq->q_index[MXK_QUEUE_OLD_PRESS], mq->q_index[MXK_QUEUE_NEW_RELEASE], mq->q_index[MXK_QUEUE_OLD_RELEASE]);
#endif
    mq->q_index[MXK_QUEUE_OLD_PRESS] = 0; /* all pressed keys are old */
    if(length > mq->q_index[MXK_QUEUE_NEW_RELEASE]){
      mq->q_index[MXK_QUEUE_OLD_RELEASE] = mq->q_index[MXK_QUEUE_NEW_RELEASE]; /* all released keys are old */
    }
#ifdef DEBUG
    mxk_dump_queue(mq, stderr, "large");
#endif
    return;
  }

#if 0
  if((length >= mq->q_index[MXK_QUEUE_OLD_PRESS]) && (length <= mq->q_index[MXK_QUEUE_NEW_RELEASE])){
#ifdef DEBUG
    fprintf(stderr, "consume: consumed %d: within [%d]%d,[%d]%d: marking all presses old\n", length, MXK_QUEUE_OLD_PRESS, mq->q_index[MXK_QUEUE_OLD_PRESS], MXK_QUEUE_NEW_RELEASE, mq->q_index[MXK_QUEUE_NEW_RELEASE]);
#endif
    mq->q_index[MXK_QUEUE_OLD_PRESS] = 0;
#ifdef DEBUG
    mxk_dump_queue(mq, stderr, "allpress");
#endif
    return;
  }

  if(length > mq->q_index[MXK_QUEUE_NEW_RELEASE]){
#ifdef DEBUG
    fprintf(stderr, "consume: marked all old: released: total=%d, discarded=%d\n", mq->q_index[MXK_QUEUE_OLD_RELEASE] - mq->q_index[MXK_QUEUE_NEW_RELEASE], mq->q_index[MXK_QUEUE_OLD_RELEASE] - length);
#endif
    mq->q_index[MXK_QUEUE_OLD_PRESS] = 0;
    mq->q_index[MXK_QUEUE_OLD_RELEASE] = mq->q_index[MXK_QUEUE_NEW_RELEASE];
#ifdef DEBUG
    mxk_dump_queue(mq, stderr, "everything");
#endif
    return;
  }
#endif

  /* could be replaced with "struct mxk_qitem xq[mq->q_index[MXK_QUEUE_OLD_PRESS]];" */
  d = mq->q_index[MXK_QUEUE_OLD_PRESS] - length;
  mi = mq->q_queue;
  xi = alloca(d * sizeof(struct mxk_qitem));
  if(xi == NULL){
    return;
  }

#ifdef DEBUG
  fprintf(stderr, "consume: %d pressed entries still left, moving from %d\n", d, length);
#endif

  memcpy(xi, mi + length, d * sizeof(struct mxk_qitem));
  memmove(mi + d, mi, length * sizeof(struct mxk_qitem));
  memcpy(mi, xi, d * sizeof(struct mxk_qitem));

  mq->q_index[MXK_QUEUE_OLD_PRESS] = d;

#ifdef DEBUG
  mxk_dump_queue(mq, stderr, "partial");
#endif
}

/***********************************************************************/

static int mxk_find_queue(struct mxk_queue *mq, int code, int id, int index)
{
  int x, i;

#ifdef DEBUG
  if((index < 0) || (index >= MXK_QUEUE_TOTAL_SIZE)){
    fprintf(stderr, "queue: problem: index out of range\n");
    abort();
  }
#endif

  x = mq->q_index[index];

  for(i = 0; i < x; i++){
    /* TODO: handle id too */
    if(mq->q_queue[i].i_code == code){
      return i;
    }
  }

  return -1;
}

static void mxk_shiftup_queue(struct mxk_queue *mq, int pos, int index)
{
  int x, s, t, w, i;
  int code, id;

#ifdef DEBUG
  if((index < 0) || (index >= MXK_QUEUE_TOTAL_SIZE)){
    fprintf(stderr, "queue: problem: index out of range\n");
    abort();
  }
#endif

  x = mq->q_index[index];
#ifdef DEBUG
  if(pos > x){
    fprintf(stderr, "queue: problem: attempting to move target %d back to %d\n", pos, x);
    abort();
  }
#endif

  if((pos + 1) < x){
    code = mq->q_queue[pos].i_code;
    id = mq->q_queue[pos].i_id;

    s = pos + 1;
    t = pos;
    w = x - s;

    memmove(mq->q_queue + t, mq->q_queue + s, w * sizeof(struct mxk_qitem));
    mq->q_queue[x - 1].i_code = code;
    mq->q_queue[x - 1].i_id = id;
  }

  for(i = 0; i <= index; i++){
    if(mq->q_index[i] > pos){
      mq->q_index[i] = mq->q_index[i] - 1;
    }
  }
}

static void mxk_insert_queue(struct mxk_queue *mq, int code, int id)
{
  int i, w;

  w = mq->q_index[MXK_QUEUE_TOTAL_SIZE] - 1;

  memmove(mq->q_queue + 1, mq->q_queue, w * sizeof(struct mxk_qitem));

  mq->q_queue[0].i_code = code;
  mq->q_queue[0].i_id = id;

  for(i = 0; i < MXK_QUEUE_TOTAL_SIZE; i++){
    if(mq->q_index[i] < mq->q_index[MXK_QUEUE_TOTAL_SIZE]){
      mq->q_index[i] = mq->q_index[i] + 1;
    }
  }

#ifdef DEBUG
  mxk_dump_queue(mq, stderr, "postinsert");
#endif

}

#ifdef DEBUG
static void mxk_dump_queue(struct mxk_queue *mq, FILE *fp, char *prefix)
{
  int i, j, id;

  fprintf(fp, "%s dump[%d]: ", prefix, mq->q_index[MXK_QUEUE_TOTAL_SIZE]);

  j = 0;
  for(i = 0; i < mq->q_index[MXK_QUEUE_TOTAL_SIZE]; i++){
    while((j < MXK_QUEUE_TOTAL_SIZE) && (mq->q_index[j] <= i)){
      fprintf(fp, " [%d]", j);
      j++;
    }
    id = mq->q_queue[i].i_id;
    fprintf(fp, " %d", mq->q_queue[i].i_code);
    if(id){
      fprintf(fp, "/%d", id);
    }
  }

  fprintf(fp, "\n");
}
#endif

