/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* reduce module: map input keypresses to single output keypress */

struct mxk_reduce{
  unsigned char *r_map;
  unsigned int r_size;

  int r_invert;
  int r_self;
  int r_merge;

  unsigned int r_replace;
  unsigned int r_press;

  int r_sync;

  struct mxk_client *r_vector;
  int r_count;
};

static void mxk_destroy_reduce(struct mxk_reduce *reduce);
static struct mxk_reduce *mxk_create_reduce();
static int mxk_run_reduce(struct list_loop *ls, struct task_loop *tl, int mask);

/* have function to add individual key to mask */
/* have function to add key ranges (button, virtual, normal) to mask */

/**********************************************************************/

struct mxk_reduce *mxk_create_reduce()
{
  struct mxk_reduce *reduce;

  /* allocate state for reduce node, used by start command */

  reduce = malloc(sizeof(struct mxk_reduce));
  if(reduce == NULL){
    return NULL;
  }

  reduce->r_vector = NULL;
  reduce->r_count = 0;

  reduce->r_invert = 0;
  reduce->r_self = 1;
  reduce->r_merge = 0;

  reduce->r_replace = 0;
  reduce->r_press = 0;

  reduce->r_map = NULL;
  reduce->r_size = 0;

  return reduce;
}

static void mxk_destroy_reduce(struct mxk_reduce *reduce)
{
  if(reduce == NULL){
    return;
  }

  reduce->r_size = 0;
  if(reduce->r_map){
    free(reduce->r_map);
    reduce->r_map = NULL;
  }

  mxk_shutdown_client(reduce->r_vector, reduce->r_count);
#ifdef DEBUG
  reduce->r_vector = NULL;
  reduce->r_count = 0;
#endif

  free(reduce);
}

/**********************************************************************/

int mxk_invert_reduce(struct list_loop *ls, struct task_loop *tx, int invert, char *args)
{
  struct mxk_reduce *reduce;

  reduce = user_loop(ls, tx);

  reduce->r_invert = invert;

  return 0;
}

int mxk_emit_reduce(struct list_loop *ls, struct task_loop *tx, int self, char *args)
{
  struct mxk_reduce *reduce;
  char buffer[MXK_NAME_SIZE];
  int idx, key;

  reduce = user_loop(ls, tx);

  if(self == 0){
    idx = start_word(args, 0);
    if(idx < 0){
#ifdef DEBUG
      fprintf(stderr, "reduce: need key to substitute\n");
#endif
      return -1;
    }

    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key < 0){
#ifdef DEBUG
      fprintf(stderr, "reduce: key not known to substitution\n");
#endif
      return -1;
    }

    reduce->r_replace = key;
  }

  reduce->r_self = self;

  return 0;
}

/**********************************************************************/

int mxk_merge_reduce(struct list_loop *ls, struct task_loop *tx, int merge, char *args)
{
  struct mxk_reduce *reduce;

  reduce = user_loop(ls, tx);

  reduce->r_merge = merge;

  return 0;
}

int mxk_add_reduce(struct mxk_reduce *reduce, int key)
{
  unsigned char *tmp;
  int i;

  /* TODO: map could be a bitmap, rather than byte */

  if(reduce->r_size <= key){
    tmp = realloc(reduce->r_map, sizeof(unsigned char) * (key + 1));
    if(tmp == NULL){
      return -1;
    }

#ifdef DEBUG
    fprintf(stderr, "reduce: padding range [%d:%d], value=%d\n", reduce->r_size, key - 1, reduce->r_invert);
#endif

    for(i = reduce->r_size; i < key; i++){
      tmp[i] = reduce->r_invert;
    }

    reduce->r_map = tmp;
    reduce->r_size = key + 1;
  }

  reduce->r_map[key] = 1 - reduce->r_invert; /* in invert mode clear the requested entries, otherwise set them */

  return 0;
}

int mxk_load_reduce(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  int idx, key;
  char buffer[MXK_NAME_SIZE];
  struct mxk_reduce *reduce;

  reduce = user_loop(ls, tx);
  if(reduce == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
#ifdef DEBUG
    fprintf(stderr, "array: need key for load\n");
#endif
    return -1;
  }

  while(idx >= 0){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      return -1;
    }

    key = name_to_code(mxk_key_codes, buffer);
    if(key < 0){
      return -1;
    }

    if(mxk_add_reduce(reduce, key)){
      return -1;
    }

    idx = next_word(args, idx);
  }

  return 0;
}

/**********************************************************************/

int mxk_discard_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_OPTIONAL, 1, mxk_invert_reduce);

  return 0;
}

int mxk_relay_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_OPTIONAL, 0, mxk_invert_reduce);

  return 0;
}

int mxk_self_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_OPTIONAL, 1, mxk_emit_reduce);

  return 0;
}

int mxk_replace_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_REQUIRED, 0, mxk_emit_reduce);

  return 0;
}

int mxk_merge_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_OPTIONAL, 1, mxk_merge_reduce);

  return 0;
}

int mxk_distinct_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_OPTIONAL, 0, mxk_merge_reduce);

  return 0;
}

int mxk_key_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "reduce", MXK_MISC_REQUIRED, 0, mxk_load_reduce);

  return 0;
}

/**********************************************************************/

int mxk_start_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_reduce *reduce;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type;
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "reduce");
  if(type <= 0){
    print_control(cl, cd, 1, "reduce: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "reduce: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "reduce: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
#if 0
  if(*end != ' '){
    print_control(cl, cd, 1, "reduce: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }
#endif

  reduce = mxk_create_reduce();
  if(reduce == NULL){
    print_control(cl, cd, 1, "reduce: unable to allocate state\n");
    return 0;
  }

  /* allocated state, now register run function to be called when event arrives */

  tx = add_tag_loop(ls, -1, reduce, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_reduce, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "reduce: unable to create task\n");
    mxk_destroy_reduce(reduce);
    return 0;
  }

  text_loop(ls, tx, "reduce:%d", id);
  print_control(cl, cd, 1, "reduce: created task\n");

  return 0;
}

int mxk_list_reduce_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, count, i, limit;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_reduce *reduce;
  char *ptr;

  /* list the nodes of this type */

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "reduce");
  if(type <= 0){
    print_control(cl, cd, 1, "reduce: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    reduce = user_loop(ls, tx);

    if(reduce){
      if(reduce->r_self){
        ptr = "<no>";
      } else {
        ptr = code_to_name(mxk_key_codes, reduce->r_replace);
        if(ptr == NULL){
          ptr = "<unknown>";
        }
      }
      print_control(cl, cd, 0, "reduce[%d]: match=%s, chords=%s, replacement=%s\n", tag_loop(ls, tx), reduce->r_invert ? "discard" : "relay", reduce->r_merge ? "merge" : "distinct", ptr);
      /* WARNING: need to include virtual keys too */
      limit = (reduce->r_invert) ? KEY_MAX : reduce->r_size;
      for(i = 0; i < limit; i++){
        if((i >= reduce->r_size) || (reduce->r_map[i])){
          ptr = code_to_name(mxk_key_codes, i);
          if(ptr){
            print_control(cl, cd, 0, "reduce[%d]: key %d: %s\n", tag_loop(ls, tx), i, ptr);
#if 0 /* tends to print unallocated codes */
          } else {
            print_control(cl, cd, 0, "reduce[%d]: key 0x%x\n", tag_loop(ls, tx), i);
#endif
          }
        }
      }
      count++;
    } else {
      print_control(cl, cd, 1, "reduce: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

/**********************************************************************/

static int mxk_run_reduce(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_reduce *reduce;

  struct input_event *vp;
  struct link_loop *nk;

  int found;

  /* called when an event arrives on the input of this node */

  mg = global_loop(ls);
  reduce = user_loop(ls, tl);

  if(reduce == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_reduce(reduce);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;

        switch(vp->type){
          case EV_KEY :
#ifdef DEBUG
          fprintf(stderr, "reduce[%d]: key=%d, size=%d, invert=%d\n", tag_loop(ls, tl), vp->code, reduce->r_size, reduce->r_invert);
#endif
            if(vp->code < reduce->r_size){
              found = reduce->r_map[vp->code] ? 1 : 0;
            } else {
              found = reduce->r_invert; /* if we are in invert mode, unspecified entries should be forwarded */
            }

            if(found){
#ifdef DEBUG
              fprintf(stderr, "reduce: found event, sending 0x%x\n", reduce->r_self ? vp->code : reduce->r_replace);
#endif
              if(reduce->r_self){
                mxk_send_client(ls, tl, reduce->r_vector, reduce->r_count, vp->type, vp->code, vp->value);
              } else {
                switch(vp->value){
                  case 0 :
                  case 1 :

                    if(reduce->r_press == vp->value){
                      if(reduce->r_merge){
                        break; /* WARNING: partial and indented switch finish */
                      }
                      mxk_send_client(ls, tl, reduce->r_vector, reduce->r_count, EV_KEY, reduce->r_replace, 1 - vp->value);
                      mxk_send_client(ls, tl, reduce->r_vector, reduce->r_count, EV_SYN, 0, 0);
                    }

                    mxk_send_client(ls, tl, reduce->r_vector, reduce->r_count, EV_KEY, reduce->r_replace, vp->value);
                    reduce->r_press = vp->value;
                    reduce->r_sync = 0;
                    break;
                }
              }
            }
            break;
          case EV_SYN :
            if(reduce->r_sync == 0){
              mxk_send_client(ls, tl, reduce->r_vector, reduce->r_count, EV_SYN, 0, 0);
              reduce->r_sync = 1;
            }
            break;
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

static int mxk_insert_reduce(void *state, int type, int id)
{
  struct mxk_reduce *reduce;
  struct mxk_client *mc;

  /* used to connect the output of this node to others */
  /* vector+count are used to point at other nodes */

  reduce = state;

  mc = mxk_insert_client(reduce->r_vector, reduce->r_count, type, id);
  if(mc == NULL){
    return -1;
  }

  reduce->r_vector = mc;
  reduce->r_count++;

  return 0;
}

int mxk_startup_reduce(struct list_loop *ls)
{
  int type;

  /* register type with rest of system */

  type = mxk_enroll_find(ls, "reduce", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_reduce);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_reduce(struct mxk_global *mg)
{
  /* could be useful if dynamic unloading ever gets implemented */
}
