/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* repeat pressed keys */

struct mxk_repeat{
  struct timeval r_rate;
  struct timeval r_delay;

  unsigned int r_base;
  unsigned int r_faster;
  unsigned int r_slower;
  int r_delta;

  int r_pass;
  int r_down;
  int r_working;
  unsigned int r_last;

  struct mxk_client *r_vector;
  int r_count;
};

static int mxk_start_repeat(struct list_loop *ls, int id, unsigned int delay, unsigned int rate);
static void mxk_stop_repeat(struct mxk_repeat *ma);

static int mxk_run_repeat(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************/

static int mxk_start_repeat(struct list_loop *ls, int id, unsigned int delay, unsigned int rate)
{
  struct mxk_repeat *mr;
  struct task_loop *tx;
  int type;

  type = lookup_task_type_loop(ls, "repeater");
  if(type <= 0){
#ifdef DEBUG
    fprintf(stderr, "repeat: type not defined\n");
#endif
    return -1;
  }

  mr = malloc(sizeof(struct mxk_repeat));
  if(mr == NULL){
    return -1;
  }

  mr->r_base = rate;

  mxk_milli_to_tv(&(mr->r_delay), delay);
  mxk_milli_to_tv(&(mr->r_rate), mr->r_base);

  mr->r_pass = 1;
  mr->r_working = 0;

  mr->r_down = 0;
  mr->r_last = KEY_RESERVED;

  mr->r_faster = KEY_RESERVED;
  mr->r_slower = KEY_RESERVED;
  mr->r_delta = 0;

  mr->r_vector = NULL;
  mr->r_count = 0;

  tx = add_tag_loop(ls, -1, mr, LO_WAIT_MASK | LO_STOP_MASK, mxk_run_repeat, NULL, type, id);
  if(tx == NULL){
#ifdef DEBUG
    fprintf(stderr, "repeat: task creation failed\n");
#endif
    mxk_stop_repeat(mr);
    return -1;
  }

  text_loop(ls, tx, "repeat:%d fixed=%ums", id, mr->r_base);

  return 0;
}

static int mxk_adjust_repeat(struct mxk_repeat *mr, int delta, unsigned int faster, unsigned int slower)
{
  if(slower == faster){
    return -1;
  }

  mr->r_delta = delta;
  if(delta){
    mr->r_faster = faster;
    mr->r_slower = slower;
  } else {
    mr->r_faster = KEY_RESERVED;
    mr->r_slower = KEY_RESERVED;
  }

  return 0;
}

static void mxk_stop_repeat(struct mxk_repeat *mr)
{
  if(mr){

    mxk_shutdown_client(mr->r_vector, mr->r_count);

    free(mr);
  }
}

/*************************************************************************/

int mxk_start_repeat_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i;
  struct control_loop *cl;
  char *value, *end;
  unsigned long values[2];
  struct mxk_global *mg;

  mg = global_loop(ls);
  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  if(args == NULL){
    print_control(cl, cd, 1, "repeat: need parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "repeat: need nonspace parameters\n");
    return 0;
  }

  value = args + idx;

  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "repeat: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  for(i = 0; (i < 2) && (idx = next_word(args, idx)) > 0; i++){
    value = args + idx;
    values[i] = strtoul(value, &end, 10);
    if(values[i] == 0){
      print_control(cl, cd, 1, "repeat: missing or non-natural parameter <%s>\n", value);
      return 0;
    }
  }

  if(mxk_start_repeat(ls, id, values[0], values[1])){
    print_control(cl, cd, 1, "repeat: unable to start task for id %d\n", id);
  } else {
    if(mg->g_verbose > 1){
      print_control(cl, cd, 1, "repeat: created id %d\n", id, values[0], values[1]);
    }
  }

  return 0;
}

int mxk_adjust_repeat_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int type, id, idx, cd, delta, i, count;
  struct task_loop *tx;
  struct control_loop *cl;
  struct mxk_repeat *mr;
  char *value, *end;
  unsigned int keys[2];
  char buffer[MXK_NAME_SIZE];

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "repeater");
  if(type <= 0){
    print_control(cl, cd, 1, "repeat: type not available\n");
    return 0;
  }

  if(args == NULL){
    print_control(cl, cd, 1, "repeat: usage: adjust-repeat id delta faster slower\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "repeat: unusable id argument\n");
    return 0;
  }

  value = args + idx;
  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "repeat: id parameter not a number\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    print_control(cl, cd, 1, "repeat: unusable increment argument\n");
    return 0;
  }

  value = args + idx;
  delta = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "repeat: increment parameter not a number\n");
    return 0;
  }

  for(i = 0; (i < 2) && (idx = next_word(args, idx)) > 0; i++){
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      print_control(cl, cd, 1, "repeat: unable to acquire key value\n");
      return 0;
    }

    keys[i] = name_to_code(mxk_key_codes, buffer);
#ifdef DEBUG
    fprintf(stderr, "adjust: key[%d]: %s->%d\n", i, buffer, keys[i]);
#endif
    if(keys[i] <= 0){
      print_control(cl, cd, 1, "repeat: unknown key <%s>\n", buffer);
      return 0;
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    mr = user_loop(ls, tx);
    if(mr){
      if(mxk_adjust_repeat(mr, delta, keys[0], keys[1])){
        print_control(cl, cd, 1, "repeat: unable to adjust entry for id %d\n", id);
      } else {
        text_loop(ls, tx, "repeat:%d %s/%s", id, code_to_name(mxk_key_codes, keys[0]), code_to_name(mxk_key_codes, keys[1]));
        count++;
      }
    } else {
      print_control(cl, cd, 1, "repeat[%d]: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "repeat: adjusted %d %s\n", count, (count == 1) ? "repeater" : "repeaters");

  return 0;
}

int mxk_list_repeat_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_repeat *mr;
  char *value;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "repeater");
  if(type <= 0){
    print_control(cl, cd, 1, "repeat: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      value = args + idx;
      id = strtoul(value, NULL, 10);
    }
  }

  count = 0;

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    mr = user_loop(ls, tx);
    if(mr){
      print_control(cl, cd, 1, "repeat[%d]: delay=%ld.%06lds rate=%ld.%06lds change=%ums slower=%s faster=%s\n", id, mr->r_delay.tv_sec, mr->r_delay.tv_usec, mr->r_rate.tv_sec, mr->r_rate.tv_usec, mr->r_delta, (mr->r_faster != KEY_RESERVED) ? code_to_name(mxk_key_codes, mr->r_faster) : "<unset>", (mr->r_slower != KEY_RESERVED) ? code_to_name(mxk_key_codes, mr->r_slower) : "<unset>");
      count++;
    } else {
      print_control(cl, cd, 1, "repeat[%d]: unavailable\n", id);
    }
  }
  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "repeater" : "repeaters");

  return 0;
}

/*************************************************************************/

static int mxk_run_repeat(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event *vp;
  struct link_loop *nk;
  struct mxk_repeat *mr;
  struct mxk_global *mg;
  int skip;

  mr = user_loop(ls, tl);
  mg = global_loop(ls);

  if(mask & LO_STOP_MASK){
    mxk_stop_repeat(mr);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        switch(vp->type){
          case EV_SYN :
            /* TODO: could put arm in here, so that syn only emitted if something else happend before */
            skip = 0;
            break;
          case EV_KEY :
            skip = 0;

            if((mr->r_faster == vp->code) || (mr->r_slower == vp->code)){
              if(vp->value == 1){
                if(mr->r_slower == vp->code){
                  if(mr->r_base > mr->r_delta){
                    mr->r_base -= mr->r_delta;
                  } else {
                    mr->r_base = 1;
                  }
                } else {
                  mr->r_base += mr->r_delta;
                }
#ifdef DEBUG
                fprintf(stderr, "repeat: new rate=%ums\n", mr->r_base);
#endif
                mxk_milli_to_tv(&(mr->r_rate), mr->r_base);
                if(mr->r_working){
                  set_time_loop(ls, tl, &(mr->r_rate));
                }
              }
              skip = 1;
              break; /* WARNING: breaks */
            }

            switch(vp->value){
              case 0 :
                if(mr->r_last == vp->code){
                  mr->r_last = KEY_RESERVED;
                }
                if(mr->r_down > 0){
                  mr->r_down--;
                }
                if(mr->r_down == 0){ /* released everything */
                  set_mask_loop(ls, tl, LO_TIME_MASK, -1);
                  mr->r_working = 0;
                }
                break;
              case 1 :
#if 0
                mr->r_last = vp->code;
                mr->r_down++;
                set_mask_loop(ls, tl, LO_TIME_MASK, 1);
                if(mr->r_working == 0){
                  set_time_loop(ls, tl, &(mr->r_delay));
                }
#endif
                mr->r_down++;
                set_mask_loop(ls, tl, LO_TIME_MASK, 1);
                if((mr->r_working == 0) || (mr->r_last != vp->code)){
                  mr->r_working = 0;
                  set_time_loop(ls, tl, &(mr->r_delay));
                }
                mr->r_last = vp->code;
                break;
              default :
                skip = 1;
                break;
            }
            break;
          default :
            skip = mr->r_pass ? 0 : 1; /* TODO: have a toggle bypass command */
            break;
        }

        if(skip == 0){
          mxk_send_client(ls, tl, mr->r_vector, mr->r_count, vp->type, vp->code, vp->value);
#ifdef DEBUG
        } else {
          fprintf(stderr, "repeat: skipping event 0x%x:0x%x:%d\n", vp->type, vp->code, vp->value);
#endif
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  if(mask & LO_TIME_MASK){
    if(mr->r_last != KEY_RESERVED){
      if(mr->r_working == 0){
        mr->r_working = 1;
        set_time_loop(ls, tl, &(mr->r_rate));
      }
      if(mr->r_pass){
        mxk_send_client(ls, tl, mr->r_vector, mr->r_count, EV_KEY, mr->r_last, 0);
        mxk_send_client(ls, tl, mr->r_vector, mr->r_count, EV_SYN, 0, 0);
        mxk_send_client(ls, tl, mr->r_vector, mr->r_count, EV_KEY, mr->r_last, 1);
      }
#ifdef DEBUG
    } else {
      fprintf(stderr, "repeat: logic problem: repeating, but nothing to repeat\n");
#endif
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_repeat(void *repeat, int type, int id)
{
  struct mxk_repeat *mr;
  struct mxk_client *mc;

  mr = repeat;

  mc = mxk_insert_client(mr->r_vector, mr->r_count, type, id);
  if(mc == NULL){
    return -1;
  }

  mr->r_vector = mc;
  mr->r_count++;

  return 0;
}

int mxk_startup_repeat(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "repeater", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_repeat);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_repeat(struct mxk_global *mg)
{
}
