/*
 * (c) Marc Welz 2005, 2006, 2007, 2008:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <loop.h>
#include <code-name.h>
#include <word-work.h>

#include "mxk.h"

/* single axis motion to keypress mapper */

#define MXK_SINGLE_MODE_QUICK     0
#define MXK_SINGLE_MODE_TIMED     1
#define MXK_SINGLE_MODE_SUSTAINED 2

struct mxk_single{
  int i_type;
  int i_axis;

  int i_mode;
  unsigned int i_cutoff;

  int i_last;

  unsigned int i_keys[2];

  struct timeval i_tv;

  struct mxk_client *i_vector;
  int i_count;
};

static struct task_loop *mxk_start_single(struct list_loop *ls, int id, int mode, int axis, int down, int up);
static int mxk_run_single(struct list_loop *ls, struct task_loop *tl, int mask);

static int mxk_mode_single(struct list_loop *ls, struct task_loop *tx, int mode, char *args);
static int mxk_threshold_single(struct list_loop *ls, struct task_loop *tx, int index, char *args);
static int mxk_timed_single(struct list_loop *ls, struct task_loop *tx, int mode, char *args);

/******************************************/

static struct code_name mxk_single_codes[] = {
  { "quick" ,     MXK_SINGLE_MODE_QUICK }, 
  { "timed" ,     MXK_SINGLE_MODE_TIMED }, 
  { "sustained" , MXK_SINGLE_MODE_SUSTAINED }, 
  {  NULL,        -1 }, 
};

/******************************************/

static struct task_loop *mxk_start_single(struct list_loop *ls, int id, int mode, int axis, int down, int up)
{
  struct mxk_global *mg;
  struct mxk_single *mi;
  struct task_loop *tx;
  int type;

  mg = global_loop(ls);
  type = lookup_task_type_loop(ls, "single");

  mi = malloc(sizeof(struct mxk_single));
  if(mi == NULL){
    return NULL;
  }

  mi->i_vector = NULL;
  mi->i_count = 0;

  mi->i_type = mode;
  mi->i_axis = axis;

  mi->i_tv.tv_sec = 0;
  mi->i_tv.tv_usec = 0;

  mi->i_mode = (mode == EV_REL) ? MXK_SINGLE_MODE_QUICK : MXK_SINGLE_MODE_SUSTAINED;
  mi->i_cutoff = 1;

  mi->i_last = (-1);

  mi->i_keys[0] = down;
  mi->i_keys[1] = up;

#ifdef DEBUG
  fprintf(stderr, "single[%d]: down=0x%x, up=0x%x\n", id, down, up);
#endif

  tx = add_tag_loop(ls, -1, mi, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_single, NULL, type, id);
  if(tx == NULL){
    free(mi);
    return NULL;
  }

  text_loop(ls, tx, "single:%d 0x%x/0x%x", mi->i_keys[0], mi->i_keys[1]);

  return tx;
}

static int mxk_mode_single(struct list_loop *ls, struct task_loop *tx, int mode, char *args)
{
  struct mxk_single *mi;

  mi = user_loop(ls, tx);
  mi->i_mode = mode;

  return 0;
}

static int mxk_threshold_single(struct list_loop *ls, struct task_loop *tx, int index, char *args)
{
  struct mxk_single *mi;
  char *ptr;
  int idx, value;

  mi = user_loop(ls, tx);

  idx = start_word(args, 0);
  if(idx < 0){
    mi->i_cutoff = 0;
    return 0;
  }

  ptr = args + idx;
  value = strtoul(ptr, NULL, 10);
  if(value <= 0){
    return -1;
  }

  mi->i_cutoff = value;

  return 0;
}

static int mxk_timed_single(struct list_loop *ls, struct task_loop *tx, int mode, char *args)
{
  struct mxk_single *mi;
  char *ptr;
  int idx;
  unsigned int value;

  mi = user_loop(ls, tx);

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  ptr = args + idx;
  value = strtoul(ptr, NULL, 10);
  if(value <= 0){
    return -1;
  }
  
  mi->i_mode = MXK_SINGLE_MODE_TIMED;
  mxk_milli_to_tv(&(mi->i_tv), value);

  return 0;
}

/****************************************************************/

int mxk_quick_single_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "single", MXK_MISC_OPTIONAL, MXK_SINGLE_MODE_QUICK, mxk_mode_single);

  return 0;
}

int mxk_sustained_single_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "single", MXK_MISC_OPTIONAL, MXK_SINGLE_MODE_SUSTAINED, mxk_mode_single);

  return 0;
}

int mxk_timed_single_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "single", MXK_MISC_REQUIRED, MXK_SINGLE_MODE_TIMED, mxk_timed_single);

  return 0;
}

int mxk_threshold_single_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "single", MXK_MISC_REQUIRED, 0, mxk_threshold_single);

  return 0;
}

/****************************************************************/

int mxk_start_single_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  char buffer[MXK_NAME_SIZE];
  struct control_loop *cl;
  struct code_name *atable;
  char *end, *value;
  int cd, idx, id, type, axis, i;
  unsigned int key[2];
  struct mxk_global *mg;

  mg = global_loop(ls);
  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "single: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "single: unable to acquire first parameter\n");
    return 0;
  }

  value = args + idx;
  id = strtoul(value, &end, 10);
  if(*end != ' '){
    print_control(cl, cd, 1, "single: require multiple parameters, first parameter needs to be an integer\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, cd, 1, "single: missing or malformed type parameter\n");
    return 0;
  }

  type = name_to_code(mxk_event_types, buffer);
  switch(type){
    case EV_REL :
      atable = mxk_rel_codes;
      break;
    case EV_ABS :
      atable = mxk_abs_codes;
      break;
    default :
      print_control(cl, cd, 1, "single: unsupported type <%s>\n", buffer);
      return 0;
  }

  idx = next_word(args, idx);
  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    print_control(cl, cd, 1, "single: missing or malformed axis parameter\n");
    return 0;
  }

  axis = name_to_code(atable, buffer);
  if(axis < 0){
    print_control(cl, cd, 1, "single: unknown axis <%s>\n", buffer);
    return 0;
  }

  for(i = 0; i < 2; i++){
    idx = next_word(args, idx);
    if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
      print_control(cl, cd, 1, "single: missing or malformed key %d\n", i);
      return 0;
    }

    key[i] = name_to_code(mxk_key_codes, buffer);
    if(key[i] < 0){
      print_control(cl, cd, 1, "single: unknown key <%s>\n", buffer);
      return 0;
    }
  }

  if(mxk_start_single(ls, id, type, axis, key[0], key[1]) == NULL){
    print_control(cl, cd, 1, "single: unable to start task for id %d\n", id);
  } else {
    if(mg->g_verbose > 0){
      print_control(cl, cd, 1, "single: created id %d\n", id);
    }
  }

  return 0;
}

int mxk_list_single_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_single *mi;
  char *mode, *axis, *up, *down, *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "single");
  if(type <= 0){
    print_control(cl, cd, 1, "single: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;

  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    mi = user_loop(ls, tx);
    if(mi){

      mode = code_to_name(mxk_event_types, mi->i_type);
      switch(mi->i_type){
        case EV_REL :
          axis = code_to_name(mxk_rel_codes, mi->i_axis);
          break;
        case EV_ABS :
          axis = code_to_name(mxk_abs_codes, mi->i_axis);
          break;
        default :
          axis = NULL;
          break;
      }

      down = code_to_name(mxk_key_codes, mi->i_keys[0]);
      up   = code_to_name(mxk_key_codes, mi->i_keys[1]);

      ptr = code_to_name(mxk_single_codes, mi->i_mode);

      print_control(cl, cd, 0, "single[%d]: duration=%s, ", tag_loop(ls, tx), ptr ? ptr : "<unknown>");
      if(mi->i_mode == MXK_SINGLE_MODE_TIMED){
        print_control(cl, cd, 0, "delay=%ld.%06lds, ", mi->i_tv.tv_sec, mi->i_tv.tv_usec);
      }
      print_control(cl, cd, 0, "cutoff=%d, motion=%s, axis=%s, down=%s, up=%s\n", mi->i_cutoff, mode ? mode : "unknown", axis ? axis : "unknown", up ? up : "<unset>", down ? down : "<unset>");

      count++;
    } else {
      print_control(cl, cd, 1, "array[%d]: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}
/***********************************************************************/

static int mxk_run_single(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_single *mi;
  struct mxk_global *mg;
  struct input_event *vp;
  struct link_loop *nk;
  unsigned int key;
  int cut;

  mg = global_loop(ls);
  mi = user_loop(ls, tl);

  if(mi == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_shutdown_client(mi->i_vector, mi->i_count);
#ifdef DEBUG
    mi->i_vector = NULL;
    mi->i_count = 0;
#endif
    free(mi);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        if(mi->i_type == vp->type){
#ifdef DEBUG
          fprintf(stderr, "single[%d]: got matching event type 0x%x (axis 0x%x ? 0x%x)\n", tag_loop(ls, tl), vp->type, mi->i_axis, vp->code);
#endif
          if(mi->i_axis == vp->code){

            /* candidate keys */
            if(vp->value < 0){
              cut = vp->value * (-1);
              key = mi->i_keys[0];
            } else if(vp->value > 0){
              cut = vp->value;
              key = mi->i_keys[1];
            } else {
              cut = 0;
              key = (-1);
            }

            /* if there was an old key and it differs from current, release it */
            if((key != mi->i_last) && (mi->i_last >= 0)){
              mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, mi->i_last, 0);
#ifdef DEBUG
              fprintf(stderr, "single[%d]: released 0x%x\n", tag_loop(ls, tl), mi->i_last);
#endif
              mi->i_last = (-1);
            }

            /* if we are above cutoff (and there is a key), then handle it */
            if((cut >= mi->i_cutoff) && (key >= 0)){
              switch(mi->i_mode){
                case MXK_SINGLE_MODE_TIMED :
                  set_time_loop(ls, tl, &(mi->i_tv));
                  set_mask_loop(ls, tl, LO_TIME_MASK, 1);
                  mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, key, 1);
                  mi->i_last = key;
#ifdef DEBUG
                  fprintf(stderr, "single[%d]: timed press 0x%x\n", tag_loop(ls, tl), key);
#endif
                  break;
                case MXK_SINGLE_MODE_SUSTAINED :
                  mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, key, 1);
                  mi->i_last = key;
#ifdef DEBUG
                  fprintf(stderr, "single[%d]: sustained press 0x%x\n", tag_loop(ls, tl), key);
#endif
                  break;
                  /* case MXK_SINGLE_MODE_QUICK : */
                default :
                  mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, key, 1);
                  mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_SYN, 0, 0);
                  mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, key, 0);
#ifdef DEBUG
                  fprintf(stderr, "single[%d]: quick press 0x%x\n", tag_loop(ls, tl), key);
#endif
                  mi->i_last = (-1);
                  break;
              }
              mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_SYN, 0, 0);
            }
          }
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  if(mask & LO_TIME_MASK){
    if(mi->i_last >= 0){
      mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_KEY, mi->i_last, 0);
      mxk_send_client(ls, tl, mi->i_vector, mi->i_count, EV_SYN, 0, 0);
      mi->i_last = (-1);
      set_mask_loop(ls, tl, LO_TIME_MASK, -1);
    }
  }

  return 0;
}

/***********************************************************************/

static int mxk_insert_single(void *single, int type, int id)
{
  struct mxk_single *mi;
  struct mxk_client *mc;

  mi = single;

  mc = mxk_insert_client(mi->i_vector, mi->i_count, type, id);
  if(mc == NULL){
    return -1;
  }

  mi->i_vector = mc;
  mi->i_count++;

  return 0;
}

int mxk_startup_single(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "single", MXK_EVT_SOURCE | MXK_EVT_TARGET, 32, mxk_insert_single);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_single(struct mxk_global *mg)
{
}
