/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* sorter module */

struct mxk_range{
  int r_type;
  int r_default;
  int r_size;
  unsigned int *r_range;
};

struct mxk_change{
  struct mxk_client *c_vector;
  int c_count;
};

struct mxk_sorter{
  struct mxk_range *s_range;
  int s_size;

  int s_current; /* only used for additions, modeless otherwise */
  struct mxk_change *s_change;
  int s_top;

  int s_default;
};

static void mxk_destroy_sorter(struct mxk_sorter *sorter);
static struct mxk_sorter *mxk_create_sorter();
static int mxk_run_sorter(struct list_loop *ls, struct task_loop *tl, int mask);

static int update_change_sorter(struct mxk_sorter *sorter, int code);
static struct mxk_range *update_range_sorter(struct mxk_sorter *sorter, int type);

/**********************************************************************/

struct mxk_sorter *mxk_create_sorter()
{
  struct mxk_sorter *sorter;

  sorter = malloc(sizeof(struct mxk_sorter));
  if(sorter == NULL){
    return NULL;
  }

  sorter->s_range = NULL;
  sorter->s_size = 0;

  sorter->s_current = (-1);

  sorter->s_change = NULL;
  sorter->s_top = 0;

  sorter->s_default = (-1);

  return sorter;
}

static void mxk_destroy_sorter(struct mxk_sorter *sorter)
{
  int i;
  struct mxk_range *rf;
  struct mxk_change *cf;

  if(sorter == NULL){
    return;
  }

  if(sorter->s_range){
    for(i = 0; i < sorter->s_size; i++){
      rf = &(sorter->s_range[i]);
      if(rf->r_range){
        free(rf->r_range);
        rf->r_range = NULL;
      }
    }
    free(sorter->s_range);
    sorter->s_range = NULL;
  }

  sorter->s_size = 0;
  sorter->s_current = -1;

  if(sorter->s_change){
    for(i = 0; i < sorter->s_top; i++){
      cf = &(sorter->s_change[i]);
      mxk_shutdown_client(cf->c_vector, cf->c_count);
#ifdef DEBUG
      cf->c_vector = NULL;
      cf->c_count = 0;
#endif
    }
    free(sorter->s_change);
    sorter->s_change = NULL;
  }
  sorter->s_top = 0;

  sorter->s_default = -1;

  free(sorter);
}

/**********************************************************************/

int mxk_bin_sorter(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_sorter *sorter;
  int bin, idx;

  sorter = user_loop(ls, tx);

  if(args == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  bin = strtoul(args + idx, NULL, 10);

  if(bin >= sorter->s_top){
    return -1;
  }

  if(bin < 0){
    sorter->s_current = (-1);
  } else {
    sorter->s_current = bin;
  }

  return 0;
}

static struct mxk_range *update_range_sorter(struct mxk_sorter *sorter, int type)
{
  int i;
  struct mxk_range *rf;

  for(i = 0; (i < sorter->s_size) && (sorter->s_range[i].r_type != type); i++);

  if(i == sorter->s_size){
    rf = realloc(sorter->s_range, sizeof(struct mxk_range) * (sorter->s_size + 1));
    if(rf == NULL){
      return NULL;
    }
    sorter->s_range = rf;
    rf = &(sorter->s_range[sorter->s_size]);
    sorter->s_size++;

    rf->r_type = type;
    rf->r_default = sorter->s_default;
    rf->r_size = 0;
    rf->r_range = NULL;
  } else {
    rf = &(sorter->s_range[i]);
  }

  return rf;
}

static int update_change_sorter(struct mxk_sorter *sorter, int code)
{
  struct mxk_change *cf;
  int i;

  if(code < 0){
    return 0;
  }

  if(sorter->s_top > code){
    return 0;
  }

  cf = realloc(sorter->s_change, sizeof(struct mxk_change) * (code + 1));
  if(cf == NULL){
    return -1;
  }
  sorter->s_change = cf;
  for(i = sorter->s_top; i <= code; i++){
    cf = &(sorter->s_change[i]);
    cf->c_vector = NULL;
    cf->c_count = 0;
  }
  sorter->s_top = code + 1;

  return 0;
}

int mxk_reserve_sorter(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  int idx, type, bin;
  char *end;
  char buffer[MXK_NAME_SIZE];
  struct mxk_range *rf;
  struct mxk_sorter *sorter;

  sorter = user_loop(ls, tx);

  if(args == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
    return -1;
  }

  type = name_to_code(mxk_event_types, buffer);
  if(type < 0){
    return -1;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    return -1;
  }

  bin = strtoul(args + idx, &end, 10);
  if((*end != ' ') && (*end != '\0')){
    return -1;
  }

  rf = update_range_sorter(sorter, type);
  if(rf == NULL){
    return -1;
  }

  if(rf->r_size > 0){
    return -1;
  }

  if(update_change_sorter(sorter, bin)){
    return -1;
  }

  rf->r_default = bin;
  return 0;
}

int mxk_match_sorter(struct list_loop *ls, struct task_loop *tx, int special, char *args)
{
  struct mxk_sorter *sorter;
  int idx, bin, i;
  struct input_event event;
  struct input_event *vp;
  struct mxk_range *rf;
  char *end;
  unsigned int *tmp;

  sorter = user_loop(ls, tx);
  vp = &event;

  if(args == NULL){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  if(mxk_parse_evt(args + idx, vp) < 0){
#ifdef DEBUG
    fprintf(stderr, "unable to parse event\n");
#endif
    return -1;
  }

  idx = next_word(args, idx);
  if(idx < 0){
    return -1;
  }

  bin = strtoul(args + idx, &end, 10);
  if((*end != ' ') && (*end != '\0')){
#ifdef DEBUG
    fprintf(stderr, "unable to acquire bin number\n");
#endif
    return -1;
  }

  if(update_change_sorter(sorter, bin)){
    return -1;
  }

  rf = update_range_sorter(sorter, vp->type);
  if(rf == NULL){
    return -1;
  }

  if(rf->r_size <= vp->code){
    tmp = realloc(rf->r_range, sizeof(unsigned int) * (vp->code + 1));
    if(tmp == NULL){
      return -1;
    }
    rf->r_range = tmp;
    for(i = rf->r_size; i <= vp->code; i++){
      rf->r_range[i] = rf->r_default;
    }
    rf->r_size = vp->code + 1;
  }

  rf->r_range[vp->code] = bin;

  return 0;
}

/**********************************************************************/

static int mxk_run_sorter(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_sorter *sorter;

  struct mxk_range *rf;
  struct mxk_change *cf;

  struct input_event *vp;
  struct link_loop *nk;

  int i, action;

  mg = global_loop(ls);
  sorter = user_loop(ls, tl);

  if(sorter == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_sorter(sorter);
    return 0;
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
        action = sorter->s_default;
        for(i = 0; i < sorter->s_size; i++){
          rf = &(sorter->s_range[i]);
          if(vp->type == rf->r_type){
#if 0
            if(vp->code < 0){
              fprintf(stderr, "sorter: logic problem: code %d\n", vp->code);
              abort();
            }
#endif
            if(vp->code < rf->r_size){
              action = rf->r_range[vp->code];
            } else {
              action = rf->r_default;
            }
            /* WARNING: break */
            break;
          }
        }
        if((action < 0) || (action >= sorter->s_top)){
#ifdef DEBUG
          fprintf(stderr, "sorter: doing multicast: action %d\n", action);
#endif
          for(i = 0; i < sorter->s_top; i++){
            cf = &(sorter->s_change[i]);
            mxk_send_client(ls, tl, cf->c_vector, cf->c_count, vp->type, vp->code, vp->value);
          }
        } else {
          cf = &(sorter->s_change[action]);
#ifdef DEBUG
          fprintf(stderr, "sorter: action %d: relay to %d\n", action, cf->c_count);
#endif
          mxk_send_client(ls, tl, cf->c_vector, cf->c_count, vp->type, vp->code, vp->value);
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

int mxk_bin_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "sorter", MXK_MISC_REQUIRED, 0, mxk_bin_sorter);

  return 0;
}

int mxk_reserve_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "sorter", MXK_MISC_REQUIRED, 0, mxk_reserve_sorter);

  return 0;
}

int mxk_match_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "sorter", MXK_MISC_REQUIRED, 0, mxk_match_sorter);

  return 0;
}

int mxk_start_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_sorter *sorter;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type, bin;
  char *end;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);
  bin = 0;

  type = lookup_task_type_loop(ls, "sorter");
  if(type <= 0){
    print_control(cl, cd, 1, "sorter: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "sorter: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "sorter: unable to acquire first parameter\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);
  if((*end != ' ') && (*end != '\0')){
    print_control(cl, cd, 1, "sorter: require an integer id as parameter\n");
    return 0;
  }

  sorter = mxk_create_sorter();
  if(sorter == NULL){
    print_control(cl, cd, 1, "sorter: unable to allocate state\n");
    return 0;
  }

  idx = next_word(args, idx);
  if(idx > 0){
    bin = strtoul(args + idx, &end, 10);
    if((*end != ' ') && (*end != '\0')){
      print_control(cl, cd, 1, "sorter: default bin malformed\n");
      mxk_destroy_sorter(sorter);
      return 0;
    }
    if(update_change_sorter(sorter, bin)){
      print_control(cl, cd, 1, "sorter: unable to allocate default bin %d\n", bin);
      mxk_destroy_sorter(sorter);
      return 0;
    }
    sorter->s_default = bin;
  }

  tx = add_tag_loop(ls, -1, sorter, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_sorter, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "sorter: unable to create task\n");
    mxk_destroy_sorter(sorter);
    return 0;
  }

  text_loop(ls, tx, "sorter:%d default=%d", id, bin);
  if(mg->g_verbose > 1){
    print_control(cl, cd, 1, "sorter: created id %d\n", id);
  }

  return 0;
}

int mxk_list_sorter_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, type, count, i, j;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_sorter *sorter;
  struct mxk_range *rf;
  struct code_name *codes;
  char tname[MXK_NAME_SIZE];
  char *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "sorter");
  if(type <= 0){
    print_control(cl, cd, 1, "sorter: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    sorter = user_loop(ls, tx);

    if(sorter){
      print_control(cl, cd, 0, "sorter[%d]: bins=%d, default=%d, current=%d\n", tag_loop(ls, tx), sorter->s_top, sorter->s_default, sorter->s_current);

      for(i = 0; i < sorter->s_size; i++){
        rf = &(sorter->s_range[i]);
        ptr = code_to_name(mxk_event_types, rf->r_type);
        if(ptr){
          strncpy(tname, ptr, MXK_NAME_SIZE - 1);
        } else {
          snprintf(tname, MXK_NAME_SIZE - 1, "0x%x", rf->r_type);
        }
        tname[MXK_NAME_SIZE - 1] = '\0';

        print_control(cl, cd, 0, "default:%s %d\n", tname, rf->r_default);

        switch(rf->r_type){
          case EV_KEY :
            codes = mxk_key_codes;
            break;
          case EV_REL :
            codes = mxk_rel_codes;
            break;
          case EV_ABS :
            codes = mxk_abs_codes;
            break;
          default :
            codes = NULL;
        }

        for(j = 0; j < rf->r_size; j++){

          if(rf->r_range[j] != rf->r_default){
            ptr = NULL;
            if(codes){
              ptr = code_to_name(codes, j);
            }

            if(ptr){
              print_control(cl, cd, 0, "%s:%s %d\n", tname, ptr, rf->r_range[j]);
            } else {
              print_control(cl, cd, 0, "%s:0x%02x %d\n", tname, j, rf->r_range[j]);
            }
          }
        }
      }

      count++;
    } else {
      print_control(cl, cd, 1, "sorter: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}
/**********************************************************************/

static int mxk_insert_sorter(void *state, int type, int id)
{
  struct mxk_sorter *sorter;
  struct mxk_change *cf;
  struct mxk_client *mc;
  int i;

  sorter = state;

  if((sorter->s_current < 0) || (sorter->s_current >= sorter->s_top)){
    for(i = 0; i < sorter->s_top; i++){
      cf = &(sorter->s_change[i]);
      mc = mxk_insert_client(cf->c_vector, cf->c_count, type, id);
      if(mc == NULL){
        return -1;
      }
      cf->c_vector = mc;
      cf->c_count++;
    }
  } else {
#ifdef DEBUG
    fprintf(stderr, "sorter: unicast insert for %d\n", sorter->s_current);
#endif
    cf = &(sorter->s_change[sorter->s_current]);
    mc = mxk_insert_client(cf->c_vector, cf->c_count, type, id);
    if(mc == NULL){
      return -1;
    }
    cf->c_vector = mc;
    cf->c_count++;
  }

  return 0;
}

int mxk_startup_sorter(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "sorter", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_sorter);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_sorter(struct mxk_global *mg)
{
}
