/*
 * (c) Marc Welz 2005-2008
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>

#include "mxk.h"

#define MXK_SOURCE_BYPASS 0x0
#define MXK_SOURCE_ACTIVE 0x1
#define MXK_SOURCE_LOCKED 0x2

static struct code_name mxk_state_codes[] = {
  { "bypass", MXK_SOURCE_BYPASS },
  { "active", MXK_SOURCE_ACTIVE },
  { "locked", MXK_SOURCE_LOCKED },
  { NULL, -1 }
};

#define MXK_FLAKE_UNPRESSED   0x1
#define MXK_FLAKE_OVERPRESSED 0x2
#define MXK_FLAKE_PARTIAL     0x4

static struct code_name mxk_flaky_codes[] = {
  { "phantom-release", MXK_FLAKE_UNPRESSED },
  { "duplicate-press", MXK_FLAKE_OVERPRESSED },
  { "partial-read",    MXK_FLAKE_PARTIAL },
  {  NULL, 0 },
};

#define MXK_HWREP_DISCARD  0x1
#define MXK_HWREP_FORWARD  0x2
#define MXK_HWREP_REPLACE  0x3

static struct code_name mxk_hwrep_codes[] = {
  { "discard", MXK_HWREP_DISCARD },
  { "forward", MXK_HWREP_FORWARD },
  { "replace", MXK_HWREP_REPLACE },
  {  NULL, 0 },
};

#define MXK_KEY_HELDDOWN    0x1
#define MXK_KEY_DOESAUTO    0x2

struct mxk_source{
  int s_version;
  unsigned long s_mask[NBITS(EV_MAX)];
  unsigned long s_leds[NBITS(LED_MAX)];
  char s_name[MXK_NAME_SIZE];
#ifdef EVIOCGUNIQ
  char s_uniq[MXK_NAME_SIZE];
#endif
  int s_state;
  dev_t s_dev; /* source inode */

  struct mxk_client *s_vector;
  int s_count;

  char *s_repeat;
  int s_sync;

  int s_flaky;
  int s_hwrep;
  int s_presses;

  unsigned long s_seq_count;
  unsigned int s_seq_value;
};

#define MXK_PRESS_HONEYMOON 5

static int mxk_flaky_source(struct list_loop *ls, struct task_loop *tx, int check, char *args);
static int mxk_hwrep_source(struct list_loop *ls, struct task_loop *tx, int hwrep, char *args);
static int mxk_state_source(struct list_loop *ls, struct task_loop *tx, int state, char *args);
static int mxk_grab_source(struct mxk_source *ms, int fd, int state);

static int mxk_list_sources(struct list_loop *ls, struct task_loop *tl);
static int mxk_used_sources(struct list_loop *ls, struct task_loop *tl);

static void mxk_free_source(struct mxk_source *ms);

static int mxk_parse_source(struct list_loop *ls, struct task_loop *tl, char *args, int search);
static int mxk_run_source(struct list_loop *ls, struct task_loop *tl, int mask);

#define MXK_PROBE_SHOW    0x0   /* display only */
#define MXK_PROBE_CAPS    0x1   /* acquire caps */
#define MXK_PROBE_SINGLE  0x2   /* more verbose error messages */

#define MXK_MATCH_FILE      0
#define MXK_MATCH_NAME      1
#define MXK_MATCH_CAPS      2

/*************************************************************************/

int mxk_file_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  return mxk_parse_source(ls, tl, args, MXK_MATCH_FILE);
}

int mxk_name_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  return mxk_parse_source(ls, tl, args, MXK_MATCH_NAME);
}

int mxk_caps_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  return mxk_parse_source(ls, tl, args, MXK_MATCH_CAPS);
}

int mxk_list_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_list_sources(ls, tl);

  return 0;
}

int mxk_used_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_used_sources(ls, tl);

  return 0;
}

int mxk_bypass_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, MXK_SOURCE_BYPASS, mxk_state_source);

  return 0;
}

int mxk_activate_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, MXK_SOURCE_ACTIVE, mxk_state_source);

  return 0;
}

int mxk_lock_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, MXK_SOURCE_LOCKED, mxk_state_source);

  return 0;
}

int mxk_forward_repeat_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, MXK_HWREP_FORWARD, mxk_hwrep_source);

  return 0;
}

int mxk_discard_repeat_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, MXK_HWREP_DISCARD, mxk_hwrep_source);

  return 0;
}

int mxk_replace_repeat_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, MXK_HWREP_REPLACE, mxk_hwrep_source);

  return 0;
}

int mxk_check_flaky_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, 1, mxk_flaky_source);

  return 0;
}

int mxk_believe_flaky_source_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "source", MXK_MISC_OPTIONAL, 0, mxk_flaky_source);

  return 0;
}

/*************************************************************************/

static int mxk_flaky_source(struct list_loop *ls, struct task_loop *tx, int check, char *args)
{
  struct mxk_source *ms;
  int i;

  ms = user_loop(ls, tx);

  if(check){
    if(ms->s_repeat == NULL){
      ms->s_repeat = malloc(sizeof(char) * KEY_MAX);
      if(ms->s_repeat == NULL){
        return -1;
      }
      for(i = 0; i < KEY_MAX; i++){
        ms->s_repeat[i] = 0;
      }
    }
  } else {
    if(ms->s_repeat){
      free(ms->s_repeat);
      ms->s_repeat = NULL;
    }
  }

  return 0;
}

static int mxk_hwrep_source(struct list_loop *ls, struct task_loop *tx, int hwrep, char *args)
{
  struct mxk_source *ms;

  ms = user_loop(ls, tx);
  if(ms == NULL){
    return 0;
  }

  ms->s_hwrep = hwrep;

  return 0;
}

static int mxk_state_source(struct list_loop *ls, struct task_loop *tx, int state, char *args)
{
  struct mxk_source *ms;
  int fx;

#ifdef DEBUG
  fprintf(stderr,  "state: setting state to %d\n", state);
#endif

  ms = user_loop(ls, tx);
  fx = fd_loop(ls, tx);
  if(ms == NULL){
    return -1;
  }

  return mxk_grab_source(ms, fx, state);
}

static int mxk_grab_source(struct mxk_source *ms, int fd, int state)
{
  int result, value;

  if((MXK_SOURCE_LOCKED & state) == (MXK_SOURCE_LOCKED & ms->s_state)){
    ms->s_state = state;
    return 0;
  }

  /* KERNEL BUG ? unable to release grabbed device */

  value = (MXK_SOURCE_LOCKED & state) ? 1 : 0;

  result = ioctl(fd, EVIOCGRAB, &value);
#ifdef DEBUG
  fprintf(stderr, "grab: %s\n", result ? strerror(errno) : "zero (ok)");
#endif
  if(result == 0){
    ms->s_state = state;
  }

  return result;
}

/************************************************************/

struct mxk_source *mxk_is_active(struct list_loop *ls, dev_t dev)
{
  struct mxk_global *mg;
  struct task_loop *tx;
  struct mxk_source *ms;
  int type;

  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "source");
  if(type > 0){
    for(tx = find_task_tag_loop(ls, NULL, NULL, type, 0); tx; tx = find_task_tag_loop(ls, NULL, tx, type, 0)){

      ms = user_loop(ls, tx);
      if(ms){
        if(dev == ms->s_dev){
          return ms;
        }
      }
    }
  }

  return NULL;
}

int mxk_is_target(char *name)
{
  if(strncmp(name, "MXK/", 4)){
    return 0;
  }

  return 1;
}

/***********************************************************************/

static void mxk_free_source(struct mxk_source *ms)
{
  if(ms == NULL){
    return;
  }

  mxk_shutdown_client(ms->s_vector, ms->s_count);
#ifdef DEBUG
  ms->s_vector = NULL;
  ms->s_count = 0;
#endif

  if(ms->s_repeat){
    free(ms->s_repeat);
    ms->s_repeat = NULL;
  }

  free(ms);
}

int mxk_filter_source(struct mxk_source *ms, char *filter, int search)
{
  char buffer[MXK_NAME_SIZE];
  char *ptr;
  int idx, no, type, result;

  if(filter == NULL){
    return -1;
  }

  switch(search){
    case MXK_MATCH_NAME :
      if(strstr(ms->s_name, filter)){
        return 1;
      } else {
        return 0;
      }
    case MXK_MATCH_CAPS :
      idx = start_word(filter, 0);
      if(idx < 0){
        return -1;
      }

      do{
        if(load_word(buffer, MXK_NAME_SIZE - 1, filter, idx) < 0){
          return -1;
        }

#ifdef DEBUG
       fprintf(stderr, "filter: examing token <%s>\n", buffer);
#endif

        ptr = buffer;
        switch(*ptr){
          case '+' :
            no = 0;
            ptr++;
            break;
          case '!' :
          case '-' :
            no = 1;
            ptr++;
            break;
          default :
            no = 0;
            break;
        }

        type = name_to_code(mxk_event_types, ptr);
        if(type < 0){
          return -1;
        }

        result = mxk_has_bit(ms->s_mask, EV_MAX, type);
        if(result < 0){
          return -1;
        }

#ifdef DEBUG
        fprintf(stderr, "filter: not=%d, cap=%s/%d, result=%d\n", no, ptr, type, result);
#endif

        if(no){
          if(result > 0){
            return 0;
          }
        } else {
          if(result <= 0){
            return 0;
          }
        }

        idx = next_word(filter, idx);
      } while(idx > 0);

      return 1;
    default :
      return 0;
  }
}

struct mxk_source *mxk_copy_source(struct mxk_source *ms)
{
  struct mxk_source *cs;
  int i;

  cs = malloc(sizeof(struct mxk_source));
  if(cs == NULL){
    return NULL;
  }

  cs->s_version = ms->s_version;

  for(i = 0; i < NBITS(EV_MAX); i++){
    cs->s_mask[i] = ms->s_mask[i];
  }

  for(i = 0; i < NBITS(LED_MAX); i++){
    cs->s_leds[i] = ms->s_leds[i];
  }

  memcpy(cs->s_name, ms->s_name, MXK_NAME_SIZE);
#ifdef EVIOCGUNIQ
  memcpy(cs->s_uniq, ms->s_uniq, MXK_NAME_SIZE);
#endif

  cs->s_state = ms->s_state;
  cs->s_dev = ms->s_dev;

  cs->s_vector = ms->s_vector;
  ms->s_vector = NULL;

  cs->s_count = ms->s_count;
  ms->s_count = 0;

  cs->s_repeat = ms->s_repeat;
  ms->s_repeat = NULL;

  cs->s_sync = ms->s_sync;
  cs->s_flaky = ms->s_flaky;
  cs->s_hwrep = ms->s_hwrep;
  cs->s_presses = 0;

  cs->s_seq_count = 0;
  cs->s_seq_value = 0;

  return cs;
}

int mxk_probe_source(struct list_loop *ls, struct task_loop *tl, struct mxk_source *ms, char *file, dev_t dev, int flags)
{
  struct control_loop *cl;
  int cd, fd, i;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  ms->s_version = 0;

  for(i = 0; i < NBITS(EV_MAX); i++){
    ms->s_mask[i] = 0;
  }

  for(i = 0; i < NBITS(LED_MAX); i++){
    ms->s_leds[i] = 0;
  }

  memset(ms->s_name, '\0', MXK_NAME_SIZE);
#ifdef EVIOCGUNIQ
  memset(ms->s_uniq, '\0', MXK_NAME_SIZE);
#endif

  ms->s_state = MXK_SOURCE_BYPASS;
  ms->s_dev = dev;

  ms->s_vector = NULL;
  ms->s_count = 0;

  ms->s_repeat = NULL;
  ms->s_sync = 0;
  ms->s_flaky = 0;
  ms->s_hwrep = MXK_HWREP_REPLACE;
  ms->s_presses = 0;

  /* now do work */

  fd = open(file, O_RDWR | O_NONBLOCK);
  if(fd < 0){
    switch(errno){
      case ENOENT :
      case ENODEV :
        break;
      default :
        print_control(cl, cd, 1, "source: unable to open %s: %s\n", file, strerror(errno));
        break;
    }
    return -1;
  }

  fcntl(fd, F_SETFD, FD_CLOEXEC);

  if(ioctl(fd, EVIOCGNAME(MXK_NAME_SIZE - 1), ms->s_name) < 0){
    print_control(cl, cd, 1, "source: unable to request name of %s: %s\n", file, strerror(errno));
    close(fd);
    return -1;
  }

  ms->s_name[MXK_NAME_SIZE - 1] = '\0';

  if(flags & MXK_PROBE_CAPS){
    if(ioctl(fd, EVIOCGBIT(0, NBITS(EV_MAX)), ms->s_mask) < 0){
      print_control(cl, cd, 1, "source: unable to establish capabilities of %s: %s\n", file, strerror(errno));
      close(fd);
      return -1;
    }
  }

  return fd;
}

struct task_loop *mxk_complete_source(struct list_loop *ls, struct task_loop *tl, struct mxk_source *ms, int fd, int id)
{
  struct stat st;
  struct mxk_global *mg;
  struct mxk_source *cs;
  struct control_loop *cl;
  struct task_loop *tx;
  unsigned long check;
  int cd, type, i;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);
  type = lookup_task_type_loop(ls, "source");

  if(mxk_is_target(ms->s_name)){
    print_control(cl, cd, 1, "source: refusing to make target <%s> a source\n", ms->s_name);
    return NULL;
  }

  if(ioctl(fd, EVIOCGVERSION, &(ms->s_version)) < 0){
    print_control(cl, cd, 1, "source: unable to acquire version information for <%s>: %s\n", ms->s_name, strerror(errno));
    return NULL;
  }

  check = 0;
  for(i = 0; i < NBITS(EV_MAX); i++){
    check |= ms->s_mask[i];
  }
  if(check == 0){
    if(ioctl(fd, EVIOCGBIT(0, NBITS(EV_MAX)), ms->s_mask) < 0){
      print_control(cl, cd, 1, "source: unable to establish capabilities of <%s>: %s\n", ms->s_name, strerror(errno));
      return NULL;
    }
  }

  for(i = 0; i < NBITS(LED_MAX); i++){
    ms->s_leds[i] = 0;
  }
  if(mxk_has_bit(ms->s_mask, EV_MAX, EV_LED)){
#ifdef DEBUG
    fprintf(stderr, "source: has leds, checking mask\n");
#endif
    if(ioctl(fd, EVIOCGBIT(EV_LED, NBITS(LED_MAX)), ms->s_leds) < 0){
      print_control(cl, cd, 1, "source: warning: unable to establish leds of <%s>: %s\n", ms->s_name, strerror(errno));
    }
  }

  if(ms->s_dev == 0){
    if(fstat(fd, &st) < 0){
      print_control(cl, cd, 1, "source: unable to stat file descriptor of <%s>: %s\n", ms->s_name, strerror(errno));
      return NULL;
    }
  }

#ifdef EVIOCGUNIQ
  memset(ms->s_uniq, '\0', MXK_NAME_SIZE);
  ioctl(fd, EVIOCGUNIQ(MXK_NAME_SIZE), ms->s_uniq);
  ms->s_uniq[MXK_NAME_SIZE - 1] = '\0';
#endif

  cs = mxk_copy_source(ms);
  if(cs == NULL){
    return NULL;
  }

  tx = add_tag_loop(ls, fd, cs, LO_READ_MASK | LO_WAIT_MASK | LO_STOP_MASK, mxk_run_source, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "source: unable to start source task for <%s> (%d,%d)\n", ms->s_name, type, id);
    mxk_free_source(cs);
    return NULL;
  }

  text_loop(ls, tx, "source:%d <%s>", id, cs->s_name);

  if(mxk_has_bit(cs->s_mask, EV_MAX, EV_KEY)){
    /* ignore failure, it is a nice to have in a way */
    mxk_flaky_source(ls, tx, 1, NULL);
  }

  return tx;
}

/**************************************************************************/

struct task_loop *mxk_file_source(struct list_loop *ls, struct task_loop *tl, char *device, int id)
{
  struct task_loop *tx;
  struct control_loop *cl;
  struct mxk_source msrc, *ms;
  struct stat st;
  int cd, fd;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  if(stat(device, &st)){
    print_control(cl, cd, 1, "source: unable to stat %s: %s\n", device, strerror(errno));
    return NULL;
  }

  if(!S_ISCHR(st.st_mode)){
    print_control(cl, cd, 1, "source: %s not a character device file\n", device);
    return NULL;
  }

  if(mxk_is_active(ls, st.st_rdev)){
    print_control(cl, cd, 1, "source: %s seems already active\n", device);
    return NULL;
  }

  ms = &msrc;
  fd = mxk_probe_source(ls, tl, ms, device, st.st_rdev, MXK_PROBE_SINGLE);
  if(fd < 0){
    return NULL;
  }

  tx = mxk_complete_source(ls, tl, ms, fd, id);
  if(tx == NULL){
    close(fd);
    return NULL;
  }

  return tx;
}

struct task_loop *mxk_match_source(struct list_loop *ls, struct task_loop *tl, char *filter, int id, int search)
{
  struct stat st;
  struct control_loop *cl;
  struct dirent *de;
  DIR *dh;
  struct task_loop *tx;
  struct mxk_source msrc, *ms;
  char path[MXK_NAME_SIZE];
  int cd, fd;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  dh = opendir(MXK_SOURCE_DIR);
  if(dh == NULL){
    print_control(cl, cd, 1, "source: open of %s failed: %s\n", MXK_SOURCE_DIR, strerror(errno));
    return NULL;
  }

  do{
    de = readdir(dh);
    if(de){
      if(!strncmp(de->d_name, "event", 5)){

        snprintf(path, MXK_NAME_SIZE - 1, "%s/%s", MXK_SOURCE_DIR, de->d_name);
        path[MXK_NAME_SIZE - 1] = '\0';

        if(stat(path, &st) == 0){
          ms = mxk_is_active(ls, st.st_rdev);
          if(ms){
            if(mxk_filter_source(ms, filter, search) > 0){
              print_control(cl, cd, 1, "source: skipping %s: already in use\n", path);
            }
          } else {
            ms = &msrc;
            fd = mxk_probe_source(ls, tl, ms, path, st.st_rdev, (search == MXK_MATCH_CAPS) ? MXK_PROBE_CAPS : MXK_PROBE_SHOW);
            if(fd >= 0){
              if(mxk_filter_source(ms, filter, search) > 0){
                if(mxk_is_target(ms->s_name)){
                  print_control(cl, cd, 1, "source: skipping %s: target device\n", path);
                } else {
                  tx = mxk_complete_source(ls, tl, ms, fd, id);
                  if(tx){
                    closedir(dh);
                    return tx; /* leave on success */
                  }
                }
              }
              close(fd);
            }
          }
        } else {
          switch(errno){
            case ENOENT :
            case ENODEV :
              break;
            default :
              print_control(cl, cd, 1, "source: unable to query %s: %s\n", path, strerror(errno));
              break;
          }
        }
      }
    }
  } while(de);

  closedir(dh);

  if(strchr(filter, '*')){
    print_control(cl, cd, 1, "source: warning: matches are exact, wildcards are not permitted\n");
  }


  return NULL;
}

/*************************************************************************/

static int mxk_parse_source(struct list_loop *ls, struct task_loop *tl, char *args, int search)
{
  int fd, id, idx;
  struct control_loop *cl;
  struct task_loop *tx;
  char *label, *name, *end;

  cl = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  if(args == NULL){
    print_control(cl, fd, 1, "source: need parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, fd, 1, "source: need nonspace parameters\n");
    return 0;
  }

  label = args + idx;

  id = strtoul(label, &end, 10);
  if(*end != ' '){
    print_control(cl, fd, 1, "source: require two parameters, first parameter needs to be an integer\n");
    return 0;
  }

  name = mxk_skip_space(end);

#ifdef DEBUG
  fprintf(stderr, "make: attempting to build source: name=%s, id=%d\n", name, id);
#endif
  if(search == MXK_MATCH_FILE){
    tx = mxk_file_source(ls, tl, name, id);
  } else {
    tx = mxk_match_source(ls, tl, name, id, search);
  }

  if(tx == NULL){
    print_control(cl, fd, 1, "source: unable to add source matching <%s>\n", name);
  } else {
    print_control(cl, fd, 1, "source: added source matching <%s>\n", name);
  }

  return 0;
}

/*************************************************************************/

static int mxk_used_sources(struct list_loop *ls, struct task_loop *tl)
{
  struct mxk_global *mg;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_source *ms;
  int cd, count, type, i;
  char *ptr, *extra;
  char delim;

  mg = global_loop(ls);
  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  count = 0;

  type = lookup_task_type_loop(ls, "source");
  if(type > 0){
    for(tx = find_task_tag_loop(ls, NULL, NULL, type, 0); tx; tx = find_task_tag_loop(ls, NULL, tx, type, 0)){
      ms = user_loop(ls, tx);
      if(ms){
        print_control(cl, cd, 0, "source[%d]: name=<%s> state=%s presses=%d ", tag_loop(ls, tx), ms->s_name, code_to_name(mxk_state_codes, ms->s_state), ms->s_presses);

        if(ms->s_seq_count){
          /* number of longest event sequences, double check for READ_GROUP */
          print_control(cl, cd, 0, "maxevseq=%lu*%u ", ms->s_seq_count, ms->s_seq_value);
        }

        if(ms->s_flaky) {
          print_control(cl, cd, 0, "flaky=");
          delim = '<';
          for(i = 1; i <= ms->s_flaky; i <<= 1){
            ptr = code_to_name(mxk_flaky_codes, i);
            if(ptr){
              print_control(cl, cd, 0, "%c%s", delim, ptr);
            } else {
              print_control(cl, cd, 0, "%c0x%x", delim, i);
            }
            delim = ' ';
          }
          print_control(cl, cd, 0, "> ");
        }

        print_control(cl, cd, 0, "repeats=");
        ptr = code_to_name(mxk_hwrep_codes, ms->s_hwrep);
        if(ptr){
          print_control(cl, cd, 0, "%s ", ptr);
        } else {
          print_control(cl, cd, 0, "0x%x ", ms->s_hwrep);
        }

        print_control(cl, cd, 0, "checks=%s ", ms->s_repeat ? "on" : "off");

        if(ms->s_repeat){
          extra = "pressed=<";
          for(i = 0; i < KEY_MAX; i++){
            if(ms->s_repeat[i] & MXK_KEY_HELDDOWN){
              ptr = code_to_name(mxk_key_codes, i);
              if(ptr){
                print_control(cl, cd, 0, "%s%s", extra, ptr);
              } else {
                print_control(cl, cd, 0, "%skey:%d", extra, i);
              }
              extra = " ";
            }
          }
          if(!strcmp(extra, " ")){
            print_control(cl, cd, 0, "> ");
          }
        }

        print_control(cl, cd, 0, "capabilities=");
        delim = '<';
        for(i = 0; i < EV_MAX; i++){
          if(mxk_has_bit(ms->s_mask, EV_MAX, i)){
            ptr = code_to_name(mxk_event_types, i);
            if(ptr){
              print_control(cl, cd, 0, "%c%s", delim, ptr);
            } else {
              print_control(cl, cd, 0, "%c0x%x", delim, i);
            }
            delim = ' ';
          }
        }
        if(delim == ' '){
          print_control(cl, cd, 0, ">\n");
        } else {
          print_control(cl, cd, 0, "none\n");
        }
        count++;
      }
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

static int mxk_list_sources(struct list_loop *ls, struct task_loop *tl)
{
  struct stat st;
  struct control_loop *cl;
  struct dirent *de;
  DIR *dh;
  struct mxk_source *ms, msrc;
  char path[MXK_NAME_SIZE];
  int fd, cd, count;
  char *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  count = 0;

  dh = opendir(MXK_SOURCE_DIR);
  if(dh == NULL){
    print_control(cl, cd, 1, "source: open of %s failed: %s\n", MXK_SOURCE_DIR, strerror(errno));
    return -1;
  }

  do{
    de = readdir(dh);
    if(de){
      if(!strncmp(de->d_name, "event", 5)){

        snprintf(path, MXK_NAME_SIZE - 1, "%s/%s", MXK_SOURCE_DIR, de->d_name);
        path[MXK_NAME_SIZE - 1] = '\0';
        if(stat(path, &st) == 0){
          ms = mxk_is_active(ls, st.st_rdev);
          if(ms){
            ptr = code_to_name(mxk_state_codes, ms->s_state);
            if(ptr == NULL){
              ptr = "broken";
            }
            print_control(cl, cd, 0, "%s: %s: %s\n", path, ptr, ms->s_name);
            count++;
          } else {
            ms = &msrc;
            fd = mxk_probe_source(ls, tl, ms, path, 0, MXK_PROBE_SHOW);
            if(fd >= 0){
              if(mxk_is_target(ms->s_name)){
                print_control(cl, cd, 0, "%s: target: %s\n", path, ms->s_name);
              } else {
                print_control(cl, cd, 0, "%s: unused: %s\n", path, ms->s_name);
              }
              close(fd);
              count++;
            }
          }
        } else {
          switch(errno){
            case ENOENT :
            case ENODEV :
              break;
            default :
              print_control(cl, cd, 0, "%s: error: %s\n", path, strerror(errno));
              break;
          }
        }

      }
    }
  } while(de);

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  closedir(dh);

  return 0;
}

/*************************************************************************/

#define READ_GROUP 8

static int mxk_run_source(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event vt[READ_GROUP];

  struct input_event *vp;
  struct mxk_global *mg;
  struct mxk_source *ms;
  struct link_loop *nk;
  int fd, wr, problem, sz, rr, i, mx, fail, skip;

  mg = global_loop(ls);
  ms = user_loop(ls, tl);
  fd = fd_loop(ls, tl);
  vp = NULL;
  sz = sizeof(struct input_event);

  if(mask & LO_STOP_MASK){
    mxk_grab_source(ms, fd, MXK_SOURCE_BYPASS);
    mxk_free_source(ms);
    return 0;
  }

  if(mask & LO_READ_MASK){
    /* WARNING: possible kernel bug, may return broken data */
    /* also may not report all releases - see flaky field */
    rr = read(fd, vt, READ_GROUP * sz);
#if DEBUG > 1
    fprintf(stderr, "source[%d]: read %d\n", tag_loop(ls, tl), rr);
#endif
    if((rr > 0) && ((rr % sz) == 0)){
      if(ms->s_state != MXK_SOURCE_BYPASS){
        mx = rr / sz;

        if(mx > ms->s_seq_value){
          ms->s_seq_value = mx;
          ms->s_seq_count = 1;
        } else if (mx == ms->s_seq_value){
          ms->s_seq_count++;
        }

        for(i = 0; i < mx; i++){
          vp = &(vt[i]);

          switch(vp->type){
            case EV_SYN :
              if(ms->s_sync == 0){
                mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, vp->value);
                ms->s_sync = 1;
              }
              break;
            case EV_KEY :
              if(ms->s_repeat && (vp->code < KEY_MAX)){
                switch(vp->value){
                  case 0 : /* release key */
                    if(ms->s_repeat[vp->code] & MXK_KEY_HELDDOWN){
                      mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, 0);
                      ms->s_sync = 0;
                    } else {
                      /* needs to start counting at point where flaky check gets enabled */
                      if(ms->s_presses > MXK_PRESS_HONEYMOON){
                        ms->s_flaky |= MXK_FLAKE_UNPRESSED;
                      } /* else ignore infelicities at startup */
                    }
                    ms->s_repeat[vp->code] &= (~MXK_KEY_HELDDOWN);
                  break;
                  case 1 : /* press key */
                    if(ms->s_repeat[vp->code] & MXK_KEY_HELDDOWN){
                      ms->s_flaky |= MXK_FLAKE_OVERPRESSED;
                      if(ms->s_hwrep != MXK_HWREP_DISCARD){
                        mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, 0);
                        mxk_send_client(ls, tl, ms->s_vector, ms->s_count, EV_SYN, 0, 0);
                        mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, 1);
                        ms->s_sync = 0;
                        ms->s_presses++;
                      }
                    } else {
                      mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, vp->value);
                      ms->s_sync = 0;
                      ms->s_presses++;
                    }
                    ms->s_repeat[vp->code] |= MXK_KEY_HELDDOWN;
                  break;
                  default : /* assume repeat */
                  switch(ms->s_hwrep){
                    case MXK_HWREP_DISCARD :
                      break;
                    case MXK_HWREP_FORWARD :
                      mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, vp->value);
                      ms->s_sync = 0;
                      break;
                    case MXK_HWREP_REPLACE :
                      mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, 0);
                      mxk_send_client(ls, tl, ms->s_vector, ms->s_count, EV_SYN, 0, 0);
                      mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, 1);
                      ms->s_sync = 0;
                      break;
                    }
                    /* NOTE: doesauto not used at the moment */
                    /* may be useful to recover from keys falsely */
                    /* reported as pressed */
                    ms->s_repeat[vp->code] |= MXK_KEY_DOESAUTO;
                  break;
                }
              } else {
                if(vp->value == 1){
                  ms->s_presses++;
                }
                mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, vp->value);
                ms->s_sync = 0;
              }
              break;
            default :
#ifdef DEBUG
              fprintf(stderr, "source[%d]: read[%d]: type=0x%02hx, code=0x%02hx, value=0x%x (forward to %d)\n", tag_loop(ls, tl), i, vp->type, vp->code, vp->value, ms->s_count);
#endif
              mxk_send_client(ls, tl, ms->s_vector, ms->s_count, vp->type, vp->code, vp->value);
              ms->s_sync = 0;
              break;
          }

        }
      }
    } else {
      if(rr > 0){
#ifdef DEBUG
        fprintf(stderr, "source[%d]: kernel flakiness: ignoring partial read %d\n", tag_loop(ls, tl), rr);
#endif
        ms->s_flaky |= MXK_FLAKE_PARTIAL;
        fail = 0;
      } else if(rr == 0){
        fail = 1;
      } else {
        switch(errno){
          case EAGAIN :
          case EINTR  :
            fail = 0;
            break;
          default :
#ifdef DEBUG
            fprintf(stderr, "source[%d]: read failed: %s\n", tag_loop(ls, tl), strerror(errno));
#endif
            fail = 1;
            break;
        }
      }
      if(fail){
        mxk_write_watch(ls, tl, 0, "#error: source[%d] (%s) unreadable: disabling\n");
        stop_loop(ls, tl);
      }
    }
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
#ifdef DEBUG
        vp = nk->k_data;
        fprintf(stderr, "source[%d]: link: type=0x%02x code=0x%02x value=0x%x\n", tag_loop(ls, tl), vp->type, vp->code, vp->value);
#endif
        skip = 0;
        if(ms->s_state == MXK_SOURCE_BYPASS){
          skip = 1;
        }

#if 0 /* Gah, LED caps not always reported ? */
        switch(vp->type){
          case EV_LED :
            if(mxk_has_bit(ms->s_leds, LED_MAX, vp->code) == 0){
#ifdef DEBUG
              fprintf(stderr, "source[%d]: do not have %s led (0x%x): discarding\n", tag_loop(ls, tl), code_to_name(mxk_led_codes, vp->code), vp->code);
#endif
              skip = 1;
            }
            break;
          default :
            if(mxk_has_bit(ms->s_mask, EV_MAX, vp->type) == 0){
#ifdef DEBUG
              fprintf(stderr, "source[%d]: do not have %s capability (0x%x): discarding\n", tag_loop(ls, tl), code_to_name(mxk_event_types, vp->type), vp->type);
#endif
              skip = 1;
            }
          break;
        }
#endif

        if(skip == 0){
          problem = 1;
          wr = write(fd, nk->k_data, sizeof(struct input_event));
          if(wr == sizeof(struct input_event)){
            problem = 0;
          } else if(wr < 0){
            switch(errno){
              case EAGAIN :
              case EINTR :
                problem = 0;
                break;
            }
          }
          if(problem){
            mxk_write_watch(ls, tl, 0, "#error: source[%d] (%s) unwriteable : disabling\n");
            stop_loop(ls, tl);
          }
        }
      } else /* if ... */ {
#ifdef DEBUG
        fprintf(stderr, "source: received odd link type %d\n", nk->k_type);
#endif
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*************************************************************************/

int mxk_insert_source(void *source, int type, int id)
{
  struct mxk_source *ms;
  struct mxk_client *mc;

  ms = source;

  mc = mxk_insert_client(ms->s_vector, ms->s_count, type, id);
  if(mc == NULL){
    return -1;
  }

  ms->s_vector = mc;
  ms->s_count++;

  return 0;
}

int mxk_startup_source(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "source", MXK_EVT_SOURCE, -1, mxk_insert_source);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_source(struct mxk_global *mg)
{
}
