/*
 * (c) Marc Welz 2005, 2006, 2007, 2011:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <linux/input.h>
#include <linux/uinput.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

struct mxk_abs{
  unsigned int a_axis;
  int a_max;
  int a_min;
  int a_fuzz;
  int a_flat;
};

struct mxk_target{
  int t_syn;
  struct uinput_user_dev t_device;

  /* something about type of key selection would be nice here */
  
  unsigned int *t_keys;
  unsigned int t_ksize;

  unsigned int *t_relative;
  unsigned int t_rsize;

  struct mxk_abs *t_absolute;
  unsigned int t_asize;
};

static struct mxk_target *mxk_create_target();
static void mxk_destroy_target(struct mxk_target *mt, int fd);

static int mxk_parse_target(struct mxk_target *mt, char *buffer);
static int mxk_default_target(struct mxk_target *mt);

static int mxk_run_target(struct list_loop *ls, struct task_loop *tl, int mask);

static int mxk_start_target(struct control_loop *cl, int cd, struct mxk_target *mt, int id);

/***********************************************************************/

static struct mxk_target *mxk_create_target()
{
  struct mxk_target *mt;

  mt = malloc(sizeof(struct mxk_target));
  if(mt == NULL){
    return NULL;
  }


  mt->t_keys = NULL;
  mt->t_relative = NULL;
  mt->t_absolute = NULL;

  mt->t_ksize = 0;
  mt->t_asize = 0;
  mt->t_rsize = 0;

  mt->t_syn = 0;

  return mt;
}

static void mxk_destroy_target(struct mxk_target *mt, int fd)
{
  if(mt){
    if(mt->t_relative){
      free(mt->t_relative);
      mt->t_relative = NULL;
    }
    mt->t_rsize = 0;

    if(mt->t_absolute){
      free(mt->t_absolute);
      mt->t_absolute = NULL;
    }
    mt->t_asize = 0;

    if(mt->t_keys){
      free(mt->t_keys);
      mt->t_keys = NULL;
    }
    mt->t_ksize = 0;

    free(mt);
  }

  if(fd >= 0){
    ioctl(fd, UI_DEV_DESTROY);
  }
}

static int mxk_addabs_target(struct mxk_target *mt, char *abs)
{
  int axis;
  struct mxk_abs *tmp;
  char *ptr, *end;

#ifdef DEBUG
  fprintf(stderr, "target: considering absolute <%s>\n", abs);
#endif

  ptr = strchr(abs, '@');
  if(ptr == NULL){
    return -1;
  }

  *ptr = '\0';
  axis = name_to_code(mxk_abs_codes, abs);
  *ptr = '@';
  ptr++;

  if(axis < 0){
    return -1;
  }

  tmp = realloc(mt->t_absolute, sizeof(struct mxk_abs) * (mt->t_asize + 1));
  if(tmp == NULL){
    return -1;
  }

  mt->t_absolute = tmp;
  tmp = &(mt->t_absolute[mt->t_asize]);

  tmp->a_axis = axis;

  tmp->a_min = strtoul(ptr, &end, 10);
  if((ptr == end) || (*end != ',')){
    return -1;
  }

  ptr = end + 1;

  tmp->a_max = strtoul(ptr, &end, 10);
  if((ptr == end) || ((*end != '(') && (*end != '\0'))){
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr, "target: %d-%d\n", tmp->a_min, tmp->a_max);
#endif

  if(tmp->a_max <= tmp->a_min){
    return -1;
  }

  if(*end == '\0'){
    tmp->a_fuzz = 0;
    tmp->a_flat = 0;
    mt->t_asize++;
    return 0;
  }

  ptr = end + 1;
  tmp->a_fuzz = strtoul(ptr, &end, 10);
  if((ptr == end) || (*end != ',')){
    return -1;
  }

  ptr = end + 1;
  tmp->a_flat = strtoul(ptr, &end, 10);
  if((ptr == end) || (*end != ')')){
    return -1;
  }

#ifdef DEBUG
fprintf(stderr, "target: got %d,%d(%d,%d)\n", tmp->a_min, tmp->a_max, tmp->a_fuzz, tmp->a_flat);
#endif

  /* TODO: some checks for sane fuzz and flat ?  */

  mt->t_asize++;

  return 0;
}

static int mxk_addkey_target(struct mxk_target *mt, char *key)
{
  int code;
  unsigned int *tmp;

  code = name_to_code(mxk_key_codes, key);
  if(code < 0){
    return -1;
  }

  tmp = realloc(mt->t_keys, sizeof(unsigned int) * (mt->t_ksize + 1));
  if(tmp == NULL){
    return -1;
  }

  mt->t_keys = tmp;
  mt->t_keys[mt->t_ksize] = code;
  mt->t_ksize++;

  return 0;
}

static int mxk_addrel_target(struct mxk_target *mt, char *rel)
{
  int axis;
  unsigned int *tmp;

  axis = name_to_code(mxk_rel_codes, rel);
  if(axis < 0){
    return -1;
  }

  tmp = realloc(mt->t_relative, sizeof(unsigned int) * (mt->t_rsize + 1));
  if(tmp == NULL){
    return -1;
  }

  mt->t_relative = tmp;
  mt->t_relative[mt->t_rsize] = axis;
  mt->t_rsize++;

  return 0;
}

static int mxk_parse_target(struct mxk_target *mt, char *buffer)
{
  char *args, *type;
  int code;

  type = buffer;

  args = strchr(type, ':');
  if(args == NULL){
    return -1;
  }
  
  *args = '\0';
  code = name_to_code(mxk_event_types, type);
  *args = ':'; /* undo, so that error messages are more useful */
  args++;

  switch(code){
    case EV_REL :
      return mxk_addrel_target(mt, args);
    case EV_ABS :
      return mxk_addabs_target(mt, args);
    case EV_KEY : 
      return mxk_addkey_target(mt, args);
      return 0;
    default :
      return -1;
  }
}

/* if we don't parse things, we pick a default set */

static int mxk_default_target(struct mxk_target *mt)
{
  mt->t_relative = malloc(sizeof(unsigned int) * 2);
  if(mt->t_relative == NULL){
    return -1;
  }

  mt->t_rsize = 2;

  mt->t_relative[0] = REL_X;
  mt->t_relative[1] = REL_Y;

  return 0;
}

/***********************************************************************/

int mxk_target_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_global *mg;
  struct mxk_target *mt;
  char *value, *end;
  char buffer[MXK_NAME_SIZE];
  int fd, cd, idx, type, id;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "target");
  if(type <= 0){
    print_control(cl, cd, 1, "target: type not found\n");
    return 0;
  }

  if(find_task_tag_loop(ls, NULL, NULL, type, 0)){
    print_control(cl, cd, 0, "target: starting multiple targets is bad form\n");
  }

  mt = mxk_create_target();
  if(mt == NULL){
    print_control(cl, cd, 1, "target: unable to allocate state\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx >= 0){
    value = args + idx;
    id = strtoul(value, &end, 10);
    if((*end != ' ') && (*end != '\0')){
      print_control(cl, cd, 1, "target: optional first parameter needs to be an integer\n");
      mxk_destroy_target(mt, -1);
      return 0;
    }

    while((idx = next_word(args, idx)) > 0){
      if(load_word(buffer, MXK_NAME_SIZE - 1, args, idx) < 0){
        print_control(cl, cd, 1, "target: missing or malformed argument\n");
        mxk_destroy_target(mt, -1);
        return 0;
      }

      if(mxk_parse_target(mt, buffer)){
        print_control(cl, cd, 1, "target: unable to parse output capability %s\n", buffer);
        mxk_destroy_target(mt, -1);
        return 0;
      }
    }
  } else {
    id = 0;
    if(mxk_default_target(mt)){
      print_control(cl, cd, 1, "target: unable to set up default output capabilities\n");
      mxk_destroy_target(mt, -1);
      return 0;
    }
  }

  fd = mxk_start_target(cl, cd, mt, id);
  if(fd < 0){
    mxk_destroy_target(mt, -1);
    return 0;
  }

  tx = add_tag_loop(ls, fd, mt, LO_WAIT_MASK | LO_READ_MASK | LO_STOP_MASK, mxk_run_target, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "unable to start target task\n");
    mxk_destroy_target(mt, fd);
    close(fd);
    return 0;
  }

  text_loop(ls, tx, "target device=%s", mt->t_device.name);

  return 0;
}

/***************************************************************/

static int mxk_start_target(struct control_loop *cl, int cd, struct mxk_target *mt, int id)
{
#if 0
  /* target looks like a keyboard/mouse combination - could be extended */
  unsigned int events[] = { EV_SYN, EV_KEY, EV_REL, EV_MAX };
  unsigned int mouse[] = { REL_X, REL_Y, REL_Z, REL_HWHEEL, REL_DIAL, REL_WHEEL, REL_MISC, REL_MAX};
#endif
  unsigned int events[4];
  char *devices[] = { MXK_SOURCE_DIR "/uinput", "/dev/misc/uinput", "/dev/uinput", NULL };
  int error, fd, i, result, axis, size;

  for(fd = (-1), i = 0; devices[i] && (fd < 0) ; i++){
    if(i == 1){
      error = errno;
    }
    fd = open(devices[i], O_RDWR | O_NONBLOCK);
  }

  if(fd < 0){
    print_control(cl, cd, 1, "unable to open user input %s: %s\n", devices[0], strerror(errno));
    return -1;
  }

  fcntl(fd, F_SETFD, FD_CLOEXEC);

#ifdef DEBUG
  memset(&mt->t_device, 0, sizeof(struct uinput_user_dev));
#endif

  snprintf(mt->t_device.name, UINPUT_MAX_NAME_SIZE, "MXK/%s (%d:%d)", VERSION, getpid(), id);
  mt->t_device.name[UINPUT_MAX_NAME_SIZE - 1] = '\0';

  mt->t_device.id.bustype = 0;
  mt->t_device.id.vendor = 0;
  mt->t_device.id.product = 0;
  mt->t_device.id.version = 0;

  mt->t_device.ff_effects_max = 0;

  size = 0;
  events[size++] = EV_SYN;
  events[size++] = EV_KEY;
  if(mt->t_rsize > 0){
    events[size++] = EV_REL;
  }
  if(mt->t_asize > 0){
    events[size++] = EV_ABS;
  }

  /* events */
  for(i = 0; i < size; i++){
    if(ioctl(fd, UI_SET_EVBIT, events[i])){
      print_control(cl, cd, 1, "target: unable to enable event 0x%02x: %s\n", events[i], strerror(errno));
      close(fd);
      return -1;
    }
  }

  /* keys */
  if(mt->t_keys != NULL){
    for(i = 0; i < mt->t_ksize; i++){
      if(ioctl(fd, UI_SET_KEYBIT, mt->t_keys[i])){
        print_control(cl, cd, 1, "target: unable to enable key 0x%02x: %s\n", mt->t_keys[i], strerror(errno));
        close(fd);
        return -1;
      }
    }
  } else {
    for(i = 0; mxk_key_codes[i].code < KEY_MAX; i++){
      if(ioctl(fd, UI_SET_KEYBIT, mxk_key_codes[i].code)){
        print_control(cl, cd, 1, "target: unable to enable key 0x%02x (%s): %s\n", mxk_key_codes[i].code, mxk_key_codes[i].name, strerror(errno));
        close(fd);
        return -1;
      }
    }
  }

  /* relative setup */
  for(i = 0; i < mt->t_rsize; i++){
    if(ioctl(fd, UI_SET_RELBIT, mt->t_relative[i])){
      print_control(cl, cd, 1, "target: unable to enable relative motion [%d]=0x%02x: %s\n", i, mt->t_relative[i], strerror(errno));
      close(fd);
      return -1;
    }
  }

  /* absolute setup */
  for(i = 0; i < ABS_MAX; i++){
    mt->t_device.absmax[i] = 0;
    mt->t_device.absmin[i] = 0;
    mt->t_device.absfuzz[i] = 0;
    mt->t_device.absflat[i] = 0;
  }
  for(i = 0; i < mt->t_asize; i++){
    axis = mt->t_absolute[i].a_axis;
    mt->t_device.absmax[axis]  = mt->t_absolute[i].a_max;
    mt->t_device.absmin[axis]  = mt->t_absolute[i].a_min;
    mt->t_device.absfuzz[axis] = mt->t_absolute[i].a_fuzz;
    mt->t_device.absflat[axis] = mt->t_absolute[i].a_flat;

    if(ioctl(fd, UI_SET_ABSBIT, axis)){
      print_control(cl, cd, 1, "target: unable to enable absolute motion 0x%02x: %s\n", axis, strerror(errno));
      close(fd);
      return -1;
    }
  }

  result = write(fd, &mt->t_device, sizeof(struct uinput_user_dev));
  if(result != sizeof(struct uinput_user_dev)){
    print_control(cl, cd, 1, "unable to write configuration: %s\n", (result < 0) ? strerror(errno) : "short write");
    close(fd);
    return -1;
  }

#if 0 
  /* shouldn't this be this instead ? */
  value = 1;
  ioctl(fd, UI_DEV_CREATE, &value);
#endif

  if(ioctl(fd, UI_DEV_CREATE)){
    print_control(cl, cd, 1, "unable to create target: %s\n", strerror(errno));
    close(fd);
    return -1;
  }

  return fd;
}

#if 0
struct task_loop *mxk_start_target(struct list_loop *ls, struct task_loop *tl)
{
  /* target looks like a keyboard/mouse combination - could be extended */
  unsigned int events[] = { EV_SYN, EV_KEY, EV_REL, EV_MAX };
  unsigned int mouse[] = { REL_X, REL_Y, REL_Z, REL_HWHEEL, REL_DIAL, REL_WHEEL, REL_MISC, REL_MAX};
  char *devices[] = { MXK_SOURCE_DIR "/uinput", "/dev/misc/uinput", NULL };
  struct mxk_target *mt;
  struct mxk_global *mg;
  struct control_loop *cl;
  struct task_loop *tx;
  int cd, fd, i, result, type;
#ifdef DEBUG
  int value;
#endif

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);
  type = lookup_task_type_loop(ls, "target");

  if(find_task_tag_loop(ls, NULL, NULL, type, 0)){
    print_control(cl, cd, 1, "will not start target: a target is already running\n");
    return NULL;
  }

  for(fd = (-1), i = 0; devices[i] && (fd < 0) ; i++){
    fd = open(devices[i], O_RDWR | O_NONBLOCK);
  }

  if(fd < 0){
    print_control(cl, cd, 1, "unable to open user input: %s\n", strerror(errno));
    return NULL;
  }

  mt = malloc(sizeof(struct mxk_target));
  if(mt == NULL){
    print_control(cl, cd, 1, "unable to allocate target: %s\n", strerror(errno));
    close(fd);
    return NULL;
  }

  fcntl(fd, F_SETFD, FD_CLOEXEC);

#ifdef DEBUG
  memset(&mt->t_device, 0, sizeof(struct uinput_user_dev));
#endif

  snprintf(mt->t_device.name, UINPUT_MAX_NAME_SIZE, "MXK/%s (%d)", VERSION, getpid());
  mt->t_device.name[UINPUT_MAX_NAME_SIZE - 1] = '\0';

  mt->t_device.id.bustype = 0;
  mt->t_device.id.vendor = 0;
  mt->t_device.id.product = 0;
  mt->t_device.id.version = 0;

  mt->t_device.ff_effects_max = 0;

  for(i = 0; i < ABS_MAX; i++){
    mt->t_device.absmax[i] = 0;
    mt->t_device.absmin[i] = 0;
    mt->t_device.absfuzz[i] = 0;
    mt->t_device.absflat[i] = 0;
  }

  for(i = 0; events[i] < EV_MAX; i++){
    if(ioctl(fd, UI_SET_EVBIT, events[i])){
      print_control(cl, cd, 1, "target: unable to enable event 0x%02x: %s\n", events[i], strerror(errno));
      mxk_destroy_target(mt);
      close(fd);
      return NULL;
    }
  }

  for(i = 0; mxk_key_codes[i].code < KEY_MAX; i++){
    if(ioctl(fd, UI_SET_KEYBIT, mxk_key_codes[i].code)){
      print_control(cl, cd, 1, "target: unable to enable key 0x%02x (%s): %s\n", mxk_key_codes[i].code, mxk_key_codes[i].name, strerror(errno));
      mxk_destroy_target(mt);
      close(fd);
      return NULL;
    }
  }

  for(i = 0; mouse[i] != REL_MAX; i++){
    if(ioctl(fd, UI_SET_RELBIT, mouse[i])){
      print_control(cl, cd, 1, "target: unable to enable relative motion 0x%02x: %s\n", mouse[i], strerror(errno));
      mxk_destroy_target(mt);
      close(fd);
      return NULL;
    }
  }

  result = write(fd, &mt->t_device, sizeof(struct uinput_user_dev));
  if(result != sizeof(struct uinput_user_dev)){
    print_control(cl, cd, 1, "unable to write configuration: %s\n", (result < 0) ? strerror(errno) : "short write");
    mxk_destroy_target(mt);
    close(fd);
    return NULL;
  }

#ifdef DEBUG
  value = 0;
  if(ioctl(fd, UI_DEV_CREATE, &value)){
#else
  if(ioctl(fd, UI_DEV_CREATE, 0)){
#endif
    print_control(cl, cd, 1, "unable to create target: %s\n", strerror(errno));
    mxk_destroy_target(mt);
    close(fd);
    return NULL;
  }

  tx = add_tag_loop(ls, fd, mt, LO_WAIT_MASK | LO_READ_MASK | LO_STOP_MASK, mxk_run_target, NULL, type, 0);
  if(tx == NULL){
    print_control(cl, cd, 1, "unable to start target task\n");
    mxk_destroy_target(mt);
    close(fd);
    return NULL;
  }

  text_loop(ls, tx, "target device=%s", mt->t_device.name);

  return tx;
}
#endif

static int mxk_run_target(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct input_event vt;
  struct input_event *vp;
  struct link_loop *nk;
  struct mxk_target *mt;
  struct mxk_global *mg;
  int fd, rr, wr, use;

  mg = global_loop(ls);
  mt = user_loop(ls, tl);
  fd = fd_loop(ls, tl);

  if(mask & LO_STOP_MASK){
    mxk_destroy_target(mt, fd);
    return 0;
  }

  if(mask & LO_READ_MASK){
    rr = read(fd, &vt, sizeof(struct input_event));
#ifdef DEBUG
    if(rr <= 0){
      fprintf(stderr, "target: reverse synthetic event: failure\n");
    } else {
      fprintf(stderr, "target: discarding reverse synthetic event: %d, %d, %d\n", vt.type, vt.code, vt.value);
    }
#endif
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;

        switch(vp->type){ /* sanity checks */
          case EV_SYN :
            if(mt->t_syn){
              use = 0;
            } else {
              use = 1;
              mt->t_syn = 1;
            }
            break;
          case EV_KEY :
            use = (vp->code < KEY_MAX) ? 1 : 0;
            mt->t_syn = 0;
            break;
          case EV_REL :
            use = 1;
            mt->t_syn = 0;
            break;
          case EV_ABS :
            use = 1;
            mt->t_syn = 0;
            break;
          default :
            use = 0;
            mt->t_syn = 0;
#ifdef DEBUG
            fprintf(stderr, "target: ignoring event type 0x%x\n", vp->type);
#endif
            break;
        }

        if(use){
          gettimeofday(&(vp->time), NULL);
          wr = write(fd, vp, sizeof(struct input_event));
          if(wr < sizeof(struct input_event)){
#ifdef DEBUG
            fprintf(stderr, "target: write failed: %d/%d: %s\n", wr, sizeof(struct input_event), (wr < 0) ? strerror(errno) : "short write");
#endif
            if(wr < 0){
              switch(errno){
                case EAGAIN :
                case EINTR  :
                  break;
                default:
                  stop_loop(ls, tl);
                  break;
              }
            } else { /* partial writes will confuse things, so give up */
              stop_loop(ls, tl);
            }
          }
        }
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

int mxk_startup_target(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "target", MXK_EVT_TARGET, 1, NULL);
  if(type <= 0){
    return -1;
  }

  return 0;
}

