/*
 * (c) Marc Welz 2005, 2006, 2007:
 *
 * Only to be distributed under the conditions of the GNU GPL v3 or newer
 * as published by the Free Software Foundation, with all anti-DRM and
 * anti-patent measures enabled
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <linux/input.h>
#include <linux/kdev_t.h>

#include <loop.h>
#include <word-work.h>
#include <code-name.h>

#include "mxk.h"

/* timster module */

#define TIMSTER_IDLE   0
#define TIMSTER_ACTIVE 1

#define TIMSTER_KEY_TRIGGER 0
#define TIMSTER_KEY_RESET   1
#define TIMSTER_KEY_CLEAR   2
#define TIMSTER_MAX_KEYS    3

static struct code_name mxk_timster_keys[] = {
  { "tigger",  TIMSTER_KEY_TRIGGER },
  { "reset",   TIMSTER_KEY_RESET },
  { "clear",   TIMSTER_KEY_CLEAR },

  { NULL, -1 },
};

struct mxk_timster{
  int t_keys[TIMSTER_MAX_KEYS];

#if 0
  int t_state;
#endif
  int t_auto;
  int t_arm;

  int t_relay; /* key to emit on timer expiry */
  struct timeval t_period;

  struct mxk_client *t_vector;
  int t_count;
};

static void mxk_destroy_timster(struct mxk_timster *timster);
static struct mxk_timster *mxk_create_timster();
static int mxk_run_timster(struct list_loop *ls, struct task_loop *tl, int mask);

/**********************************************************************/

struct mxk_timster *mxk_create_timster(unsigned int period, int relay)
{
  struct mxk_timster *timster;
  int i;

  timster = malloc(sizeof(struct mxk_timster));
  if(timster == NULL){
    return NULL;
  }

  mxk_milli_to_tv(&(timster->t_period), period);
  timster->t_relay = relay;
  timster->t_auto = 0;
  timster->t_arm = 0;

  for(i = 0; i < TIMSTER_MAX_KEYS; i++){
    timster->t_keys[i] = KEY_RESERVED;
  }

  timster->t_vector = NULL;
  timster->t_count = 0;

  return timster;
}

static void mxk_destroy_timster(struct mxk_timster *timster)
{
  if(timster == NULL){
    return;
  }

  mxk_shutdown_client(timster->t_vector, timster->t_count);
#ifdef DEBUG
  timster->t_vector = NULL;
  timster->t_count = 0;
#endif

  free(timster);
}

/**********************************************************************/

int mxk_auto_timster(struct list_loop *ls, struct task_loop *tx, int autorestart, char *args)
{
  struct mxk_timster *timster;

  timster = user_loop(ls, tx);
  timster->t_auto = autorestart;

  return 0;
}

int mxk_arm_timster(struct list_loop *ls, struct task_loop *tx, int arm, char *args)
{
  struct mxk_timster *timster;

  timster = user_loop(ls, tx);

  if(arm){
    set_time_loop(ls, tx, &(timster->t_period));
    set_mask_loop(ls, tx, LO_TIME_MASK, 1);
  } else {
    set_mask_loop(ls, tx, LO_TIME_MASK, -1);
  }

  timster->t_arm = arm;

  return 0;
}

int mxk_keys_timster(struct list_loop *ls, struct task_loop *tx, int index, char *args)
{
  struct mxk_timster *timster;
  char *ptr;
  int idx, key;

  timster = user_loop(ls, tx);

  if((index < 0) || (index >= TIMSTER_MAX_KEYS)){
    return -1;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    return -1;
  }

  ptr = args + idx;

  idx = end_word(args, idx);
  if(idx >= 0){
    args[idx] = '\0';
  }

  key = name_to_code(mxk_key_codes, ptr);
  if(key < 0){
    return -1;
  }


  timster->t_keys[index] = key;

  return 0;
}

/**********************************************************************/

int mxk_start_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  struct control_loop *cl;
  struct mxk_timster *timster;
  struct mxk_global *mg;
  struct task_loop *tx;
  int cd, idx, id, type;
  unsigned int period;
  int relay;
  char *end, *ptr;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);
  mg = global_loop(ls);

  type = lookup_task_type_loop(ls, "timer");
  if(type <= 0){
    print_control(cl, cd, 1, "timer: type not found\n");
    return 0;
  }

  idx = (-1);
  id = 0;
  if(args == NULL){
    print_control(cl, cd, 1, "timer: insufficient parameters\n");
    return 0;
  }

  idx = start_word(args, 0);
  if(idx < 0){
    print_control(cl, cd, 1, "timer: unable to acquire id\n");
    return 0;
  }

  id = strtoul(args + idx, &end, 10);

  idx = next_word(args, idx);
  if(idx < 0){
    print_control(cl, cd, 1, "timer: unable to acquire millisecond period\n");
    return 0;
  }

  if(*end != ' '){
    print_control(cl, cd, 1, "timer: require multiple parameters, first parameter needs to be an integer id\n");
    return 0;
  }

  period = strtoul(args + idx, &end, 10);

  idx = next_word(args, idx);
  if(idx < 0){
    print_control(cl, cd, 1, "timer: unable to acquire relayed key event\n");
    return 0;
  }

  if(*end != ' '){
    print_control(cl, cd, 1, "timer: require multiple parameters, second parameter needs to be integer millisecond period\n");
    return 0;
  }

  ptr = args + idx;
  idx = end_word(args, idx);
  if(idx > 0){
    args[idx] = '\0';
  }

  relay = name_to_code(mxk_key_codes, ptr);
  if(relay < 0){
    print_control(cl, cd, 1, "timer: unknown key <%s>\n", ptr);
    return 0;
  }

  timster = mxk_create_timster(period, relay);
  if(timster == NULL){
    print_control(cl, cd, 1, "timer: unable to allocate state\n");
    return 0;
  }

  tx = add_tag_loop(ls, -1, timster, LO_STOP_MASK | LO_WAIT_MASK, mxk_run_timster, NULL, type, id);
  if(tx == NULL){
    print_control(cl, cd, 1, "timer: unable to create task\n");
    mxk_destroy_timster(timster);
    return 0;
  }

  text_loop(ls, tx, "timer:%d", id);
  print_control(cl, cd, 1, "timer: created task\n");

  return 0;
}

int mxk_list_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  int cd, idx, id, i, type, count;
  struct control_loop *cl;
  struct task_loop *tx;
  struct mxk_timster *timster;
  char *ptr, *tmp;

  cl = user_loop(ls, tl);
  cd = fd_loop(ls, tl);

  type = lookup_task_type_loop(ls, "timer");
  if(type <= 0){
    print_control(cl, cd, 1, "timer: type not available\n");
    return 0;
  }

  id = 0;
  if(args){
    idx = start_word(args, 0);
    if(idx >= 0){
      id = strtoul(args + idx, NULL, 10);
    }
  }

  count = 0;
  for(tx = find_task_tag_loop(ls, NULL, NULL, type, id); tx; tx = find_task_tag_loop(ls, NULL, tx, type, id)){
    timster = user_loop(ls, tx);

    if(timster){
      print_control(cl, cd, 0, "timer[%d]: period=%ld.%06lds", tag_loop(ls, tx), timster->t_period.tv_sec, timster->t_period.tv_usec);

      if((ptr = code_to_name(mxk_key_codes, timster->t_relay)) != NULL){
        print_control(cl, cd, 0, " key=%s", ptr);
      }

      print_control(cl, cd, 0, " trigger=%s", timster->t_auto ? "automatic" : "single");

      print_control(cl, cd, 0, " state=%s", timster->t_arm ? "armed" : "off");

      for(i = 0; i < TIMSTER_MAX_KEYS; i++){
        if((timster->t_keys[i] != KEY_RESERVED) && ((ptr = code_to_name(mxk_key_codes, timster->t_keys[i])) != NULL)){
          tmp = code_to_name(mxk_timster_keys, i);
          if(tmp){
            print_control(cl, cd, 0, " %s:%s", tmp, ptr);
          } else { /* really should not happen */
            print_control(cl, cd, 0, " %d:%s", i, ptr);
          }
        }
      }

      print_control(cl, cd, 0, "\n", ptr);

      count++;
    } else {
      print_control(cl, cd, 1, "timer: unavailable\n", id);
    }
  }

  print_control(cl, cd, 1, "%d %s\n", count, (count == 1) ? "entry" : "entries");

  return 0;
}

int mxk_trigger_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_REQUIRED, TIMSTER_KEY_TRIGGER, mxk_keys_timster);

  return 0;
}

int mxk_reset_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_REQUIRED, TIMSTER_KEY_RESET, mxk_keys_timster);

  return 0;
}

int mxk_clear_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_REQUIRED, TIMSTER_KEY_CLEAR, mxk_keys_timster);

  return 0;
}

int mxk_arm_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_OPTIONAL, 1, mxk_arm_timster);

  return 0;
}

int mxk_off_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_OPTIONAL, 0, mxk_arm_timster);

  return 0;
}

int mxk_auto_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_OPTIONAL, 1, mxk_auto_timster);

  return 0;
}

int mxk_single_timster_command(struct list_loop *ls, struct task_loop *tl, char *args)
{
  mxk_parse_apply(ls, tl, args, "timer", MXK_MISC_OPTIONAL, 0, mxk_auto_timster);

  return 0;
}

/**********************************************************************/

static int mxk_run_timster(struct list_loop *ls, struct task_loop *tl, int mask)
{
  struct mxk_global *mg;
  struct mxk_timster *timster;

  struct input_event *vp;
  struct link_loop *nk;
  int k;

  mg = global_loop(ls);
  timster = user_loop(ls, tl);

  if(timster == NULL){
    stop_loop(ls, tl);
    return 0;
  }

  if(mask & LO_STOP_MASK){
    mxk_destroy_timster(timster);
    return 0;
  }

  if(mask & LO_TIME_MASK){
#ifdef DEBUG
    fprintf(stderr, "timer: got timeout, sneding %d\n", timster->t_relay);
#endif
    mxk_send_client(ls, tl, timster->t_vector, timster->t_count, EV_KEY, timster->t_relay, 1);
    mxk_send_client(ls, tl, timster->t_vector, timster->t_count, EV_SYN, 0, 0);
    mxk_send_client(ls, tl, timster->t_vector, timster->t_count, EV_KEY, timster->t_relay, 0);
    mxk_send_client(ls, tl, timster->t_vector, timster->t_count, EV_SYN, 0, 0);
    if(timster->t_auto){
      set_time_loop(ls, tl, &(timster->t_period));
      set_mask_loop(ls, tl, LO_TIME_MASK, 1);
      timster->t_arm = 1;
    } else {
      set_mask_loop(ls, tl, LO_TIME_MASK, -1);
      timster->t_arm = 0;
    }
  }

  if(mask & LO_WAIT_MASK){
    while((nk = receive_link_loop(ls, tl)) != NULL){
      if(nk->k_type == mg->g_event){
        vp = nk->k_data;
#ifdef DEBUG
        fprintf(stderr, "timer: got event %d,%d,%d\n", vp->type, vp->code, vp->value);
#endif
        if((vp->type == EV_KEY) && (vp->value == 1)){
          for(k = 0; (k < TIMSTER_MAX_KEYS) && (timster->t_keys[k] != vp->code); k++);
          switch(k){
            case TIMSTER_KEY_TRIGGER :
              if(timster->t_arm){
#ifdef DEBUG
                fprintf(stderr, "timer: already active (key=%d)\n", vp->code);
#endif
                break;
              }
              /* WARNING: partial fall */
            case TIMSTER_KEY_RESET   :
#ifdef DEBUG
              fprintf(stderr, "timer: activating (key=%d)\n", vp->code);
#endif
              set_time_loop(ls, tl, &(timster->t_period));
              set_mask_loop(ls, tl, LO_TIME_MASK, 1);
              timster->t_arm = 1;
              break;
            case TIMSTER_KEY_CLEAR   :
#ifdef DEBUG
              fprintf(stderr, "timer: clearing (key=%d)\n", vp->code);
#endif
              set_mask_loop(ls, tl, LO_TIME_MASK, -1);
              timster->t_arm = 0;
              break;
          }
        }
#ifdef DEBUG
      } else {
        fprintf(stderr, "timer: unhandled message type %d\n", nk->k_type);
#endif
      }
      discard_link_loop(ls, nk);
    }
  }

  return 0;
}

/*********************************************************************/

static int mxk_insert_timster(void *state, int type, int id)
{
  struct mxk_timster *timster;
  struct mxk_client *mc;

  timster = state;

  mc = mxk_insert_client(timster->t_vector, timster->t_count, type, id);
  if(mc == NULL){
    return -1;
  }

  timster->t_vector = mc;
  timster->t_count++;

  return 0;
}

int mxk_startup_timster(struct list_loop *ls)
{
  int type;

  type = mxk_enroll_find(ls, "timer", MXK_EVT_SOURCE | MXK_EVT_TARGET, -1, mxk_insert_timster);
  if(type <= 0){
    return -1;
  }

  return 0;
}

void mxk_shutdown_timster(struct mxk_global *mg)
{
}
